import numpy as np
import matplotlib.pyplot as plt
from typing import List
from sklearn.preprocessing import PolynomialFeatures
from sklearn.linear_model import LinearRegression
from sklearn.metrics import r2_score
from scipy.fft import fft



def movmean(x, w):
    return np.convolve(x, np.ones(w), 'same') / w


def rescale(x, l=0, u=1):
    return l + ((x-np.min(x))/(np.max(x)-np.min(x)))*(u-l)


def get_noise_index_vec(fs: float, mtg_labels: List[str], mtg_signals: np.ndarray, notched: bool = False, notch_freqs: np.ndarray = None, is_ieeg:bool=False) -> np.ndarray:
    nr_samples = mtg_signals.shape[1]
    nr_mtgs = mtg_signals.shape[0]
    wdw_length = int(fs*10)
    wdw_step = wdw_length
    start_sample_vec = np.arange(0, nr_samples-wdw_length, wdw_step, dtype=int)

    time_vec = np.zeros(len(start_sample_vec))
    chspec_ni_vec = np.zeros((nr_mtgs, len(start_sample_vec)))
    chavg_ni_vec = np.zeros(len(start_sample_vec))

    poly_order = 6
    ideal_freqs, ideal_spctrm_regression = get_ideal_signal_spectrum(
        fs, wdw_length, poly_order, notched, notch_freqs, is_ieeg)

    for si, start_sample in enumerate(start_sample_vec):
        end_sample = start_sample + wdw_length
        timestamp = start_sample/fs
        data_wdw = mtg_signals[:, start_sample:end_sample]

        results = get_ffts_and_regressions(fs, start_sample, mtg_labels, data_wdw,
                                           poly_order, ideal_freqs, ideal_spctrm_regression, notched, notch_freqs)

        mtgs_ni = getNI(results, ideal_spctrm_regression)

        time_vec[si] = timestamp
        chspec_ni_vec[:, si] = mtgs_ni
        chavg_ni_vec[si] = np.mean(mtgs_ni)
        pass

    return time_vec, chspec_ni_vec, chavg_ni_vec


def getNI(results, ideal_spctrm_regression):
    fft_freqs = results[0]
    all_ch_spctrms = results[1]
    all_ch_spctrm_regs = results[2]
    channs_avg_spctrm = results[3]
    channs_avg_spctrm_reg = results[4]

    nr_mtgs = all_ch_spctrm_regs.shape[0]
    mtgs_ni = np.zeros(nr_mtgs)
    for mi in range(nr_mtgs):

        # NI_1, Channel Spectrum vs Channel Model, catch artefats that don't follow a linear adjustment to the spctrm
        ni_val_1 = 1-r2_score(all_ch_spctrms[mi, :], all_ch_spctrm_regs[mi, :])
        if ni_val_1 > 1:
            ni_val_1 = 1

        # NI_2, Channel Model vs Ch.Average Model, catch artifcats based on a clean spectrum (i.e. ch.avg.spctrm)
        ni_val_2 = 1-r2_score(channs_avg_spctrm_reg, all_ch_spctrm_regs[mi, :])
        if ni_val_2 > 1:
            ni_val_2 = 1

        # NI_3, Channel Model vs Ideal Model, catch artefacts in case all channels artifactual
        # ni_val_3 = 1-r2_score(ideal_spctrm_regression, all_ch_spctrm_regs[mi,:])
        ni_val_3 = 1-r2_score(ideal_spctrm_regression, channs_avg_spctrm_reg)

        if ni_val_3 > 1:
            ni_val_3 = 1

        ni = np.mean([ni_val_1, ni_val_2, ni_val_3])
        mtgs_ni[mi] = ni
        pass
    return mtgs_ni


def get_ffts_and_regressions(fs: float, start_sample: int, mtg_labels: List[str], data_wdw: np.ndarray, poly_order: int, ideal_freqs, ideal_spctrm_regression, notched: bool, notch_freqs: List[int]) -> float:

    nr_samples = data_wdw.shape[1]
    nr_mtgs = data_wdw.shape[0]

    # determine which frequencies to use for extrapolation of the ideal spectrum
    fft_freqs = np.arange(0, np.round(nr_samples/2))*(fs/nr_samples)

    # remove the notched frequencies from the spectrum analysis
    keep_freqs_sel = np.logical_and(fft_freqs >= 1, fft_freqs <= 341.33)
    if notched:
        for hfi in range(0, len(notch_freqs)):
            # range of frequecies of the electro-magnetic interference (EMI)
            emi_freq_low = notch_freqs[hfi, 1]
            emi_freq_high = notch_freqs[hfi, 2]
            sel_vec = np.logical_and(
                fft_freqs >= emi_freq_low, fft_freqs <= emi_freq_high)
            keep_freqs_sel[sel_vec] = False
    fft_freqs = fft_freqs[keep_freqs_sel]
    # plot_sel = np.linspace(0, len(fft_freqs), num=len(ideal_freqs), endpoint=False, dtype=int)
    time = np.arange(start_sample, start_sample+nr_samples)/fs

    # Get ChannSpec and ChannAvg Spectrum Regression
    channs_avg_spctrm = np.zeros((1, len(fft_freqs)))
    all_ch_spctrms = np.zeros((nr_mtgs, len(fft_freqs)))
    all_ch_spctrm_regs = np.zeros((nr_mtgs, len(fft_freqs)))
    movmean_order = 10
    for mi in range(nr_mtgs):
        # Chann Spectrum
        wdw_signal = data_wdw[mi, :]
        wdw_signal = wdw_signal - np.mean(wdw_signal)

        P2 = np.abs(fft(wdw_signal)/nr_samples)
        chann_spctrm = P2[0:int(nr_samples/2)]*2
        # chann_spctrm[1:-2] = 2*chann_spctrm[1:-2]
        chann_spctrm = np.real(chann_spctrm)
        chann_spctrm = chann_spctrm[keep_freqs_sel]

        # Clip possible artifacts from fft at first and last sample
        chann_spctrm[0] = chann_spctrm[1]
        chann_spctrm[-1] = chann_spctrm[-2]

        chann_spctrm = movmean(chann_spctrm, movmean_order)

        # Clip possible artifacts from fft at first and last sample, consider movmean_order
        chann_spctrm[0:movmean_order] = chann_spctrm[movmean_order]
        chann_spctrm[-1*movmean_order:] = chann_spctrm[-1*movmean_order-1]

        chann_spctrm = rescale(np.log10(chann_spctrm))

        # Get estimate of channel spectrum
        poly = PolynomialFeatures(degree=poly_order)
        lin_reg = LinearRegression()
        x_poly = poly.fit_transform(fft_freqs.reshape(-1, 1))
        lin_reg.fit(x_poly, chann_spctrm)
        chDataSpctrm_reg = lin_reg.predict(x_poly)

        all_ch_spctrms[mi, :] = chann_spctrm
        all_ch_spctrm_regs[mi, :] = chDataSpctrm_reg
        channs_avg_spctrm = channs_avg_spctrm+chann_spctrm

        plot_ch_fft = False
        if plot_ch_fft:
            plt.plot(fft_freqs, chann_spctrm, label="Chann Spectrum")
            plt.plot(fft_freqs, chDataSpctrm_reg,
                     label="Chann Spectrum Regression")
            plt.plot(ideal_freqs, ideal_spctrm_regression,
                     label="Ideal Spectrum")
            plt.title('Spectrum')
            plt.xlabel('Frequency[Hz]')
            plt.ylabel('Amplitude')
            plt.legend()
            plt.show()
            plt.close()

    # Model window average spectrum
    channs_avg_spctrm /= nr_mtgs
    channs_avg_spctrm = channs_avg_spctrm.reshape(-1, 1)
    poly = PolynomialFeatures(degree=poly_order)
    lin_reg = LinearRegression()
    x_poly = poly.fit_transform(fft_freqs.reshape(-1, 1))
    lin_reg.fit(x_poly, channs_avg_spctrm)
    channs_avg_spctrm_reg = lin_reg.predict(x_poly)
    ni_val = 1-r2_score(channs_avg_spctrm, channs_avg_spctrm_reg)

    plot_ch_avg_fft = False
    if plot_ch_avg_fft:
        plt.plot(fft_freqs, channs_avg_spctrm, label="Average Spectrum")
        plt.plot(fft_freqs, np.mean(all_ch_spctrm_regs, axis=0),
                 label=f"Average of ch.specific estimations")
        plt.plot(fft_freqs, channs_avg_spctrm_reg,
                 label=f"Estimation of Average Spectrum (NI={ni_val:.2f})")
        plt.plot(ideal_freqs, ideal_spctrm_regression, label="Ideal Spectrum")
        plt.legend()
        plt.title(
            f"Average Spectrum\nWindow {time[0]:.2f}-{time[-1]:.2f} seconds")
        plt.show(block=False)
        plt.close()
        pass

    return (fft_freqs, all_ch_spctrms, all_ch_spctrm_regs, channs_avg_spctrm, channs_avg_spctrm_reg)


################################################################


def get_ideal_signal_spectrum(fs, wdw_length, poly_order, notched=False, harmon_freqs=None, is_ieeg:bool=False):
    nr_samples = wdw_length

    # determine which frequencies to use for extrapolation of the ideal spectrum
    fft_freqs = np.arange(0, np.round(nr_samples/2))*(fs/nr_samples)

    # remove the notched frequencies from the spectrum analysis
    keep_freqs_sel = np.logical_and(fft_freqs >= 1, fft_freqs <= 341.33)
    if notched:
        for hfi in range(0, len(harmon_freqs)):
            # range of frequecies of the electro-magnetic interference (EMI)
            emi_freq_low = harmon_freqs[hfi, 1]
            emi_freq_high = harmon_freqs[hfi, 2]
            sel_vec = np.logical_and(
                fft_freqs >= emi_freq_low, fft_freqs <= emi_freq_high)
            keep_freqs_sel[sel_vec] = False
    fft_freqs = fft_freqs[keep_freqs_sel]

    # get ideal scalp EEG spectrum and determine which frequencies to use for extrapolation
    ideal_freqs, ideal_spctrm = get_perf_EEG_chann_avg_regression_vals()
    
    if is_ieeg:
        ideal_freqs, ideal_spctrm = get_perf_iEEG_chann_avg_regression_vals()

    # remove the notched frequencies from the saved "ideal"  scalp EEG spectrum
    keep_freqs_sel = np.logical_and(ideal_freqs >= 1, ideal_freqs <= 341.33)
    if notched:
        for hfi in range(0, len(harmon_freqs)):
            # range of frequecies of the electro-magnetic interference (EMI)
            emi_freq_low = harmon_freqs[hfi, 1]
            emi_freq_high = harmon_freqs[hfi, 2]
            sel_vec = np.logical_and(
                ideal_freqs >= emi_freq_low, ideal_freqs <= emi_freq_high)
            keep_freqs_sel[sel_vec] = False
    ideal_freqs = ideal_freqs[keep_freqs_sel]
    ideal_spctrm = ideal_spctrm[keep_freqs_sel]

    # Extrapolate ideal spectrum
    poly = PolynomialFeatures(degree=poly_order)
    lin_reg = LinearRegression()
    x_poly = poly.fit_transform(ideal_freqs.reshape(-1, 1))
    lin_reg.fit(x_poly, ideal_spctrm)

    # Extrapolate ideal spectrum
    x_poly = poly.fit_transform(fft_freqs.reshape(-1, 1))
    ideal_spctrm_regression = lin_reg.predict(x_poly)
    ni_val = 1-r2_score(ideal_spctrm, ideal_spctrm_regression)

    plot_ideal_spctrm = False
    if plot_ideal_spctrm:
        plt.plot(ideal_freqs, ideal_spctrm, label="Ideal Spectrum")
        plt.plot(fft_freqs, ideal_spctrm_regression,
                 label=f"Estimation of Ideal Spectrum (NI={ni_val:.2f})")
        plt.legend()
        plt.show(block=False)
        plt.close()

    return ideal_freqs, ideal_spctrm_regression


def get_perf_EEG_chann_avg_regression_vals():
    freqsPerfEEG = [1, 1.10000000000000, 1.20000000000000, 1.30000000000000, 1.40000000000000, 1.50000000000000, 1.60000000000000, 1.70000000000000, 1.80000000000000, 1.90000000000000, 2, 2.10000000000000, 2.20000000000000, 2.30000000000000, 2.40000000000000, 2.50000000000000, 2.60000000000000, 2.70000000000000, 2.80000000000000, 2.90000000000000, 3, 3.10000000000000, 3.20000000000000, 3.30000000000000, 3.40000000000000, 3.50000000000000, 3.60000000000000, 3.70000000000000, 3.80000000000000, 3.90000000000000, 4, 4.10000000000000, 4.20000000000000, 4.30000000000000, 4.40000000000000, 4.50000000000000, 4.60000000000000, 4.70000000000000, 4.80000000000000, 4.90000000000000, 5, 5.10000000000000, 5.20000000000000, 5.30000000000000, 5.40000000000000, 5.50000000000000, 5.60000000000000, 5.70000000000000, 5.80000000000000, 5.90000000000000, 6, 6.10000000000000, 6.20000000000000, 6.30000000000000, 6.40000000000000, 6.50000000000000, 6.60000000000000, 6.70000000000000, 6.80000000000000, 6.90000000000000, 7, 7.10000000000000, 7.20000000000000, 7.30000000000000, 7.40000000000000, 7.50000000000000, 7.60000000000000, 7.70000000000000, 7.80000000000000, 7.90000000000000, 8, 8.10000000000000, 8.20000000000000, 8.30000000000000, 8.40000000000000, 8.50000000000000, 8.60000000000000, 8.70000000000000, 8.80000000000000, 8.90000000000000, 9, 9.10000000000000, 9.20000000000000, 9.30000000000000, 9.40000000000000, 9.50000000000000, 9.60000000000000, 9.70000000000000, 9.80000000000000, 9.90000000000000, 10, 10.1000000000000, 10.2000000000000, 10.3000000000000, 10.4000000000000, 10.5000000000000, 10.6000000000000, 10.7000000000000, 10.8000000000000, 10.9000000000000, 11, 11.1000000000000, 11.2000000000000, 11.3000000000000, 11.4000000000000, 11.5000000000000, 11.6000000000000, 11.7000000000000, 11.8000000000000, 11.9000000000000, 12, 12.1000000000000, 12.2000000000000, 12.3000000000000, 12.4000000000000, 12.5000000000000, 12.6000000000000, 12.7000000000000, 12.8000000000000, 12.9000000000000, 13, 13.1000000000000, 13.2000000000000, 13.3000000000000, 13.4000000000000, 13.5000000000000, 13.6000000000000, 13.7000000000000, 13.8000000000000, 13.9000000000000, 14, 14.1000000000000, 14.2000000000000, 14.3000000000000, 14.4000000000000, 14.5000000000000, 14.6000000000000, 14.7000000000000, 14.8000000000000, 14.9000000000000, 15, 15.1000000000000, 15.2000000000000, 15.3000000000000, 15.4000000000000, 15.5000000000000, 15.6000000000000, 15.7000000000000, 15.8000000000000, 15.9000000000000, 16, 16.1000000000000, 16.2000000000000, 16.3000000000000, 16.4000000000000, 16.5000000000000, 16.6000000000000, 16.7000000000000, 16.8000000000000, 16.9000000000000, 17, 17.1000000000000, 17.2000000000000, 17.3000000000000, 17.4000000000000, 17.5000000000000, 17.6000000000000, 17.7000000000000, 17.8000000000000, 17.9000000000000, 18, 18.1000000000000, 18.2000000000000, 18.3000000000000, 18.4000000000000, 18.5000000000000, 18.6000000000000, 18.7000000000000, 18.8000000000000, 18.9000000000000, 19, 19.1000000000000, 19.2000000000000, 19.3000000000000, 19.4000000000000, 19.5000000000000, 19.6000000000000, 19.7000000000000, 19.8000000000000, 19.9000000000000, 20, 20.1000000000000, 20.2000000000000, 20.3000000000000, 20.4000000000000, 20.5000000000000, 20.6000000000000, 20.7000000000000, 20.8000000000000, 20.9000000000000, 21, 21.1000000000000, 21.2000000000000, 21.3000000000000, 21.4000000000000, 21.5000000000000, 21.6000000000000, 21.7000000000000, 21.8000000000000, 21.9000000000000, 22, 22.1000000000000, 22.2000000000000, 22.3000000000000, 22.4000000000000, 22.5000000000000, 22.6000000000000, 22.7000000000000, 22.8000000000000, 22.9000000000000, 23, 23.1000000000000, 23.2000000000000, 23.3000000000000, 23.4000000000000, 23.5000000000000, 23.6000000000000, 23.7000000000000, 23.8000000000000, 23.9000000000000, 24, 24.1000000000000, 24.2000000000000, 24.3000000000000, 24.4000000000000, 24.5000000000000, 24.6000000000000, 24.7000000000000, 24.8000000000000, 24.9000000000000, 25, 25.1000000000000, 25.2000000000000, 25.3000000000000, 25.4000000000000, 25.5000000000000, 25.6000000000000, 25.7000000000000, 25.8000000000000, 25.9000000000000, 26, 26.1000000000000, 26.2000000000000, 26.3000000000000, 26.4000000000000, 26.5000000000000, 26.6000000000000, 26.7000000000000, 26.8000000000000, 26.9000000000000, 27, 27.1000000000000, 27.2000000000000, 27.3000000000000, 27.4000000000000, 27.5000000000000, 27.6000000000000, 27.7000000000000, 27.8000000000000, 27.9000000000000, 28, 28.1000000000000, 28.2000000000000, 28.3000000000000, 28.4000000000000, 28.5000000000000, 28.6000000000000, 28.7000000000000, 28.8000000000000, 28.9000000000000, 29, 29.1000000000000, 29.2000000000000, 29.3000000000000, 29.4000000000000, 29.5000000000000, 29.6000000000000, 29.7000000000000, 29.8000000000000, 29.9000000000000, 30, 30.1000000000000, 30.2000000000000, 30.3000000000000, 30.4000000000000, 30.5000000000000, 30.6000000000000, 30.7000000000000, 30.8000000000000, 30.9000000000000, 31, 31.1000000000000, 31.2000000000000, 31.3000000000000, 31.4000000000000, 31.5000000000000, 31.6000000000000, 31.7000000000000, 31.8000000000000, 31.9000000000000, 32, 32.1000000000000, 32.2000000000000, 32.3000000000000, 32.4000000000000, 32.5000000000000, 32.6000000000000, 32.7000000000000, 32.8000000000000, 32.9000000000000, 33, 33.1000000000000, 33.2000000000000, 33.3000000000000, 33.4000000000000, 33.5000000000000, 33.6000000000000, 33.7000000000000, 33.8000000000000, 33.9000000000000, 34, 34.1000000000000, 34.2000000000000, 34.3000000000000, 34.4000000000000, 34.5000000000000, 34.6000000000000, 34.7000000000000, 34.8000000000000, 34.9000000000000, 35, 35.1000000000000, 35.2000000000000, 35.3000000000000, 35.4000000000000, 35.5000000000000, 35.6000000000000, 35.7000000000000, 35.8000000000000, 35.9000000000000, 36, 36.1000000000000, 36.2000000000000, 36.3000000000000, 36.4000000000000, 36.5000000000000, 36.6000000000000, 36.7000000000000, 36.8000000000000, 36.9000000000000, 37, 37.1000000000000, 37.2000000000000, 37.3000000000000, 37.4000000000000, 37.5000000000000, 37.6000000000000, 37.7000000000000, 37.8000000000000, 37.9000000000000, 38, 38.1000000000000, 38.2000000000000, 38.3000000000000, 38.4000000000000, 38.5000000000000, 38.6000000000000, 38.7000000000000, 38.8000000000000, 38.9000000000000, 39, 39.1000000000000, 39.2000000000000, 39.3000000000000, 39.4000000000000, 39.5000000000000, 39.6000000000000, 39.7000000000000, 39.8000000000000, 39.9000000000000, 40, 40.1000000000000, 40.2000000000000, 40.3000000000000, 40.4000000000000, 40.5000000000000, 40.6000000000000, 40.7000000000000, 40.8000000000000, 40.9000000000000, 41, 41.1000000000000, 41.2000000000000, 41.3000000000000, 41.4000000000000, 41.5000000000000, 41.6000000000000, 41.7000000000000, 41.8000000000000, 41.9000000000000, 42, 42.1000000000000, 42.2000000000000, 42.3000000000000, 42.4000000000000, 42.5000000000000, 42.6000000000000, 42.7000000000000, 42.8000000000000, 42.9000000000000, 43, 43.1000000000000, 43.2000000000000, 43.3000000000000, 43.4000000000000, 43.5000000000000, 43.6000000000000, 43.7000000000000, 43.8000000000000, 43.9000000000000, 44, 44.1000000000000, 44.2000000000000, 44.3000000000000, 44.4000000000000, 44.5000000000000, 44.6000000000000, 44.7000000000000, 44.8000000000000, 44.9000000000000, 45, 45.1000000000000, 45.2000000000000, 45.3000000000000, 45.4000000000000, 45.5000000000000, 45.6000000000000, 45.7000000000000, 45.8000000000000, 45.9000000000000, 46, 46.1000000000000, 46.2000000000000, 46.3000000000000, 46.4000000000000, 46.5000000000000, 46.6000000000000, 46.7000000000000, 46.8000000000000, 46.9000000000000, 47, 47.1000000000000, 47.2000000000000, 47.3000000000000, 47.4000000000000, 47.5000000000000, 47.6000000000000, 47.7000000000000, 47.8000000000000, 47.9000000000000, 48, 48.1000000000000, 48.2000000000000, 48.3000000000000, 48.4000000000000, 48.5000000000000, 48.6000000000000, 48.7000000000000, 48.8000000000000, 48.9000000000000, 49, 49.1000000000000, 49.2000000000000, 49.3000000000000, 49.4000000000000, 49.5000000000000, 49.6000000000000, 49.7000000000000, 49.8000000000000, 49.9000000000000, 50, 50.1000000000000, 50.2000000000000, 50.3000000000000, 50.4000000000000, 50.5000000000000, 50.6000000000000, 50.7000000000000, 50.8000000000000, 50.9000000000000, 51, 51.1000000000000, 51.2000000000000, 51.3000000000000, 51.4000000000000, 51.5000000000000, 51.6000000000000, 51.7000000000000, 51.8000000000000, 51.9000000000000, 52, 52.1000000000000, 52.2000000000000, 52.3000000000000, 52.4000000000000, 52.5000000000000, 52.6000000000000, 52.7000000000000, 52.8000000000000, 52.9000000000000, 53, 53.1000000000000, 53.2000000000000, 53.3000000000000, 53.4000000000000, 53.5000000000000, 53.6000000000000, 53.7000000000000, 53.8000000000000, 53.9000000000000, 54, 54.1000000000000, 54.2000000000000, 54.3000000000000, 54.4000000000000, 54.5000000000000, 54.6000000000000, 54.7000000000000, 54.8000000000000, 54.9000000000000, 55, 55.1000000000000, 55.2000000000000, 55.3000000000000, 55.4000000000000, 55.5000000000000, 55.6000000000000, 55.7000000000000, 55.8000000000000, 55.9000000000000, 56, 56.1000000000000, 56.2000000000000, 56.3000000000000, 56.4000000000000, 56.5000000000000, 56.6000000000000, 56.7000000000000, 56.8000000000000, 56.9000000000000, 57, 57.1000000000000, 57.2000000000000, 57.3000000000000, 57.4000000000000, 57.5000000000000, 57.6000000000000, 57.7000000000000, 57.8000000000000, 57.9000000000000, 58, 58.1000000000000, 58.2000000000000, 58.3000000000000, 58.4000000000000, 58.5000000000000, 58.6000000000000, 58.7000000000000, 58.8000000000000, 58.9000000000000, 59, 59.1000000000000, 59.2000000000000, 59.3000000000000, 59.4000000000000, 59.5000000000000, 59.6000000000000, 59.7000000000000, 59.8000000000000, 59.9000000000000, 60, 60.1000000000000, 60.2000000000000, 60.3000000000000, 60.4000000000000, 60.5000000000000, 60.6000000000000, 60.7000000000000, 60.8000000000000, 60.9000000000000, 61, 61.1000000000000, 61.2000000000000, 61.3000000000000, 61.4000000000000, 61.5000000000000, 61.6000000000000, 61.7000000000000, 61.8000000000000, 61.9000000000000, 62, 62.1000000000000, 62.2000000000000, 62.3000000000000, 62.4000000000000, 62.5000000000000, 62.6000000000000, 62.7000000000000, 62.8000000000000, 62.9000000000000, 63, 63.1000000000000, 63.2000000000000, 63.3000000000000, 63.4000000000000, 63.5000000000000, 63.6000000000000, 63.7000000000000, 63.8000000000000, 63.9000000000000, 64, 64.1000000000000, 64.2000000000000, 64.3000000000000, 64.4000000000000, 64.5000000000000, 64.6000000000000, 64.7000000000000, 64.8000000000000, 64.9000000000000, 65, 65.1000000000000, 65.2000000000000, 65.3000000000000, 65.4000000000000, 65.5000000000000, 65.6000000000000, 65.7000000000000, 65.8000000000000, 65.9000000000000, 66, 66.1000000000000, 66.2000000000000, 66.3000000000000, 66.4000000000000, 66.5000000000000, 66.6000000000000, 66.7000000000000, 66.8000000000000, 66.9000000000000, 67, 67.1000000000000, 67.2000000000000, 67.3000000000000, 67.4000000000000, 67.5000000000000, 67.6000000000000, 67.7000000000000, 67.8000000000000, 67.9000000000000, 68, 68.1000000000000, 68.2000000000000, 68.3000000000000, 68.4000000000000, 68.5000000000000, 68.6000000000000, 68.7000000000000, 68.8000000000000, 68.9000000000000, 69, 69.1000000000000, 69.2000000000000, 69.3000000000000, 69.4000000000000, 69.5000000000000, 69.6000000000000, 69.7000000000000, 69.8000000000000, 69.9000000000000, 70, 70.1000000000000, 70.2000000000000, 70.3000000000000, 70.4000000000000, 70.5000000000000, 70.6000000000000, 70.7000000000000, 70.8000000000000, 70.9000000000000, 71, 71.1000000000000, 71.2000000000000, 71.3000000000000, 71.4000000000000, 71.5000000000000, 71.6000000000000, 71.7000000000000, 71.8000000000000, 71.9000000000000, 72, 72.1000000000000, 72.2000000000000, 72.3000000000000, 72.4000000000000, 72.5000000000000, 72.6000000000000, 72.7000000000000, 72.8000000000000, 72.9000000000000, 73, 73.1000000000000, 73.2000000000000, 73.3000000000000, 73.4000000000000, 73.5000000000000, 73.6000000000000, 73.7000000000000, 73.8000000000000, 73.9000000000000, 74, 74.1000000000000, 74.2000000000000, 74.3000000000000, 74.4000000000000, 74.5000000000000, 74.6000000000000, 74.7000000000000, 74.8000000000000, 74.9000000000000, 75, 75.1000000000000, 75.2000000000000, 75.3000000000000, 75.4000000000000, 75.5000000000000, 75.6000000000000, 75.7000000000000, 75.8000000000000, 75.9000000000000, 76, 76.1000000000000, 76.2000000000000, 76.3000000000000, 76.4000000000000, 76.5000000000000, 76.6000000000000, 76.7000000000000, 76.8000000000000, 76.9000000000000, 77, 77.1000000000000, 77.2000000000000, 77.3000000000000, 77.4000000000000, 77.5000000000000, 77.6000000000000, 77.7000000000000, 77.8000000000000, 77.9000000000000, 78, 78.1000000000000, 78.2000000000000, 78.3000000000000, 78.4000000000000, 78.5000000000000, 78.6000000000000, 78.7000000000000, 78.8000000000000, 78.9000000000000, 79, 79.1000000000000, 79.2000000000000, 79.3000000000000, 79.4000000000000, 79.5000000000000, 79.6000000000000, 79.7000000000000, 79.8000000000000, 79.9000000000000, 80, 80.1000000000000, 80.2000000000000, 80.3000000000000, 80.4000000000000, 80.5000000000000, 80.6000000000000, 80.7000000000000, 80.8000000000000, 80.9000000000000, 81, 81.1000000000000, 81.2000000000000, 81.3000000000000, 81.4000000000000, 81.5000000000000, 81.6000000000000, 81.7000000000000, 81.8000000000000, 81.9000000000000, 82, 82.1000000000000, 82.2000000000000, 82.3000000000000, 82.4000000000000, 82.5000000000000, 82.6000000000000, 82.7000000000000, 82.8000000000000, 82.9000000000000, 83, 83.1000000000000, 83.2000000000000, 83.3000000000000, 83.4000000000000, 83.5000000000000, 83.6000000000000, 83.7000000000000, 83.8000000000000, 83.9000000000000, 84, 84.1000000000000, 84.2000000000000, 84.3000000000000, 84.4000000000000, 84.5000000000000, 84.6000000000000, 84.7000000000000, 84.8000000000000, 84.9000000000000, 85, 85.1000000000000, 85.2000000000000, 85.3000000000000, 85.4000000000000, 85.5000000000000, 85.6000000000000, 85.7000000000000, 85.8000000000000, 85.9000000000000, 86, 86.1000000000000, 86.2000000000000, 86.3000000000000, 86.4000000000000, 86.5000000000000, 86.6000000000000, 86.7000000000000, 86.8000000000000, 86.9000000000000, 87, 87.1000000000000, 87.2000000000000, 87.3000000000000, 87.4000000000000, 87.5000000000000, 87.6000000000000, 87.7000000000000, 87.8000000000000, 87.9000000000000, 88, 88.1000000000000, 88.2000000000000, 88.3000000000000, 88.4000000000000, 88.5000000000000, 88.6000000000000, 88.7000000000000, 88.8000000000000, 88.9000000000000, 89, 89.1000000000000, 89.2000000000000, 89.3000000000000, 89.4000000000000, 89.5000000000000, 89.6000000000000, 89.7000000000000, 89.8000000000000, 89.9000000000000, 90, 90.1000000000000, 90.2000000000000, 90.3000000000000, 90.4000000000000, 90.5000000000000, 90.6000000000000, 90.7000000000000, 90.8000000000000, 90.9000000000000, 91, 91.1000000000000, 91.2000000000000, 91.3000000000000, 91.4000000000000, 91.5000000000000, 91.6000000000000, 91.7000000000000, 91.8000000000000, 91.9000000000000, 92, 92.1000000000000, 92.2000000000000, 92.3000000000000, 92.4000000000000, 92.5000000000000, 92.6000000000000, 92.7000000000000, 92.8000000000000, 92.9000000000000, 93, 93.1000000000000, 93.2000000000000, 93.3000000000000, 93.4000000000000, 93.5000000000000, 93.6000000000000, 93.7000000000000, 93.8000000000000, 93.9000000000000, 94, 94.1000000000000, 94.2000000000000, 94.3000000000000, 94.4000000000000, 94.5000000000000, 94.6000000000000, 94.7000000000000, 94.8000000000000, 94.9000000000000, 95, 95.1000000000000, 95.2000000000000, 95.3000000000000, 95.4000000000000, 95.5000000000000, 95.6000000000000, 95.7000000000000, 95.8000000000000, 95.9000000000000, 96, 96.1000000000000, 96.2000000000000, 96.3000000000000, 96.4000000000000, 96.5000000000000, 96.6000000000000, 96.7000000000000, 96.8000000000000, 96.9000000000000, 97, 97.1000000000000, 97.2000000000000, 97.3000000000000, 97.4000000000000, 97.5000000000000, 97.6000000000000, 97.7000000000000, 97.8000000000000, 97.9000000000000, 98, 98.1000000000000, 98.2000000000000, 98.3000000000000, 98.4000000000000, 98.5000000000000, 98.6000000000000, 98.7000000000000, 98.8000000000000, 98.9000000000000, 99, 99.1000000000000, 99.2000000000000, 99.3000000000000, 99.4000000000000, 99.5000000000000, 99.6000000000000, 99.7000000000000, 99.8000000000000, 99.9000000000000, 100, 100.100000000000, 100.200000000000, 100.300000000000, 100.400000000000, 100.500000000000, 100.600000000000, 100.700000000000, 100.800000000000, 100.900000000000, 101, 101.100000000000, 101.200000000000, 101.300000000000, 101.400000000000, 101.500000000000, 101.600000000000, 101.700000000000, 101.800000000000, 101.900000000000, 102, 102.100000000000, 102.200000000000, 102.300000000000, 102.400000000000, 102.500000000000, 102.600000000000, 102.700000000000, 102.800000000000, 102.900000000000, 103, 103.100000000000, 103.200000000000, 103.300000000000, 103.400000000000, 103.500000000000, 103.600000000000, 103.700000000000, 103.800000000000, 103.900000000000, 104, 104.100000000000, 104.200000000000, 104.300000000000, 104.400000000000, 104.500000000000, 104.600000000000, 104.700000000000, 104.800000000000, 104.900000000000, 105, 105.100000000000, 105.200000000000, 105.300000000000, 105.400000000000, 105.500000000000, 105.600000000000, 105.700000000000, 105.800000000000, 105.900000000000, 106, 106.100000000000, 106.200000000000, 106.300000000000, 106.400000000000, 106.500000000000, 106.600000000000, 106.700000000000, 106.800000000000, 106.900000000000, 107, 107.100000000000, 107.200000000000, 107.300000000000, 107.400000000000, 107.500000000000, 107.600000000000, 107.700000000000, 107.800000000000, 107.900000000000, 108, 108.100000000000, 108.200000000000, 108.300000000000, 108.400000000000, 108.500000000000, 108.600000000000, 108.700000000000, 108.800000000000, 108.900000000000, 109, 109.100000000000, 109.200000000000, 109.300000000000, 109.400000000000, 109.500000000000, 109.600000000000, 109.700000000000, 109.800000000000, 109.900000000000, 110, 110.100000000000, 110.200000000000, 110.300000000000, 110.400000000000, 110.500000000000, 110.600000000000, 110.700000000000, 110.800000000000, 110.900000000000, 111, 111.100000000000, 111.200000000000, 111.300000000000, 111.400000000000, 111.500000000000, 111.600000000000, 111.700000000000, 111.800000000000, 111.900000000000, 112, 112.100000000000, 112.200000000000, 112.300000000000, 112.400000000000, 112.500000000000, 112.600000000000, 112.700000000000, 112.800000000000, 112.900000000000, 113, 113.100000000000, 113.200000000000, 113.300000000000, 113.400000000000, 113.500000000000, 113.600000000000, 113.700000000000, 113.800000000000, 113.900000000000, 114, 114.100000000000, 114.200000000000, 114.300000000000, 114.400000000000, 114.500000000000, 114.600000000000, 114.700000000000, 114.800000000000, 114.900000000000, 115, 115.100000000000, 115.200000000000, 115.300000000000, 115.400000000000, 115.500000000000, 115.600000000000, 115.700000000000, 115.800000000000, 115.900000000000, 116, 116.100000000000, 116.200000000000, 116.300000000000, 116.400000000000, 116.500000000000, 116.600000000000, 116.700000000000, 116.800000000000, 116.900000000000, 117, 117.100000000000, 117.200000000000, 117.300000000000, 117.400000000000, 117.500000000000, 117.600000000000, 117.700000000000, 117.800000000000, 117.900000000000, 118, 118.100000000000, 118.200000000000, 118.300000000000, 118.400000000000, 118.500000000000, 118.600000000000, 118.700000000000, 118.800000000000, 118.900000000000, 119, 119.100000000000, 119.200000000000, 119.300000000000, 119.400000000000, 119.500000000000, 119.600000000000, 119.700000000000, 119.800000000000, 119.900000000000, 120, 120.100000000000, 120.200000000000, 120.300000000000, 120.400000000000, 120.500000000000, 120.600000000000, 120.700000000000, 120.800000000000, 120.900000000000, 121, 121.100000000000, 121.200000000000, 121.300000000000, 121.400000000000, 121.500000000000, 121.600000000000, 121.700000000000, 121.800000000000, 121.900000000000, 122, 122.100000000000, 122.200000000000, 122.300000000000, 122.400000000000, 122.500000000000, 122.600000000000, 122.700000000000, 122.800000000000, 122.900000000000, 123, 123.100000000000, 123.200000000000, 123.300000000000, 123.400000000000, 123.500000000000, 123.600000000000, 123.700000000000, 123.800000000000, 123.900000000000, 124, 124.100000000000, 124.200000000000, 124.300000000000, 124.400000000000, 124.500000000000, 124.600000000000, 124.700000000000, 124.800000000000, 124.900000000000, 125, 125.100000000000, 125.200000000000, 125.300000000000, 125.400000000000, 125.500000000000, 125.600000000000, 125.700000000000, 125.800000000000, 125.900000000000, 126, 126.100000000000, 126.200000000000, 126.300000000000, 126.400000000000, 126.500000000000, 126.600000000000, 126.700000000000, 126.800000000000, 126.900000000000, 127, 127.100000000000, 127.200000000000, 127.300000000000, 127.400000000000, 127.500000000000, 127.600000000000, 127.700000000000, 127.800000000000, 127.900000000000, 128, 128.100000000000, 128.200000000000, 128.300000000000, 128.400000000000, 128.500000000000, 128.600000000000, 128.700000000000, 128.800000000000, 128.900000000000, 129, 129.100000000000, 129.200000000000, 129.300000000000, 129.400000000000, 129.500000000000, 129.600000000000, 129.700000000000, 129.800000000000, 129.900000000000, 130, 130.100000000000, 130.200000000000, 130.300000000000, 130.400000000000, 130.500000000000, 130.600000000000, 130.700000000000, 130.800000000000, 130.900000000000, 131, 131.100000000000, 131.200000000000, 131.300000000000, 131.400000000000, 131.500000000000, 131.600000000000, 131.700000000000, 131.800000000000, 131.900000000000, 132, 132.100000000000, 132.200000000000, 132.300000000000, 132.400000000000, 132.500000000000, 132.600000000000, 132.700000000000, 132.800000000000, 132.900000000000, 133, 133.100000000000, 133.200000000000, 133.300000000000, 133.400000000000, 133.500000000000, 133.600000000000, 133.700000000000, 133.800000000000, 133.900000000000, 134, 134.100000000000, 134.200000000000, 134.300000000000, 134.400000000000, 134.500000000000, 134.600000000000, 134.700000000000, 134.800000000000, 134.900000000000, 135, 135.100000000000, 135.200000000000, 135.300000000000, 135.400000000000, 135.500000000000, 135.600000000000, 135.700000000000, 135.800000000000, 135.900000000000, 136, 136.100000000000, 136.200000000000, 136.300000000000, 136.400000000000, 136.500000000000, 136.600000000000, 136.700000000000, 136.800000000000, 136.900000000000, 137, 137.100000000000, 137.200000000000, 137.300000000000, 137.400000000000, 137.500000000000, 137.600000000000, 137.700000000000, 137.800000000000, 137.900000000000, 138, 138.100000000000, 138.200000000000, 138.300000000000, 138.400000000000, 138.500000000000, 138.600000000000, 138.700000000000, 138.800000000000, 138.900000000000, 139, 139.100000000000, 139.200000000000, 139.300000000000, 139.400000000000, 139.500000000000, 139.600000000000, 139.700000000000, 139.800000000000, 139.900000000000, 140, 140.100000000000, 140.200000000000, 140.300000000000, 140.400000000000, 140.500000000000, 140.600000000000, 140.700000000000, 140.800000000000, 140.900000000000, 141, 141.100000000000, 141.200000000000, 141.300000000000, 141.400000000000, 141.500000000000, 141.600000000000, 141.700000000000, 141.800000000000, 141.900000000000, 142, 142.100000000000, 142.200000000000, 142.300000000000, 142.400000000000, 142.500000000000, 142.600000000000, 142.700000000000, 142.800000000000, 142.900000000000, 143, 143.100000000000, 143.200000000000, 143.300000000000, 143.400000000000, 143.500000000000, 143.600000000000, 143.700000000000, 143.800000000000, 143.900000000000, 144, 144.100000000000, 144.200000000000, 144.300000000000, 144.400000000000, 144.500000000000, 144.600000000000, 144.700000000000, 144.800000000000, 144.900000000000, 145, 145.100000000000, 145.200000000000, 145.300000000000, 145.400000000000, 145.500000000000, 145.600000000000, 145.700000000000, 145.800000000000, 145.900000000000, 146, 146.100000000000, 146.200000000000, 146.300000000000, 146.400000000000, 146.500000000000, 146.600000000000, 146.700000000000, 146.800000000000, 146.900000000000, 147, 147.100000000000, 147.200000000000, 147.300000000000, 147.400000000000, 147.500000000000, 147.600000000000, 147.700000000000, 147.800000000000, 147.900000000000, 148, 148.100000000000, 148.200000000000, 148.300000000000, 148.400000000000, 148.500000000000, 148.600000000000, 148.700000000000, 148.800000000000, 148.900000000000, 149, 149.100000000000, 149.200000000000, 149.300000000000, 149.400000000000, 149.500000000000, 149.600000000000, 149.700000000000, 149.800000000000, 149.900000000000, 150, 150.100000000000, 150.200000000000, 150.300000000000, 150.400000000000, 150.500000000000, 150.600000000000, 150.700000000000, 150.800000000000, 150.900000000000, 151, 151.100000000000, 151.200000000000, 151.300000000000, 151.400000000000, 151.500000000000, 151.600000000000, 151.700000000000, 151.800000000000, 151.900000000000, 152, 152.100000000000, 152.200000000000, 152.300000000000, 152.400000000000, 152.500000000000, 152.600000000000, 152.700000000000, 152.800000000000, 152.900000000000, 153, 153.100000000000, 153.200000000000, 153.300000000000, 153.400000000000, 153.500000000000, 153.600000000000, 153.700000000000, 153.800000000000, 153.900000000000, 154, 154.100000000000, 154.200000000000, 154.300000000000, 154.400000000000, 154.500000000000, 154.600000000000, 154.700000000000, 154.800000000000, 154.900000000000, 155, 155.100000000000, 155.200000000000, 155.300000000000, 155.400000000000, 155.500000000000, 155.600000000000, 155.700000000000, 155.800000000000, 155.900000000000, 156, 156.100000000000, 156.200000000000, 156.300000000000, 156.400000000000, 156.500000000000, 156.600000000000, 156.700000000000, 156.800000000000, 156.900000000000, 157, 157.100000000000, 157.200000000000, 157.300000000000, 157.400000000000, 157.500000000000, 157.600000000000, 157.700000000000, 157.800000000000, 157.900000000000, 158, 158.100000000000, 158.200000000000, 158.300000000000, 158.400000000000, 158.500000000000, 158.600000000000, 158.700000000000, 158.800000000000, 158.900000000000, 159, 159.100000000000, 159.200000000000, 159.300000000000, 159.400000000000, 159.500000000000, 159.600000000000, 159.700000000000, 159.800000000000, 159.900000000000, 160, 160.100000000000, 160.200000000000, 160.300000000000, 160.400000000000, 160.500000000000, 160.600000000000, 160.700000000000, 160.800000000000, 160.900000000000, 161, 161.100000000000, 161.200000000000, 161.300000000000, 161.400000000000, 161.500000000000, 161.600000000000, 161.700000000000, 161.800000000000, 161.900000000000, 162, 162.100000000000, 162.200000000000, 162.300000000000, 162.400000000000, 162.500000000000, 162.600000000000, 162.700000000000, 162.800000000000, 162.900000000000, 163, 163.100000000000, 163.200000000000, 163.300000000000, 163.400000000000, 163.500000000000, 163.600000000000, 163.700000000000, 163.800000000000, 163.900000000000, 164, 164.100000000000, 164.200000000000, 164.300000000000, 164.400000000000, 164.500000000000, 164.600000000000, 164.700000000000, 164.800000000000, 164.900000000000, 165, 165.100000000000, 165.200000000000, 165.300000000000, 165.400000000000, 165.500000000000, 165.600000000000, 165.700000000000, 165.800000000000, 165.900000000000, 166, 166.100000000000, 166.200000000000, 166.300000000000, 166.400000000000, 166.500000000000, 166.600000000000, 166.700000000000, 166.800000000000, 166.900000000000, 167, 167.100000000000, 167.200000000000, 167.300000000000, 167.400000000000, 167.500000000000, 167.600000000000, 167.700000000000, 167.800000000000, 167.900000000000, 168, 168.100000000000, 168.200000000000, 168.300000000000, 168.400000000000, 168.500000000000, 168.600000000000, 168.700000000000, 168.800000000000, 168.900000000000, 169, 169.100000000000, 169.200000000000, 169.300000000000, 169.400000000000, 169.500000000000, 169.600000000000, 169.700000000000, 169.800000000000, 169.900000000000, 170, 170.100000000000, 170.200000000000, 170.300000000000, 170.400000000000, 170.500000000000, 170.600000000000, 170.700000000000, 170.800000000000, 170.900000000000, 171, 171.100000000000, 171.200000000000, 171.300000000000, 171.400000000000,
                    171.500000000000, 171.600000000000, 171.700000000000, 171.800000000000, 171.900000000000, 172, 172.100000000000, 172.200000000000, 172.300000000000, 172.400000000000, 172.500000000000, 172.600000000000, 172.700000000000, 172.800000000000, 172.900000000000, 173, 173.100000000000, 173.200000000000, 173.300000000000, 173.400000000000, 173.500000000000, 173.600000000000, 173.700000000000, 173.800000000000, 173.900000000000, 174, 174.100000000000, 174.200000000000, 174.300000000000, 174.400000000000, 174.500000000000, 174.600000000000, 174.700000000000, 174.800000000000, 174.900000000000, 175, 175.100000000000, 175.200000000000, 175.300000000000, 175.400000000000, 175.500000000000, 175.600000000000, 175.700000000000, 175.800000000000, 175.900000000000, 176, 176.100000000000, 176.200000000000, 176.300000000000, 176.400000000000, 176.500000000000, 176.600000000000, 176.700000000000, 176.800000000000, 176.900000000000, 177, 177.100000000000, 177.200000000000, 177.300000000000, 177.400000000000, 177.500000000000, 177.600000000000, 177.700000000000, 177.800000000000, 177.900000000000, 178, 178.100000000000, 178.200000000000, 178.300000000000, 178.400000000000, 178.500000000000, 178.600000000000, 178.700000000000, 178.800000000000, 178.900000000000, 179, 179.100000000000, 179.200000000000, 179.300000000000, 179.400000000000, 179.500000000000, 179.600000000000, 179.700000000000, 179.800000000000, 179.900000000000, 180, 180.100000000000, 180.200000000000, 180.300000000000, 180.400000000000, 180.500000000000, 180.600000000000, 180.700000000000, 180.800000000000, 180.900000000000, 181, 181.100000000000, 181.200000000000, 181.300000000000, 181.400000000000, 181.500000000000, 181.600000000000, 181.700000000000, 181.800000000000, 181.900000000000, 182, 182.100000000000, 182.200000000000, 182.300000000000, 182.400000000000, 182.500000000000, 182.600000000000, 182.700000000000, 182.800000000000, 182.900000000000, 183, 183.100000000000, 183.200000000000, 183.300000000000, 183.400000000000, 183.500000000000, 183.600000000000, 183.700000000000, 183.800000000000, 183.900000000000, 184, 184.100000000000, 184.200000000000, 184.300000000000, 184.400000000000, 184.500000000000, 184.600000000000, 184.700000000000, 184.800000000000, 184.900000000000, 185, 185.100000000000, 185.200000000000, 185.300000000000, 185.400000000000, 185.500000000000, 185.600000000000, 185.700000000000, 185.800000000000, 185.900000000000, 186, 186.100000000000, 186.200000000000, 186.300000000000, 186.400000000000, 186.500000000000, 186.600000000000, 186.700000000000, 186.800000000000, 186.900000000000, 187, 187.100000000000, 187.200000000000, 187.300000000000, 187.400000000000, 187.500000000000, 187.600000000000, 187.700000000000, 187.800000000000, 187.900000000000, 188, 188.100000000000, 188.200000000000, 188.300000000000, 188.400000000000, 188.500000000000, 188.600000000000, 188.700000000000, 188.800000000000, 188.900000000000, 189, 189.100000000000, 189.200000000000, 189.300000000000, 189.400000000000, 189.500000000000, 189.600000000000, 189.700000000000, 189.800000000000, 189.900000000000, 190, 190.100000000000, 190.200000000000, 190.300000000000, 190.400000000000, 190.500000000000, 190.600000000000, 190.700000000000, 190.800000000000, 190.900000000000, 191, 191.100000000000, 191.200000000000, 191.300000000000, 191.400000000000, 191.500000000000, 191.600000000000, 191.700000000000, 191.800000000000, 191.900000000000, 192, 192.100000000000, 192.200000000000, 192.300000000000, 192.400000000000, 192.500000000000, 192.600000000000, 192.700000000000, 192.800000000000, 192.900000000000, 193, 193.100000000000, 193.200000000000, 193.300000000000, 193.400000000000, 193.500000000000, 193.600000000000, 193.700000000000, 193.800000000000, 193.900000000000, 194, 194.100000000000, 194.200000000000, 194.300000000000, 194.400000000000, 194.500000000000, 194.600000000000, 194.700000000000, 194.800000000000, 194.900000000000, 195, 195.100000000000, 195.200000000000, 195.300000000000, 195.400000000000, 195.500000000000, 195.600000000000, 195.700000000000, 195.800000000000, 195.900000000000, 196, 196.100000000000, 196.200000000000, 196.300000000000, 196.400000000000, 196.500000000000, 196.600000000000, 196.700000000000, 196.800000000000, 196.900000000000, 197, 197.100000000000, 197.200000000000, 197.300000000000, 197.400000000000, 197.500000000000, 197.600000000000, 197.700000000000, 197.800000000000, 197.900000000000, 198, 198.100000000000, 198.200000000000, 198.300000000000, 198.400000000000, 198.500000000000, 198.600000000000, 198.700000000000, 198.800000000000, 198.900000000000, 199, 199.100000000000, 199.200000000000, 199.300000000000, 199.400000000000, 199.500000000000, 199.600000000000, 199.700000000000, 199.800000000000, 199.900000000000, 200, 200.100000000000, 200.200000000000, 200.300000000000, 200.400000000000, 200.500000000000, 200.600000000000, 200.700000000000, 200.800000000000, 200.900000000000, 201, 201.100000000000, 201.200000000000, 201.300000000000, 201.400000000000, 201.500000000000, 201.600000000000, 201.700000000000, 201.800000000000, 201.900000000000, 202, 202.100000000000, 202.200000000000, 202.300000000000, 202.400000000000, 202.500000000000, 202.600000000000, 202.700000000000, 202.800000000000, 202.900000000000, 203, 203.100000000000, 203.200000000000, 203.300000000000, 203.400000000000, 203.500000000000, 203.600000000000, 203.700000000000, 203.800000000000, 203.900000000000, 204, 204.100000000000, 204.200000000000, 204.300000000000, 204.400000000000, 204.500000000000, 204.600000000000, 204.700000000000, 204.800000000000, 204.900000000000, 205, 205.100000000000, 205.200000000000, 205.300000000000, 205.400000000000, 205.500000000000, 205.600000000000, 205.700000000000, 205.800000000000, 205.900000000000, 206, 206.100000000000, 206.200000000000, 206.300000000000, 206.400000000000, 206.500000000000, 206.600000000000, 206.700000000000, 206.800000000000, 206.900000000000, 207, 207.100000000000, 207.200000000000, 207.300000000000, 207.400000000000, 207.500000000000, 207.600000000000, 207.700000000000, 207.800000000000, 207.900000000000, 208, 208.100000000000, 208.200000000000, 208.300000000000, 208.400000000000, 208.500000000000, 208.600000000000, 208.700000000000, 208.800000000000, 208.900000000000, 209, 209.100000000000, 209.200000000000, 209.300000000000, 209.400000000000, 209.500000000000, 209.600000000000, 209.700000000000, 209.800000000000, 209.900000000000, 210, 210.100000000000, 210.200000000000, 210.300000000000, 210.400000000000, 210.500000000000, 210.600000000000, 210.700000000000, 210.800000000000, 210.900000000000, 211, 211.100000000000, 211.200000000000, 211.300000000000, 211.400000000000, 211.500000000000, 211.600000000000, 211.700000000000, 211.800000000000, 211.900000000000, 212, 212.100000000000, 212.200000000000, 212.300000000000, 212.400000000000, 212.500000000000, 212.600000000000, 212.700000000000, 212.800000000000, 212.900000000000, 213, 213.100000000000, 213.200000000000, 213.300000000000, 213.400000000000, 213.500000000000, 213.600000000000, 213.700000000000, 213.800000000000, 213.900000000000, 214, 214.100000000000, 214.200000000000, 214.300000000000, 214.400000000000, 214.500000000000, 214.600000000000, 214.700000000000, 214.800000000000, 214.900000000000, 215, 215.100000000000, 215.200000000000, 215.300000000000, 215.400000000000, 215.500000000000, 215.600000000000, 215.700000000000, 215.800000000000, 215.900000000000, 216, 216.100000000000, 216.200000000000, 216.300000000000, 216.400000000000, 216.500000000000, 216.600000000000, 216.700000000000, 216.800000000000, 216.900000000000, 217, 217.100000000000, 217.200000000000, 217.300000000000, 217.400000000000, 217.500000000000, 217.600000000000, 217.700000000000, 217.800000000000, 217.900000000000, 218, 218.100000000000, 218.200000000000, 218.300000000000, 218.400000000000, 218.500000000000, 218.600000000000, 218.700000000000, 218.800000000000, 218.900000000000, 219, 219.100000000000, 219.200000000000, 219.300000000000, 219.400000000000, 219.500000000000, 219.600000000000, 219.700000000000, 219.800000000000, 219.900000000000, 220, 220.100000000000, 220.200000000000, 220.300000000000, 220.400000000000, 220.500000000000, 220.600000000000, 220.700000000000, 220.800000000000, 220.900000000000, 221, 221.100000000000, 221.200000000000, 221.300000000000, 221.400000000000, 221.500000000000, 221.600000000000, 221.700000000000, 221.800000000000, 221.900000000000, 222, 222.100000000000, 222.200000000000, 222.300000000000, 222.400000000000, 222.500000000000, 222.600000000000, 222.700000000000, 222.800000000000, 222.900000000000, 223, 223.100000000000, 223.200000000000, 223.300000000000, 223.400000000000, 223.500000000000, 223.600000000000, 223.700000000000, 223.800000000000, 223.900000000000, 224, 224.100000000000, 224.200000000000, 224.300000000000, 224.400000000000, 224.500000000000, 224.600000000000, 224.700000000000, 224.800000000000, 224.900000000000, 225, 225.100000000000, 225.200000000000, 225.300000000000, 225.400000000000, 225.500000000000, 225.600000000000, 225.700000000000, 225.800000000000, 225.900000000000, 226, 226.100000000000, 226.200000000000, 226.300000000000, 226.400000000000, 226.500000000000, 226.600000000000, 226.700000000000, 226.800000000000, 226.900000000000, 227, 227.100000000000, 227.200000000000, 227.300000000000, 227.400000000000, 227.500000000000, 227.600000000000, 227.700000000000, 227.800000000000, 227.900000000000, 228, 228.100000000000, 228.200000000000, 228.300000000000, 228.400000000000, 228.500000000000, 228.600000000000, 228.700000000000, 228.800000000000, 228.900000000000, 229, 229.100000000000, 229.200000000000, 229.300000000000, 229.400000000000, 229.500000000000, 229.600000000000, 229.700000000000, 229.800000000000, 229.900000000000, 230, 230.100000000000, 230.200000000000, 230.300000000000, 230.400000000000, 230.500000000000, 230.600000000000, 230.700000000000, 230.800000000000, 230.900000000000, 231, 231.100000000000, 231.200000000000, 231.300000000000, 231.400000000000, 231.500000000000, 231.600000000000, 231.700000000000, 231.800000000000, 231.900000000000, 232, 232.100000000000, 232.200000000000, 232.300000000000, 232.400000000000, 232.500000000000, 232.600000000000, 232.700000000000, 232.800000000000, 232.900000000000, 233, 233.100000000000, 233.200000000000, 233.300000000000, 233.400000000000, 233.500000000000, 233.600000000000, 233.700000000000, 233.800000000000, 233.900000000000, 234, 234.100000000000, 234.200000000000, 234.300000000000, 234.400000000000, 234.500000000000, 234.600000000000, 234.700000000000, 234.800000000000, 234.900000000000, 235, 235.100000000000, 235.200000000000, 235.300000000000, 235.400000000000, 235.500000000000, 235.600000000000, 235.700000000000, 235.800000000000, 235.900000000000, 236, 236.100000000000, 236.200000000000, 236.300000000000, 236.400000000000, 236.500000000000, 236.600000000000, 236.700000000000, 236.800000000000, 236.900000000000, 237, 237.100000000000, 237.200000000000, 237.300000000000, 237.400000000000, 237.500000000000, 237.600000000000, 237.700000000000, 237.800000000000, 237.900000000000, 238, 238.100000000000, 238.200000000000, 238.300000000000, 238.400000000000, 238.500000000000, 238.600000000000, 238.700000000000, 238.800000000000, 238.900000000000, 239, 239.100000000000, 239.200000000000, 239.300000000000, 239.400000000000, 239.500000000000, 239.600000000000, 239.700000000000, 239.800000000000, 239.900000000000, 240, 240.100000000000, 240.200000000000, 240.300000000000, 240.400000000000, 240.500000000000, 240.600000000000, 240.700000000000, 240.800000000000, 240.900000000000, 241, 241.100000000000, 241.200000000000, 241.300000000000, 241.400000000000, 241.500000000000, 241.600000000000, 241.700000000000, 241.800000000000, 241.900000000000, 242, 242.100000000000, 242.200000000000, 242.300000000000, 242.400000000000, 242.500000000000, 242.600000000000, 242.700000000000, 242.800000000000, 242.900000000000, 243, 243.100000000000, 243.200000000000, 243.300000000000, 243.400000000000, 243.500000000000, 243.600000000000, 243.700000000000, 243.800000000000, 243.900000000000, 244, 244.100000000000, 244.200000000000, 244.300000000000, 244.400000000000, 244.500000000000, 244.600000000000, 244.700000000000, 244.800000000000, 244.900000000000, 245, 245.100000000000, 245.200000000000, 245.300000000000, 245.400000000000, 245.500000000000, 245.600000000000, 245.700000000000, 245.800000000000, 245.900000000000, 246, 246.100000000000, 246.200000000000, 246.300000000000, 246.400000000000, 246.500000000000, 246.600000000000, 246.700000000000, 246.800000000000, 246.900000000000, 247, 247.100000000000, 247.200000000000, 247.300000000000, 247.400000000000, 247.500000000000, 247.600000000000, 247.700000000000, 247.800000000000, 247.900000000000, 248, 248.100000000000, 248.200000000000, 248.300000000000, 248.400000000000, 248.500000000000, 248.600000000000, 248.700000000000, 248.800000000000, 248.900000000000, 249, 249.100000000000, 249.200000000000, 249.300000000000, 249.400000000000, 249.500000000000, 249.600000000000, 249.700000000000, 249.800000000000, 249.900000000000, 250, 250.100000000000, 250.200000000000, 250.300000000000, 250.400000000000, 250.500000000000, 250.600000000000, 250.700000000000, 250.800000000000, 250.900000000000, 251, 251.100000000000, 251.200000000000, 251.300000000000, 251.400000000000, 251.500000000000, 251.600000000000, 251.700000000000, 251.800000000000, 251.900000000000, 252, 252.100000000000, 252.200000000000, 252.300000000000, 252.400000000000, 252.500000000000, 252.600000000000, 252.700000000000, 252.800000000000, 252.900000000000, 253, 253.100000000000, 253.200000000000, 253.300000000000, 253.400000000000, 253.500000000000, 253.600000000000, 253.700000000000, 253.800000000000, 253.900000000000, 254, 254.100000000000, 254.200000000000, 254.300000000000, 254.400000000000, 254.500000000000, 254.600000000000, 254.700000000000, 254.800000000000, 254.900000000000, 255, 255.100000000000, 255.200000000000, 255.300000000000, 255.400000000000, 255.500000000000, 255.600000000000, 255.700000000000, 255.800000000000, 255.900000000000, 256, 256.100000000000, 256.200000000000, 256.300000000000, 256.400000000000, 256.500000000000, 256.600000000000, 256.700000000000, 256.800000000000, 256.900000000000, 257, 257.100000000000, 257.200000000000, 257.300000000000, 257.400000000000, 257.500000000000, 257.600000000000, 257.700000000000, 257.800000000000, 257.900000000000, 258, 258.100000000000, 258.200000000000, 258.300000000000, 258.400000000000, 258.500000000000, 258.600000000000, 258.700000000000, 258.800000000000, 258.900000000000, 259, 259.100000000000, 259.200000000000, 259.300000000000, 259.400000000000, 259.500000000000, 259.600000000000, 259.700000000000, 259.800000000000, 259.900000000000, 260, 260.100000000000, 260.200000000000, 260.300000000000, 260.400000000000, 260.500000000000, 260.600000000000, 260.700000000000, 260.800000000000, 260.900000000000, 261, 261.100000000000, 261.200000000000, 261.300000000000, 261.400000000000, 261.500000000000, 261.600000000000, 261.700000000000, 261.800000000000, 261.900000000000, 262, 262.100000000000, 262.200000000000, 262.300000000000, 262.400000000000, 262.500000000000, 262.600000000000, 262.700000000000, 262.800000000000, 262.900000000000, 263, 263.100000000000, 263.200000000000, 263.300000000000, 263.400000000000, 263.500000000000, 263.600000000000, 263.700000000000, 263.800000000000, 263.900000000000, 264, 264.100000000000, 264.200000000000, 264.300000000000, 264.400000000000, 264.500000000000, 264.600000000000, 264.700000000000, 264.800000000000, 264.900000000000, 265, 265.100000000000, 265.200000000000, 265.300000000000, 265.400000000000, 265.500000000000, 265.600000000000, 265.700000000000, 265.800000000000, 265.900000000000, 266, 266.100000000000, 266.200000000000, 266.300000000000, 266.400000000000, 266.500000000000, 266.600000000000, 266.700000000000, 266.800000000000, 266.900000000000, 267, 267.100000000000, 267.200000000000, 267.300000000000, 267.400000000000, 267.500000000000, 267.600000000000, 267.700000000000, 267.800000000000, 267.900000000000, 268, 268.100000000000, 268.200000000000, 268.300000000000, 268.400000000000, 268.500000000000, 268.600000000000, 268.700000000000, 268.800000000000, 268.900000000000, 269, 269.100000000000, 269.200000000000, 269.300000000000, 269.400000000000, 269.500000000000, 269.600000000000, 269.700000000000, 269.800000000000, 269.900000000000, 270, 270.100000000000, 270.200000000000, 270.300000000000, 270.400000000000, 270.500000000000, 270.600000000000, 270.700000000000, 270.800000000000, 270.900000000000, 271, 271.100000000000, 271.200000000000, 271.300000000000, 271.400000000000, 271.500000000000, 271.600000000000, 271.700000000000, 271.800000000000, 271.900000000000, 272, 272.100000000000, 272.200000000000, 272.300000000000, 272.400000000000, 272.500000000000, 272.600000000000, 272.700000000000, 272.800000000000, 272.900000000000, 273, 273.100000000000, 273.200000000000, 273.300000000000, 273.400000000000, 273.500000000000, 273.600000000000, 273.700000000000, 273.800000000000, 273.900000000000, 274, 274.100000000000, 274.200000000000, 274.300000000000, 274.400000000000, 274.500000000000, 274.600000000000, 274.700000000000, 274.800000000000, 274.900000000000, 275, 275.100000000000, 275.200000000000, 275.300000000000, 275.400000000000, 275.500000000000, 275.600000000000, 275.700000000000, 275.800000000000, 275.900000000000, 276, 276.100000000000, 276.200000000000, 276.300000000000, 276.400000000000, 276.500000000000, 276.600000000000, 276.700000000000, 276.800000000000, 276.900000000000, 277, 277.100000000000, 277.200000000000, 277.300000000000, 277.400000000000, 277.500000000000, 277.600000000000, 277.700000000000, 277.800000000000, 277.900000000000, 278, 278.100000000000, 278.200000000000, 278.300000000000, 278.400000000000, 278.500000000000, 278.600000000000, 278.700000000000, 278.800000000000, 278.900000000000, 279, 279.100000000000, 279.200000000000, 279.300000000000, 279.400000000000, 279.500000000000, 279.600000000000, 279.700000000000, 279.800000000000, 279.900000000000, 280, 280.100000000000, 280.200000000000, 280.300000000000, 280.400000000000, 280.500000000000, 280.600000000000, 280.700000000000, 280.800000000000, 280.900000000000, 281, 281.100000000000, 281.200000000000, 281.300000000000, 281.400000000000, 281.500000000000, 281.600000000000, 281.700000000000, 281.800000000000, 281.900000000000, 282, 282.100000000000, 282.200000000000, 282.300000000000, 282.400000000000, 282.500000000000, 282.600000000000, 282.700000000000, 282.800000000000, 282.900000000000, 283, 283.100000000000, 283.200000000000, 283.300000000000, 283.400000000000, 283.500000000000, 283.600000000000, 283.700000000000, 283.800000000000, 283.900000000000, 284, 284.100000000000, 284.200000000000, 284.300000000000, 284.400000000000, 284.500000000000, 284.600000000000, 284.700000000000, 284.800000000000, 284.900000000000, 285, 285.100000000000, 285.200000000000, 285.300000000000, 285.400000000000, 285.500000000000, 285.600000000000, 285.700000000000, 285.800000000000, 285.900000000000, 286, 286.100000000000, 286.200000000000, 286.300000000000, 286.400000000000, 286.500000000000, 286.600000000000, 286.700000000000, 286.800000000000, 286.900000000000, 287, 287.100000000000, 287.200000000000, 287.300000000000, 287.400000000000, 287.500000000000, 287.600000000000, 287.700000000000, 287.800000000000, 287.900000000000, 288, 288.100000000000, 288.200000000000, 288.300000000000, 288.400000000000, 288.500000000000, 288.600000000000, 288.700000000000, 288.800000000000, 288.900000000000, 289, 289.100000000000, 289.200000000000, 289.300000000000, 289.400000000000, 289.500000000000, 289.600000000000, 289.700000000000, 289.800000000000, 289.900000000000, 290, 290.100000000000, 290.200000000000, 290.300000000000, 290.400000000000, 290.500000000000, 290.600000000000, 290.700000000000, 290.800000000000, 290.900000000000, 291, 291.100000000000, 291.200000000000, 291.300000000000, 291.400000000000, 291.500000000000, 291.600000000000, 291.700000000000, 291.800000000000, 291.900000000000, 292, 292.100000000000, 292.200000000000, 292.300000000000, 292.400000000000, 292.500000000000, 292.600000000000, 292.700000000000, 292.800000000000, 292.900000000000, 293, 293.100000000000, 293.200000000000, 293.300000000000, 293.400000000000, 293.500000000000, 293.600000000000, 293.700000000000, 293.800000000000, 293.900000000000, 294, 294.100000000000, 294.200000000000, 294.300000000000, 294.400000000000, 294.500000000000, 294.600000000000, 294.700000000000, 294.800000000000, 294.900000000000, 295, 295.100000000000, 295.200000000000, 295.300000000000, 295.400000000000, 295.500000000000, 295.600000000000, 295.700000000000, 295.800000000000, 295.900000000000, 296, 296.100000000000, 296.200000000000, 296.300000000000, 296.400000000000, 296.500000000000, 296.600000000000, 296.700000000000, 296.800000000000, 296.900000000000, 297, 297.100000000000, 297.200000000000, 297.300000000000, 297.400000000000, 297.500000000000, 297.600000000000, 297.700000000000, 297.800000000000, 297.900000000000, 298, 298.100000000000, 298.200000000000, 298.300000000000, 298.400000000000, 298.500000000000, 298.600000000000, 298.700000000000, 298.800000000000, 298.900000000000, 299, 299.100000000000, 299.200000000000, 299.300000000000, 299.400000000000, 299.500000000000, 299.600000000000, 299.700000000000, 299.800000000000, 299.900000000000, 300, 300.100000000000, 300.200000000000, 300.300000000000, 300.400000000000, 300.500000000000, 300.600000000000, 300.700000000000, 300.800000000000, 300.900000000000, 301, 301.100000000000, 301.200000000000, 301.300000000000, 301.400000000000, 301.500000000000, 301.600000000000, 301.700000000000, 301.800000000000, 301.900000000000, 302, 302.100000000000, 302.200000000000, 302.300000000000, 302.400000000000, 302.500000000000, 302.600000000000, 302.700000000000, 302.800000000000, 302.900000000000, 303, 303.100000000000, 303.200000000000, 303.300000000000, 303.400000000000, 303.500000000000, 303.600000000000, 303.700000000000, 303.800000000000, 303.900000000000, 304, 304.100000000000, 304.200000000000, 304.300000000000, 304.400000000000, 304.500000000000, 304.600000000000, 304.700000000000, 304.800000000000, 304.900000000000, 305, 305.100000000000, 305.200000000000, 305.300000000000, 305.400000000000, 305.500000000000, 305.600000000000, 305.700000000000, 305.800000000000, 305.900000000000, 306, 306.100000000000, 306.200000000000, 306.300000000000, 306.400000000000, 306.500000000000, 306.600000000000, 306.700000000000, 306.800000000000, 306.900000000000, 307, 307.100000000000, 307.200000000000, 307.300000000000, 307.400000000000, 307.500000000000, 307.600000000000, 307.700000000000, 307.800000000000, 307.900000000000, 308, 308.100000000000, 308.200000000000, 308.300000000000, 308.400000000000, 308.500000000000, 308.600000000000, 308.700000000000, 308.800000000000, 308.900000000000, 309, 309.100000000000, 309.200000000000, 309.300000000000, 309.400000000000, 309.500000000000, 309.600000000000, 309.700000000000, 309.800000000000, 309.900000000000, 310, 310.100000000000, 310.200000000000, 310.300000000000, 310.400000000000, 310.500000000000, 310.600000000000, 310.700000000000, 310.800000000000, 310.900000000000, 311, 311.100000000000, 311.200000000000, 311.300000000000, 311.400000000000, 311.500000000000, 311.600000000000, 311.700000000000, 311.800000000000, 311.900000000000, 312, 312.100000000000, 312.200000000000, 312.300000000000, 312.400000000000, 312.500000000000, 312.600000000000, 312.700000000000, 312.800000000000, 312.900000000000, 313, 313.100000000000, 313.200000000000, 313.300000000000, 313.400000000000, 313.500000000000, 313.600000000000, 313.700000000000, 313.800000000000, 313.900000000000, 314, 314.100000000000, 314.200000000000, 314.300000000000, 314.400000000000, 314.500000000000, 314.600000000000, 314.700000000000, 314.800000000000, 314.900000000000, 315, 315.100000000000, 315.200000000000, 315.300000000000, 315.400000000000, 315.500000000000, 315.600000000000, 315.700000000000, 315.800000000000, 315.900000000000, 316, 316.100000000000, 316.200000000000, 316.300000000000, 316.400000000000, 316.500000000000, 316.600000000000, 316.700000000000, 316.800000000000, 316.900000000000, 317, 317.100000000000, 317.200000000000, 317.300000000000, 317.400000000000, 317.500000000000, 317.600000000000, 317.700000000000, 317.800000000000, 317.900000000000, 318, 318.100000000000, 318.200000000000, 318.300000000000, 318.400000000000, 318.500000000000, 318.600000000000, 318.700000000000, 318.800000000000, 318.900000000000, 319, 319.100000000000, 319.200000000000, 319.300000000000, 319.400000000000, 319.500000000000, 319.600000000000, 319.700000000000, 319.800000000000, 319.900000000000, 320, 320.100000000000, 320.200000000000, 320.300000000000, 320.400000000000, 320.500000000000, 320.600000000000, 320.700000000000, 320.800000000000, 320.900000000000, 321, 321.100000000000, 321.200000000000, 321.300000000000, 321.400000000000, 321.500000000000, 321.600000000000, 321.700000000000, 321.800000000000, 321.900000000000, 322, 322.100000000000, 322.200000000000, 322.300000000000, 322.400000000000, 322.500000000000, 322.600000000000, 322.700000000000, 322.800000000000, 322.900000000000, 323, 323.100000000000, 323.200000000000, 323.300000000000, 323.400000000000, 323.500000000000, 323.600000000000, 323.700000000000, 323.800000000000, 323.900000000000, 324, 324.100000000000, 324.200000000000, 324.300000000000, 324.400000000000, 324.500000000000, 324.600000000000, 324.700000000000, 324.800000000000, 324.900000000000, 325, 325.100000000000, 325.200000000000, 325.300000000000, 325.400000000000, 325.500000000000, 325.600000000000, 325.700000000000, 325.800000000000, 325.900000000000, 326, 326.100000000000, 326.200000000000, 326.300000000000, 326.400000000000, 326.500000000000, 326.600000000000, 326.700000000000, 326.800000000000, 326.900000000000, 327, 327.100000000000, 327.200000000000, 327.300000000000, 327.400000000000, 327.500000000000, 327.600000000000, 327.700000000000, 327.800000000000, 327.900000000000, 328, 328.100000000000, 328.200000000000, 328.300000000000, 328.400000000000, 328.500000000000, 328.600000000000, 328.700000000000, 328.800000000000, 328.900000000000, 329, 329.100000000000, 329.200000000000, 329.300000000000, 329.400000000000, 329.500000000000, 329.600000000000, 329.700000000000, 329.800000000000, 329.900000000000, 330, 330.100000000000, 330.200000000000, 330.300000000000, 330.400000000000, 330.500000000000, 330.600000000000, 330.700000000000, 330.800000000000, 330.900000000000, 331, 331.100000000000, 331.200000000000, 331.300000000000, 331.400000000000, 331.500000000000, 331.600000000000, 331.700000000000, 331.800000000000, 331.900000000000, 332, 332.100000000000, 332.200000000000, 332.300000000000, 332.400000000000, 332.500000000000, 332.600000000000, 332.700000000000, 332.800000000000, 332.900000000000, 333, 333.100000000000, 333.200000000000, 333.300000000000, 333.400000000000, 333.500000000000, 333.600000000000, 333.700000000000, 333.800000000000, 333.900000000000, 334, 334.100000000000, 334.200000000000, 334.300000000000, 334.400000000000, 334.500000000000, 334.600000000000, 334.700000000000, 334.800000000000, 334.900000000000, 335, 335.100000000000, 335.200000000000, 335.300000000000, 335.400000000000, 335.500000000000, 335.600000000000, 335.700000000000, 335.800000000000, 335.900000000000, 336, 336.100000000000, 336.200000000000, 336.300000000000, 336.400000000000, 336.500000000000, 336.600000000000, 336.700000000000, 336.800000000000, 336.900000000000, 337, 337.100000000000, 337.200000000000, 337.300000000000, 337.400000000000, 337.500000000000, 337.600000000000, 337.700000000000, 337.800000000000, 337.900000000000, 338, 338.100000000000, 338.200000000000, 338.300000000000, 338.400000000000, 338.500000000000, 338.600000000000, 338.700000000000, 338.800000000000, 338.900000000000, 339, 339.100000000000, 339.200000000000, 339.300000000000, 339.400000000000, 339.500000000000, 339.600000000000, 339.700000000000, 339.800000000000, 339.900000000000, 340, 340.100000000000, 340.200000000000, 340.300000000000, 340.400000000000, 340.500000000000, 340.600000000000, 340.700000000000, 340.800000000000, 340.900000000000, 341, 341.100000000000, 341.200000000000, 341.300000000000]
    logPowerPerfEEG = [0.946682341221846, 0.941745520108470, 0.936838987574603, 0.931962579828465, 0.927116133820960, 0.922299487242931, 0.917512478522418, 0.912754946821932, 0.908026732035731, 0.903327674787109, 0.898657616425687, 0.894016399024719, 0.889403865378398, 0.884819858999178, 0.880264224115098, 0.875736805667118, 0.871237449306457, 0.866766001391946, 0.862322308987387, 0.857906219858913, 0.853517582472366, 0.849156245990676, 0.844822060271250, 0.840514875863369, 0.836234544005592, 0.831980916623168, 0.827753846325455, 0.823553186403347, 0.819378790826714, 0.815230514241836, 0.811108211968860, 0.807011739999257, 0.802940954993285, 0.798895714277464, 0.794875875842057, 0.790881298338556, 0.786911841077180, 0.782967364024378, 0.779047727800339, 0.775152793676510, 0.771282423573122, 0.767436480056725, 0.763614826337724, 0.759817326267933, 0.756043844338124, 0.752294245675597, 0.748568396041741, 0.744866161829620, 0.741187410061552, 0.737532008386706, 0.733899825078696, 0.730290729033191, 0.726704589765528, 0.723141277408334, 0.719600662709151, 0.716082617028077, 0.712587012335405, 0.709113721209272, 0.705662616833320, 0.702233572994356, 0.698826464080027, 0.695441165076494, 0.692077551566124, 0.688735499725176, 0.685414886321503, 0.682115588712261, 0.678837484841621, 0.675580453238486, 0.672344373014227, 0.669129123860409, 0.665934586046538, 0.662760640417809, 0.659607168392860, 0.656474051961533, 0.653361173682648, 0.650268416681775, 0.647195664649020, 0.644142801836811, 0.641109713057699, 0.638096283682159, 0.635102399636402, 0.632127947400190, 0.629172814004660, 0.626236887030157, 0.623320054604070, 0.620422205398676, 0.617543228628989, 0.614683014050622, 0.611841451957645, 0.609018433180461, 0.606213849083683, 0.603427591564013, 0.600659553048139, 0.597909626490628, 0.595177705371832, 0.592463683695796, 0.589767455988177, 0.587088917294170, 0.584427963176429, 0.581784489713015, 0.579158393495330, 0.576549571626069, 0.573957921717178, 0.571383341887815, 0.568825730762318, 0.566284987468178, 0.563761011634028, 0.561253703387621, 0.558762963353831, 0.556288692652652, 0.553830792897204, 0.551389166191746, 0.548963715129698, 0.546554342791664, 0.544160952743463, 0.541783449034172, 0.539421736194168, 0.537075719233176, 0.534745303638331, 0.532430395372238, 0.530130900871041, 0.527846727042501, 0.525577781264075, 0.523323971381006, 0.521085205704415, 0.518861393009403, 0.516652442533154, 0.514458263973050, 0.512278767484789, 0.510113863680507, 0.507963463626912, 0.505827478843416, 0.503705821300280, 0.501598403416761, 0.499505138059270, 0.497425938539525, 0.495360718612723, 0.493309392475711, 0.491271874765163, 0.489248080555763, 0.487237925358396, 0.485241325118346, 0.483258196213493, 0.481288455452521, 0.479332020073137, 0.477388807740280, 0.475458736544356, 0.473541724999460, 0.471637692041618, 0.469746557027027, 0.467868239730303, 0.466002660342734, 0.464149739470540, 0.462309398133136, 0.460481557761407, 0.458666140195979, 0.456863067685505, 0.455072262884950, 0.453293648853886, 0.451527149054789, 0.449772687351344, 0.448030188006758, 0.446299575682071, 0.444580775434479, 0.442873712715661, 0.441178313370111, 0.439494503633475, 0.437822210130898, 0.436161359875364, 0.434511880266062, 0.432873699086734, 0.431246744504049, 0.429630945065968, 0.428026229700124, 0.426432527712198, 0.424849768784312, 0.423277882973416, 0.421716800709691, 0.420166452794943, 0.418626770401021, 0.417097685068224, 0.415579128703723, 0.414071033579984, 0.412573332333194, 0.411085957961704, 0.409608843824462, 0.408141923639457, 0.406685131482177, 0.405238401784057, 0.403801669330945, 0.402374869261564, 0.400957937065985, 0.399550808584105, 0.398153420004126, 0.396765707861042, 0.395387609035131, 0.394019060750454, 0.392660000573354, 0.391310366410962, 0.389970096509714, 0.388639129453866, 0.387317404164012, 0.386004859895618, 0.384701436237551, 0.383407073110616, 0.382121710766098, 0.380845289784311, 0.379577751073150, 0.378319035866645, 0.377069085723530, 0.375827842525800, 0.374595248477295, 0.373371246102265, 0.372155778243962, 0.370948788063219, 0.369750219037047, 0.368560014957228, 0.367378119928916, 0.366204478369247, 0.365039035005946, 0.363881734875943, 0.362732523323996, 0.361591346001316, 0.360458148864192, 0.359332878172632, 0.358215480489001, 0.357105902676662, 0.356004091898630, 0.354909995616221, 0.353823561587712, 0.352744737867007, 0.351673472802301, 0.350609715034754, 0.349553413497166, 0.348504517412662, 0.347462976293378, 0.346428739939146, 0.345401758436199, 0.344381982155863, 0.343369361753267, 0.342363848166048, 0.341365392613067, 0.340373946593128, 0.339389461883699, 0.338411890539641, 0.337441184891939, 0.336477297546436, 0.335520181382580, 0.334569789552162, 0.333626075478071, 0.332688992853044, 0.331758495638430, 0.330834538062948, 0.329917074621455, 0.329006060073721, 0.328101449443201, 0.327203198015821, 0.326311261338755, 0.325425595219224, 0.324546155723279, 0.323672899174609, 0.322805782153335, 0.321944761494822, 0.321089794288487, 0.320240837876615, 0.319397849853179, 0.318560788062662, 0.317729610598885, 0.316904275803841, 0.316084742266530, 0.315270968821799, 0.314462914549187, 0.313660538771773, 0.312863801055035, 0.312072661205695, 0.311287079270595, 0.310507015535550, 0.309732430524225, 0.308963284997007, 0.308199539949881, 0.307441156613315, 0.306688096451144, 0.305940321159460, 0.305197792665508, 0.304460473126581, 0.303728324928926, 0.303001310686648, 0.302279393240622, 0.301562535657404, 0.300850701228153, 0.300143853467551, 0.299441956112732, 0.298744973122207, 0.298052868674805, 0.297365607168605, 0.296683153219882, 0.296005471662052, 0.295332527544619, 0.294664286132135, 0.294000712903153, 0.293341773549187, 0.292687433973682, 0.292037660290982, 0.291392418825301, 0.290751676109702, 0.290115398885076, 0.289483554099129, 0.288856108905369, 0.288233030662096, 0.287614286931402, 0.286999845478170, 0.286389674269073, 0.285783741471589, 0.285182015453003, 0.284584464779428, 0.283991058214824, 0.283401764720014, 0.282816553451716, 0.282235393761569, 0.281658255195168, 0.281085107491099, 0.280515920579982, 0.279950664583512, 0.279389309813511, 0.278831826770975, 0.278278186145132, 0.277728358812497, 0.277182315835941, 0.276640028463750, 0.276101468128696, 0.275566606447113, 0.275035415217971, 0.274507866421954, 0.273983932220551, 0.273463584955132, 0.272946797146048, 0.272433541491723, 0.271923790867746, 0.271417518325981, 0.270914697093664, 0.270415300572516, 0.269919302337849, 0.269426676137688, 0.268937395891884, 0.268451435691238, 0.267968769796625, 0.267489372638124, 0.267013218814148, 0.266540283090581, 0.266070540399917, 0.265603965840399, 0.265140534675172, 0.264680222331421, 0.264223004399530, 0.263768856632239, 0.263317754943797, 0.262869675409129, 0.262424594262997, 0.261982487899174, 0.261543332869611, 0.261107105883615, 0.260673783807027, 0.260243343661403, 0.259815762623199, 0.259391018022960, 0.258969087344512, 0.258549948224155, 0.258133578449860, 0.257719955960472, 0.257309058844917, 0.256900865341401, 0.256495353836628, 0.256092502865013, 0.255692291107894, 0.255294697392757, 0.254899700692457, 0.254507280124443, 0.254117414949990, 0.253730084573432, 0.253345268541392, 0.252962946542027, 0.252583098404265, 0.252205704097053, 0.251830743728606, 0.251458197545650, 0.251088045932685, 0.250720269411237, 0.250354848639120, 0.249991764409697, 0.249630997651148, 0.249272529425736, 0.248916340929082, 0.248562413489439, 0.248210728566969, 0.247861267753025, 0.247514012769433, 0.247168945467780, 0.246826047828704, 0.246485301961184, 0.246146690101839, 0.245810194614225, 0.245475797988135, 0.245143482838904, 0.244813231906716, 0.244485028055913, 0.244158854274310, 0.243834693672510, 0.243512529483219, 0.243192345060571, 0.242874123879452, 0.242557849534823, 0.242243505741056, 0.241931076331261, 0.241620545256622, 0.241311896585740, 0.241005114503966, 0.240700183312752, 0.240397087428992, 0.240095811384375, 0.239796339824732, 0.239498657509399, 0.239202749310563, 0.238908600212632, 0.238616195311593, 0.238325519814376, 0.238036559038227, 0.237749298410074, 0.237463723465904, 0.237179819850134, 0.236897573314999, 0.236616969719924, 0.236337995030912, 0.236060635319932, 0.235784876764305, 0.235510705646099, 0.235238108351522, 0.234967071370320, 0.234697581295176, 0.234429624821111, 0.234163188744892, 0.233898259964437, 0.233634825478227, 0.233372872384715, 0.233112387881747, 0.232853359265973, 0.232595773932272, 0.232339619373171, 0.232084883178274, 0.231831553033689, 0.231579616721453, 0.231329062118973, 0.231079877198453, 0.230832050026339, 0.230585568762752, 0.230340421660938, 0.230096597066706, 0.229854083417879, 0.229612869243745, 0.229372943164506, 0.229134293890735, 0.228896910222832, 0.228660781050484, 0.228425895352127, 0.228192242194411, 0.227959810731664, 0.227728590205365, 0.227498569943610, 0.227269739360593, 0.227042087956076, 0.226815605314868, 0.226590281106310, 0.226366105083754, 0.226143067084051, 0.225921157027035, 0.225700364915018, 0.225480680832279, 0.225262094944559, 0.225044597498558, 0.224828178821438, 0.224612829320316, 0.224398539481776, 0.224185299871369, 0.223973101133127, 0.223761933989068, 0.223551789238710, 0.223342657758589, 0.223134530501773, 0.222927398497379, 0.222721252850103, 0.222516084739734, 0.222311885420685, 0.222108646221522, 0.221906358544487, 0.221705013865044, 0.221504603731397, 0.221305119764041, 0.221106553655289, 0.220908897168824, 0.220712142139233, 0.220516280471559, 0.220321304140842, 0.220127205191673, 0.219933975737745, 0.219741607961406, 0.219550094113212, 0.219359426511490, 0.219169597541894, 0.218980599656968, 0.218792425375709, 0.218605067283138, 0.218418518029858, 0.218232770331635, 0.218047816968964, 0.217863650786645, 0.217680264693356, 0.217497651661238, 0.217315804725468, 0.217134716983846, 0.216954381596376, 0.216774791784854, 0.216595940832457, 0.216417822083331, 0.216240428942183, 0.216063754873878, 0.215887793403030, 0.215712538113606, 0.215537982648520, 0.215364120709239, 0.215190946055384, 0.215018452504339, 0.214846633930854, 0.214675484266657, 0.214504997500067, 0.214335167675602, 0.214165988893601, 0.213997455309834, 0.213829561135128, 0.213662300634979, 0.213495668129186, 0.213329657991461, 0.213164264649070, 0.212999482582452, 0.212835306324849, 0.212671730461945, 0.212508749631487, 0.212346358522936, 0.212184551877091, 0.212023324485737, 0.211862671191277, 0.211702586886384, 0.211543066513638, 0.211384105065178, 0.211225697582344, 0.211067839155333, 0.210910524922842, 0.210753750071732, 0.210597509836673, 0.210441799499807, 0.210286614390402, 0.210131949884517, 0.209977801404661, 0.209824164419454, 0.209671034443297, 0.209518407036037, 0.209366277802636, 0.209214642392839, 0.209063496500850, 0.208912835865002, 0.208762656267434, 0.208612953533771, 0.208463723532796, 0.208314962176137, 0.208166665417944, 0.208018829254576, 0.207871449724284, 0.207724522906901, 0.207578044923526, 0.207432011936216, 0.207286420147682, 0.207141265800977, 0.206996545179191, 0.206852254605152, 0.206708390441119, 0.206564949088484, 0.206421926987478, 0.206279320616860, 0.206137126493643, 0.205995341172776, 0.205853961246871, 0.205712983345901, 0.205572404136915, 0.205432220323749, 0.205292428646740, 0.205153025882440, 0.205014008843336, 0.204875374377566, 0.204737119368636, 0.204599240735148, 0.204461735430516, 0.204324600442696, 0.204187832793906, 0.204051429540359, 0.203915387771989, 0.203779704612178, 0.203644377217495, 0.203509402777422, 0.203374778514091, 0.203240501682024, 0.203106569567864, 0.202972979490118, 0.202839728798897, 0.202706814875654, 0.202574235132933, 0.202441987014109, 0.202310067993137, 0.202178475574298, 0.202047207291947, 0.201916260710261, 0.201785633423002, 0.201655323053254, 0.201525327253189, 0.201395643703813, 0.201266270114739, 0.201137204223924, 0.201008443797448, 0.200879986629263, 0.200751830540963, 0.200623973381539, 0.200496413027155, 0.200369147380906, 0.200242174372592, 0.200115491958481, 0.199989098121087, 0.199862990868931, 0.199737168236325, 0.199611628283143, 0.199486369094590, 0.199361388780991, 0.199236685477560, 0.199112257344183, 0.198988102565196, 0.198864219349172, 0.198740605928699, 0.198617260560173, 0.198494181523571, 0.198371367122249, 0.198248815682727, 0.198126525554477, 0.198004495109716, 0.197882722743196, 0.197761206872000, 0.197639945935334, 0.197518938394323, 0.197398182731813, 0.197277677452159, 0.197157421081032, 0.197037412165221, 0.196917649272429, 0.196798130991079, 0.196678855930117, 0.196559822718817, 0.196441030006591, 0.196322476462793, 0.196204160776530, 0.196086081656465, 0.195968237830643, 0.195850628046288, 0.195733251069622, 0.195616105685687, 0.195499190698146, 0.195382504929111, 0.195266047218959, 0.195149816426146, 0.195033811427035, 0.194918031115712, 0.194802474403808, 0.194687140220324, 0.194572027511459, 0.194457135240429, 0.194342462387298, 0.194228007948808, 0.194113770938201, 0.193999750385054, 0.193885945335113, 0.193772354850116, 0.193658978007635, 0.193545813900907, 0.193432861638669, 0.193320120344993, 0.193207589159128, 0.193095267235331, 0.192983153742719, 0.192871247865094, 0.192759548800797, 0.192648055762546, 0.192536767977276, 0.192425684685992, 0.192314805143604, 0.192204128618788, 0.192093654393817, 0.191983381764423, 0.191873310039640, 0.191763438541656, 0.191653766605666, 0.191544293579724, 0.191435018824593, 0.191325941713606, 0.191217061632519, 0.191108377979361, 0.190999890164304, 0.190891597609512, 0.190783499749000, 0.190675596028504, 0.190567885905329, 0.190460368848224, 0.190353044337232, 0.190245911863566, 0.190138970929468, 0.190032221048074, 0.189925661743286, 0.189819292549636, 0.189713113012153, 0.189607122686237, 0.189501321137523, 0.189395707941763, 0.189290282684688, 0.189185044961885, 0.189079994378676, 0.188975130549981, 0.188870453100203, 0.188765961663107, 0.188661655881686, 0.188557535408054, 0.188453599903311, 0.188349849037438, 0.188246282489162, 0.188142899945850, 0.188039701103387, 0.187936685666060, 0.187833853346442, 0.187731203865281, 0.187628736951378, 0.187526452341482, 0.187424349780175, 0.187322429019757, 0.187220689820142, 0.187119131948741, 0.187017755180361, 0.186916559297088, 0.186815544088187, 0.186714709349989, 0.186614054885794, 0.186513580505755, 0.186413286026784, 0.186313171272439, 0.186213236072832, 0.186113480264515, 0.186013903690394, 0.185914506199610, 0.185815287647456, 0.185716247895269, 0.185617386810336, 0.185518704265795, 0.185420200140535, 0.185321874319114, 0.185223726691639, 0.185125757153699, 0.185027965606255, 0.184930351955551, 0.184832916113020, 0.184735657995198, 0.184638577523632, 0.184541674624781, 0.184444949229940, 0.184348401275144, 0.184252030701082, 0.184155837453007, 0.184059821480659, 0.183963982738166, 0.183868321183965, 0.183772836780725, 0.183677529495253, 0.183582399298415, 0.183487446165055, 0.183392670073908, 0.183298071007530, 0.183203648952206, 0.183109403897875, 0.183015335838060, 0.182921444769772, 0.182827730693448, 0.182734193612864, 0.182640833535069, 0.182547650470296, 0.182454644431903, 0.182361815436283, 0.182269163502802, 0.182176688653720, 0.182084390914120, 0.181992270311838, 0.181900326877390, 0.181808560643906, 0.181716971647047, 0.181625559924954, 0.181534325518170, 0.181443268469563, 0.181352388824281, 0.181261686629663, 0.181171161935187, 0.181080814792395, 0.180990645254840, 0.180900653378007, 0.180810839219259, 0.180721202837774, 0.180631744294474, 0.180542463651976, 0.180453360974514, 0.180364436327895, 0.180275689779426, 0.180187121397865, 0.180098731253349, 0.180010519417348, 0.179922485962597, 0.179834630963048, 0.179746954493805, 0.179659456631067, 0.179572137452082, 0.179484997035083, 0.179398035459239, 0.179311252804590, 0.179224649152005, 0.179138224583127, 0.179051979180315, 0.178965913026601, 0.178880026205620, 0.178794318801587, 0.178708790899219, 0.178623442583705, 0.178538273940640, 0.178453285055998, 0.178368476016057, 0.178283846907368, 0.178199397816711, 0.178115128831031, 0.178031040037407, 0.177947131523004, 0.177863403375014, 0.177779855680632, 0.177696488526992, 0.177613302001137, 0.177530296189968, 0.177447471180208, 0.177364827058347, 0.177282363910616, 0.177200081822922, 0.177117980880847, 0.177036061169559, 0.176954322773806, 0.176872765777861, 0.176791390265496, 0.176710196319920, 0.176629184023770, 0.176548353459046, 0.176467704707092, 0.176387237848548, 0.176306952963324, 0.176226850130555, 0.176146929428563, 0.176067190934832, 0.175987634725971, 0.175908260877668, 0.175829069464671, 0.175750060560750, 0.175671234238652, 0.175592590570084, 0.175514129625682, 0.175435851474963, 0.175357756186301, 0.175279843826901, 0.175202114462769, 0.175124568158674, 0.175047204978114, 0.174970024983308, 0.174893028235147, 0.174816214793167, 0.174739584715535, 0.174663138059004, 0.174586874878891, 0.174510795229064, 0.174434899161893, 0.174359186728232, 0.174283657977400, 0.174208312957155, 0.174133151713658, 0.174058174291448, 0.173983380733441, 0.173908771080869, 0.173834345373294, 0.173760103648556, 0.173686045942763, 0.173612172290264, 0.173538482723638, 0.173464977273649, 0.173391655969253, 0.173318518837549, 0.173245565903782, 0.173172797191301, 0.173100212721560, 0.173027812514079, 0.172955596586441, 0.172883564954255, 0.172811717631160, 0.172740054628780, 0.172668575956731, 0.172597281622586, 0.172526171631857, 0.172455245987999, 0.172384504692364, 0.172313947744204, 0.172243575140653, 0.172173386876696, 0.172103382945173, 0.172033563336758, 0.171963928039932, 0.171894477040987, 0.171825210323987, 0.171756127870791, 0.171687229661004, 0.171618515671971, 0.171549985878782, 0.171481640254244, 0.171413478768860, 0.171345501390840, 0.171277708086075, 0.171210098818125, 0.171142673548205, 0.171075432235187, 0.171008374835570, 0.170941501303492, 0.170874811590693, 0.170808305646535, 0.170741983417963, 0.170675844849517, 0.170609889883309, 0.170544118459029, 0.170478530513910, 0.170413125982758, 0.170347904797912, 0.170282866889230, 0.170218012184129, 0.170153340607523, 0.170088852081853, 0.170024546527055, 0.169960423860568, 0.169896483997323, 0.169832726849745, 0.169769152327728, 0.169705760338654, 0.169642550787361, 0.169579523576164, 0.169516678604828, 0.169454015770572, 0.169391534968076, 0.169329236089458, 0.169267119024271, 0.169205183659524, 0.169143429879640, 0.169081857566490, 0.169020466599360, 0.168959256854965, 0.168898228207451, 0.168837380528366, 0.168776713686685, 0.168716227548803, 0.168655921978517, 0.168595796837046, 0.168535851983012, 0.168476087272452, 0.168416502558801, 0.168357097692912, 0.168297872523032, 0.168238826894829, 0.168179960651372, 0.168121273633122, 0.168062765677961, 0.168004436621177, 0.167946286295458, 0.167888314530898, 0.167830521155011, 0.167772905992699, 0.167715468866306, 0.167658209595559, 0.167601127997614, 0.167544223887037, 0.167487497075813, 0.167430947373354, 0.167374574586485, 0.167318378519456, 0.167262358973953, 0.167206515749081, 0.167150848641393, 0.167095357444861, 0.167040041950913, 0.166984901948415, 0.166929937223677, 0.166875147560463, 0.166820532739992, 0.166766092540943, 0.166711826739462, 0.166657735109162, 0.166603817421121, 0.166550073443909, 0.166496502943566, 0.166443105683631, 0.166389881425127, 0.166336829926578, 0.166283950944025, 0.166231244231002, 0.166178709538563, 0.166126346615289, 0.166074155207291, 0.166022135058207, 0.165970285909221, 0.165918607499055, 0.165867099564002, 0.165815761837902, 0.165764594052168, 0.165713595935793, 0.165662767215339, 0.165612107614974, 0.165561616856445, 0.165511294659121, 0.165461140739977, 0.165411154813599, 0.165361336592209, 0.165311685785676, 0.165262202101494, 0.165212885244828, 0.165163734918483, 0.165114750822960, 0.165065932656413, 0.165017280114701, 0.164968792891370, 0.164920470677676, 0.164872313162591, 0.164824320032805, 0.164776490972744, 0.164728825664576, 0.164681323788221, 0.164633985021361, 0.164586809039449, 0.164539795515715, 0.164492944121191, 0.164446254524696, 0.164399726392872, 0.164353359390179, 0.164307153178906, 0.164261107419183, 0.164215221768999, 0.164169495884205, 0.164123929418522, 0.164078522023559, 0.164033273348819, 0.163988183041715, 0.163943250747579, 0.163898476109651, 0.163853858769151, 0.163809398365211, 0.163765094534953, 0.163720946913458, 0.163676955133800, 0.163633118827043, 0.163589437622266, 0.163545911146570, 0.163502539025084, 0.163459320880984, 0.163416256335508, 0.163373345007938, 0.163330586515672, 0.163287980474175, 0.163245526497038, 0.163203224195948, 0.163161073180720, 0.163119073059342, 0.163077223437929, 0.163035523920763, 0.162993974110325, 0.162952573607271, 0.162911322010468, 0.162870218917011, 0.162829263922198, 0.162788456619606, 0.162747796601050, 0.162707283456615, 0.162666916774667, 0.162626696141878, 0.162586621143230, 0.162546691362007, 0.162506906379853, 0.162467265776747, 0.162427769131042, 0.162388416019454, 0.162349206017096, 0.162310138697482, 0.162271213632538, 0.162232430392643, 0.162193788546581, 0.162155287661630, 0.162116927303513, 0.162078707036461, 0.162040626423185, 0.162002685024909, 0.161964882401398, 0.161927218110944, 0.161889691710411, 0.161852302755217, 0.161815050799346, 0.161777935395413, 0.161740956094629, 0.161704112446818, 0.161667404000461, 0.161630830302688, 0.161594390899290, 0.161558085334736, 0.161521913152200, 0.161485873893552, 0.161449967099427, 0.161414192309138, 0.161378549060787, 0.161343036891244, 0.161307655336155, 0.161272403929959, 0.161237282205909, 0.161202289696084, 0.161167425931384, 0.161132690441600, 0.161098082755359, 0.161063602400192, 0.161029248902513, 0.160995021787641, 0.160960920579853, 0.160926944802345, 0.160893093977269, 0.160859367625738, 0.160825765267874, 0.160792286422789, 0.160758930608602, 0.160725697342471, 0.160692586140585, 0.160659596518211, 0.160626727989660, 0.160593980068357, 0.160561352266819, 0.160528844096681, 0.160496455068697, 0.160464184692800, 0.160432032478049, 0.160399997932691, 0.160368080564170, 0.160336279879128, 0.160304595383428, 0.160273026582175, 0.160241572979696, 0.160210234079624, 0.160179009384836, 0.160147898397508, 0.160116900619148, 0.160086015550557, 0.160055242691875, 0.160024581542621, 0.159994031601649, 0.159963592367210, 0.159933263336952, 0.159903044007932, 0.159872933876615, 0.159842932438923, 0.159813039190234, 0.159783253625385, 0.159753575238711, 0.159724003524028, 0.159694537974660, 0.159665178083476, 0.159635923342877, 0.159606773244826, 0.159577727280848, 0.159548784942057, 0.159519945719160, 0.159491209102493, 0.159462574582008, 0.159434041647296, 0.159405609787619, 0.159377278491900, 0.159349047248748, 0.159320915546473, 0.159292882873111, 0.159264948716398, 0.159237112563858, 0.159209373902708, 0.159181732220006, 0.159154187002545, 0.159126737736941, 0.159099383909618, 0.159072125006821, 0.159044960514664, 0.159017889919068, 0.158990912705875, 0.158964028360790, 0.158937236369404, 0.158910536217238, 0.158883927389738, 0.158857409372278, 0.158830981650191, 0.158804643708789, 0.158778395033354, 0.158752235109159, 0.158726163421502, 0.158700179455687, 0.158674282697081, 0.158648472631056, 0.158622748743080, 0.158597110518708, 0.158571557443554, 0.158546089003376, 0.158520704684004, 0.158495403971435, 0.158470186351806, 0.158445051311398, 0.158419998336654, 0.158395026914248, 0.158370136531019, 0.158345326674010, 0.158320596830507, 0.158295946488039, 0.158271375134358, 0.158246882257511, 0.158222467345807, 0.158198129887853, 0.158173869372542, 0.158149685289095, 0.158125577127084, 0.158101544376371, 0.158077586527230, 0.158053703070257, 0.158029893496472, 0.158006157297245, 0.157982493964368, 0.157958902990084, 0.157935383866996, 0.157911936088221, 0.157888559147301, 0.157865252538228, 0.157842015755509, 0.157818848294136, 0.157795749649598, 0.157772719317891, 0.157749756795581, 0.157726861579727, 0.157704033167979, 0.157681271058532, 0.157658574750177, 0.157635943742273, 0.157613377534810, 0.157590875628348, 0.157568437524139, 0.157546062724004, 0.157523750730461, 0.157501501046676, 0.157479313176476, 0.157457186624382, 0.157435120895613, 0.157413115496106, 0.157391169932493, 0.157369283712151, 0.157347456343184, 0.157325687334493, 0.157303976195682, 0.157282322437176, 0.157260725570172, 0.157239185106644, 0.157217700559436, 0.157196271442121, 0.157174897269201, 0.157153577555939, 0.157132311818497, 0.157111099573896, 0.157089940339987, 0.157068833635588, 0.157047778980335, 0.157026775894807, 0.157005823900508, 0.156984922519825, 0.156964071276139, 0.156943269693727, 0.156922517297878, 0.156901813614785, 0.156881158171673, 0.156860550496740, 0.156839990119178, 0.156819476569189, 0.156799009377980, 0.156778588077825, 0.156758212201979, 0.156737881284801, 0.156717594861695, 0.156697352469097, 0.156677153644573, 0.156656997926728, 0.156636884855280, 0.156616813971072, 0.156596784816055, 0.156576796933264, 0.156556849866946, 0.156536943162410, 0.156517076366179, 0.156497249025879, 0.156477460690370, 0.156457710909656, 0.156437999234938, 0.156418325218595, 0.156398688414223, 0.156379088376653, 0.156359524661936, 0.156339996827293, 0.156320504431299, 0.156301047033675, 0.156281624195467, 0.156262235478923, 0.156242880447631, 0.156223558666430, 0.156204269701447, 0.156185013120111, 0.156165788491186, 0.156146595384685, 0.156127433372031, 0.156108302025904, 0.156089200920372, 0.156070129630810, 0.156051087733987, 0.156032074808002, 0.156013090432345, 0.155994134187862, 0.155975205656840, 0.155956304422862, 0.155937430070994, 0.155918582187690, 0.155899760360803, 0.155880964179591, 0.155862193234781, 0.155843447118530, 0.155824725424403, 0.155806027747454, 0.155787353684159, 0.155768702832505, 0.155750074791898, 0.155731469163229, 0.155712885548928, 0.155694323552831, 0.155675782780338, 0.155657262838341, 0.155638763335180, 0.155620283880809, 0.155601824086632, 0.155583383565587, 0.155564961932202, 0.155546558802486, 0.155528173793992, 0.155509806525892, 0.155491456618846, 0.155473123695115, 0.155454807378502, 0.155436507294423, 0.155418223069841, 0.155399954333360, 0.155381700715089, 0.155363461846808, 0.155345237361885, 0.155327026895265, 0.155308830083553, 0.155290646564945, 0.155272475979235, 0.155254317967946, 0.155236172174105, 0.155218038242487, 0.155199915819418, 0.155181804552972, 0.155163704092781, 0.155145614090223, 0.155127534198260, 0.155109464071579, 0.155091403366540, 0.155073351741106, 0.155055308855040, 0.155037274369679, 0.155019247948121, 0.155001229255124, 0.154983217957169, 0.154965213722416, 0.154947216220719, 0.154929225123738, 0.154911240104677, 0.154893260838651, 0.154875287002332, 0.154857318274214, 0.154839354334536, 0.154821394865140, 0.154803439549770, 0.154785488073830, 0.154767540124431, 0.154749595390479, 0.154731653562667, 0.154713714333364, 0.154695777396749, 0.154677842448720, 0.154659909186976, 0.154641977310977, 0.154624046521908, 0.154606116522789, 0.154588187018379, 0.154570257715240, 0.154552328321642, 0.154534398547766, 0.154516468105413, 0.154498536708355, 0.154480604071997, 0.154462669913640, 0.154444733952314, 0.154426795908896, 0.154408855506039, 0.154390912468177, 0.154372966521611, 0.154355017394375, 0.154337064816357, 0.154319108519242, 0.154301148236502, 0.154283183703515, 0.154265214657346, 0.154247240836960, 0.154229261983099, 0.154211277838401, 0.154193288147234, 0.154175292655769, 0.154157291112161, 0.154139283266247, 0.154121268869711, 0.154103247676116, 0.154085219440826, 0.154067183921005, 0.154049140875740, 0.154031090065834, 0.154013031254007, 0.153994964204766, 0.153976888684519, 0.153958804461446, 0.153940711305571, 0.153922608988782, 0.153904497284808, 0.153886375969181, 0.153868244819296, 0.153850103614419, 0.153831952135579, 0.153813790165725, 0.153795617489599, 0.153777433893789, 0.153759239166739, 0.153741033098730, 0.153722815481884, 0.153704586110182, 0.153686344779387, 0.153668091287192, 0.153649825433021, 0.153631547018270, 0.153613255846060, 0.153594951721421, 0.153576634451193, 0.153558303844050, 0.153539959710557, 0.153521601863041, 0.153503230115735, 0.153484844284639, 0.153466444187687, 0.153448029644506, 0.153429600476759, 0.153411156507705, 0.153392697562659, 0.153374223468622, 0.153355734054484, 0.153337229150929, 0.153318708590520, 0.153300172207632, 0.153281619838401, 0.153263051320931, 0.153244466494991, 0.153225865202274, 0.153207247286270, 0.153188612592275, 0.153169960967407, 0.153151292260635, 0.153132606322697, 0.153113903006114, 0.153095182165365, 0.153076443656566, 0.153057687337730, 0.153038913068658, 0.153020120710981, 0.153001310128094, 0.152982481185239, 0.152963633749340, 0.152944767689333, 0.152925882875682, 0.152906979180886, 0.152888056479081, 0.152869114646236, 0.152850153560112, 0.152831173100236, 0.152812173147911, 0.152793153586282, 0.152774114300156, 0.152755055176181, 0.152735976102828, 0.152716876970217, 0.152697757670315, 0.152678618096826, 0.152659458145182, 0.152640277712619, 0.152621076698113, 0.152601855002401, 0.152582612527970, 0.152563349178948, 0.152544064861315, 0.152524759482842, 0.152505432952873, 0.152486085182637, 0.152466716084936, 0.152447325574448, 0.152427913567439, 0.152408479982054, 0.152389024737942, 0.152369547756726, 0.152350048961445, 0.152330528277082, 0.152310985630150, 0.152291420949025, 0.152271834163620, 0.152252225205661, 0.152232594008441, 0.152212940507061, 0.152193264638207, 0.152173566340222, 0.152153845553299, 0.152134102219005, 0.152114336280810, 0.152094547683797, 0.152074736374648, 0.152054902301651, 0.152035045414899, 0.152015165665963, 0.151995263008187, 0.151975337396423, 0.151955388787238, 0.151935417138798, 0.151915422410888, 0.151895404564938, 0.151875363563947, 0.151855299372515, 0.151835211956895, 0.151815101284879, 0.151794967325934, 0.151774810051014, 0.151754629432756, 0.151734425445288, 0.151714198064383, 0.151693947267304, 0.151673673032982, 0.151653375341870, 0.151633054175917, 0.151612709518696, 0.151592341355245, 0.151571949672252, 0.151551534457871, 0.151531095701787, 0.151510633395240, 0.151490147530962, 0.151469638103198, 0.151449105107736, 0.151428548541829, 0.151407968404298, 0.151387364695292, 0.151366737416720, 0.151346086571704, 0.151325412165034, 0.151304714202857, 0.151283992692847, 0.151263247644100, 0.151242479067210, 0.151221686974216, 0.151200871378573, 0.151180032295204, 0.151159169740432, 0.151138283732025, 0.151117374289244, 0.151096441432657, 0.151075485184296, 0.151054505567591, 0.151033502607384, 0.151012476329875, 0.150991426762722, 0.150970353934890, 0.150949257876758, 0.150928138620083, 0.150906996197943, 0.150885830644871, 0.150864641996628, 0.150843430290401, 0.150822195564694, 0.150800937859368, 0.150779657215613, 0.150758353675863, 0.150737027283943, 0.150715678085028, 0.150694306125489, 0.150672911453132, 0.150651494116851, 0.150630054167013, 0.150608591655211, 0.150587106634284, 0.150565599158362, 0.150544069282773, 0.150522517064210, 0.150500942560544, 0.150479345830825, 0.150457726935538, 0.150436085936180, 0.150414422895560, 0.150392737877720, 0.150371030947876, 0.150349302172482, 0.150327551619155, 0.150305779356713, 0.150283985455138, 0.150262169985619, 0.150240333020494, 0.150218474633285, 0.150196594898643, 0.150174693892374, 0.150152771691430, 0.150130828373883, 0.150108864019024, 0.150086878707122, 0.150064872519644, 0.150042845539180, 0.150020797849372, 0.149998729535023, 0.149976640681928, 0.149954531377073, 0.149932401708381, 0.149910251765009, 0.149888081637055, 0.149865891415714, 0.149843681193204, 0.149821451062809, 0.149799201118823, 0.149776931456655, 0.149754642172527, 0.149732333363928, 0.149710005129113, 0.149687657567558, 0.149665290779505, 0.149642904866442, 0.149620499930541, 0.149598076075170, 0.149575633404545, 0.149553172023932, 0.149530692039433, 0.149508193558128, 0.149485676688066, 0.149463141538195, 0.149440588218440, 0.149418016839533, 0.149395427513204, 0.149372820351996, 0.149350195469449, 0.149327552979917, 0.149304892998602, 0.149282215641654, 0.149259521026064, 0.149236809269638, 0.149214080491049, 0.149191334809817, 0.149168572346332, 0.149145793221716, 0.149122997557946, 0.149100185477916, 0.149077357105182, 0.149054512564184, 0.149031651980048, 0.149008775478850, 0.148985883187276, 0.148962975232855, 0.148940051743889, 0.148917112849445, 0.148894158679237, 0.148871189363819,
                       0.148848205034422, 0.148825205823036, 0.148802191862315, 0.148779163285730, 0.148756120227293, 0.148733062821843, 0.148709991204840, 0.148686905512483, 0.148663805881533, 0.148640692449523, 0.148617565354642, 0.148594424735591, 0.148571270731881, 0.148548103483619, 0.148524923131476, 0.148501729816724, 0.148478523681382, 0.148455304867951, 0.148432073519565, 0.148408829780007, 0.148385573793553, 0.148362305705096, 0.148339025660090, 0.148315733804544, 0.148292430285056, 0.148269115248771, 0.148245788843259, 0.148222451216814, 0.148199102518059, 0.148175742896266, 0.148152372501166, 0.148128991483005, 0.148105599992515, 0.148082198180909, 0.148058786199942, 0.148035364201669, 0.148011932338866, 0.147988490764531, 0.147965039632285, 0.147941579096082, 0.147918109310287, 0.147894630429855, 0.147871142609971, 0.147847646006357, 0.147824140775173, 0.147800627072808, 0.147777105056164, 0.147753574882584, 0.147730036709645, 0.147706490695346, 0.147682936998078, 0.147659375776644, 0.147635807190019, 0.147612231397668, 0.147588648559340, 0.147565058835078, 0.147541462385305, 0.147517859370816, 0.147494249952445, 0.147470634291642, 0.147447012549951, 0.147423384889265, 0.147399751471657, 0.147376112459703, 0.147352468015980, 0.147328818303372, 0.147305163485205, 0.147281503724805, 0.147257839185878, 0.147234170032254, 0.147210496428023, 0.147186818537547, 0.147163136525301, 0.147139450556055, 0.147115760794531, 0.147092067406018, 0.147068370555617, 0.147044670408870, 0.147020967131227, 0.146997260888543, 0.146973551846623, 0.146949840171537, 0.146926126029399, 0.146902409586528, 0.146878691009322, 0.146854970464216, 0.146831248117938, 0.146807524137153, 0.146783798688628, 0.146760071939412, 0.146736344056211, 0.146712615206302, 0.146688885556739, 0.146665155274556, 0.146641424527069, 0.146617693481446, 0.146593962305139, 0.146570231165315, 0.146546500229270, 0.146522769664607, 0.146499039638439, 0.146475310318242, 0.146451581871364, 0.146427854465176, 0.146404128267006, 0.146380403444090, 0.146356680163813, 0.146332958593388, 0.146309238900030, 0.146285521250754, 0.146261805812867, 0.146238092753195, 0.146214382238875, 0.146190674436620, 0.146166969513263, 0.146143267635599, 0.146119568970105, 0.146095873683444, 0.146072181941847, 0.146048493911718, 0.146024809759136, 0.146001129650192, 0.145977453750727, 0.145953782226503, 0.145930115243153, 0.145906452966094, 0.145882795560605, 0.145859143191842, 0.145835496024812, 0.145811854224176, 0.145788217954522, 0.145764587380352, 0.145740962665795, 0.145717343974889, 0.145693731471349, 0.145670125318772, 0.145646525680513, 0.145622932719699, 0.145599346599218, 0.145575767481674, 0.145552195529537, 0.145528630904854, 0.145505073769603, 0.145481524285287, 0.145457982613449, 0.145434448915084, 0.145410923350852, 0.145387406081457, 0.145363897267061, 0.145340397067500, 0.145316905642509, 0.145293423151303, 0.145269949752900, 0.145246485605990, 0.145223030868840, 0.145199585699515, 0.145176150255654, 0.145152724694641, 0.145129309173290, 0.145105903848386, 0.145082508876134, 0.145059124412357, 0.145035750612614, 0.145012387632114, 0.144989035625476, 0.144965694747142, 0.144942365151130, 0.144919046990931, 0.144895740419798, 0.144872445590337, 0.144849162655079, 0.144825891765909, 0.144802633074245, 0.144779386731200, 0.144756152887420, 0.144732931693041, 0.144709723297926, 0.144686527851307, 0.144663345502063, 0.144640176398459, 0.144617020688542, 0.144593878519630, 0.144570750038765, 0.144547635392524, 0.144524534726731, 0.144501448186977, 0.144478375918313, 0.144455318065225, 0.144432274771607, 0.144409246181133, 0.144386232436707, 0.144363233680811, 0.144340250055378, 0.144317281701812, 0.144294328760990, 0.144271391373295, 0.144248469678441, 0.144225563815813, 0.144202673923995, 0.144179800141133, 0.144156942604910, 0.144134101452242, 0.144111276819656, 0.144088468842945, 0.144065677657551, 0.144042903398020, 0.144020146198633, 0.143997406192775, 0.143974683513624, 0.143951978293297, 0.143929290663729, 0.143906620755869, 0.143883968700388, 0.143861334627047, 0.143838718665287, 0.143816120943744, 0.143793541590492, 0.143770980732852, 0.143748438497658, 0.143725915011048, 0.143703410398459, 0.143680924784739, 0.143658458294284, 0.143636011050324, 0.143613583175997, 0.143591174793411, 0.143568786024105, 0.143546416989090, 0.143524067808439, 0.143501738601793, 0.143479429488022, 0.143457140585241, 0.143434872010929, 0.143412623881955, 0.143390396314437, 0.143368189423798, 0.143346003324678, 0.143323838131150, 0.143301693956479, 0.143279570913296, 0.143257469113429, 0.143235388668105, 0.143213329687786, 0.143191292282141, 0.143169276560271, 0.143147282630242, 0.143125310599806, 0.143103360575650, 0.143081432663916, 0.143059526969837, 0.143037643598168, 0.143015782652588, 0.142993944236179, 0.142972128451353, 0.142950335399754, 0.142928565182159, 0.142906817898567, 0.142885093648377, 0.142863392530075, 0.142841714641486, 0.142820060079480, 0.142798428940459, 0.142776821319793, 0.142755237312196, 0.142733677011560, 0.142712140510893, 0.142690627902687, 0.142669139278471, 0.142647674728882, 0.142626234343948, 0.142604818212880, 0.142583426424004, 0.142562059064998, 0.142540716222413, 0.142519397982409, 0.142498104430082, 0.142476835649913, 0.142455591725322, 0.142434372739061, 0.142413178773109, 0.142392009908596, 0.142370866225880, 0.142349747804175, 0.142328654722506, 0.142307587058585, 0.142286544889242, 0.142265528290902, 0.142244537338843, 0.142223572107721, 0.142202632671052, 0.142181719101956, 0.142160831472221, 0.142139969853246, 0.142119134315390, 0.142098324928138, 0.142077541760231, 0.142056784879512, 0.142036054353028, 0.142015350247018, 0.141994672626706, 0.141974021556635, 0.141953397100508, 0.141932799321026, 0.141912228280201, 0.141891684039224, 0.141871166658190, 0.141850676196665, 0.141830212712931, 0.141809776265057, 0.141789366909647, 0.141768984702710, 0.141748629699464, 0.141728301954133, 0.141708001520062, 0.141687728450029, 0.141667482795605, 0.141647264607568, 0.141627073936044, 0.141606910830087, 0.141586775337879, 0.141566667507157, 0.141546587384086, 0.141526535014502, 0.141506510443255, 0.141486513714171, 0.141466544870396, 0.141446603954263, 0.141426691006990, 0.141406806069232, 0.141386949180437, 0.141367120379509, 0.141347319704323, 0.141327547191843, 0.141307802878351, 0.141288086799056, 0.141268398988601, 0.141248739480215, 0.141229108306857, 0.141209505500364, 0.141189931091647, 0.141170385110896, 0.141150867587245, 0.141131378549159, 0.141111918024222, 0.141092486038952, 0.141073082619325, 0.141053707790014, 0.141034361575280, 0.141015043998423, 0.140995755081521, 0.140976494846263, 0.140957263313179, 0.140938060501983, 0.140918886431781, 0.140899741120388, 0.140880624585009, 0.140861536842084, 0.140842477907118, 0.140823447794565, 0.140804446518140, 0.140785474091059, 0.140766530525023, 0.140747615831397, 0.140728730020367, 0.140709873101748, 0.140691045083890, 0.140672245974591, 0.140653475780923, 0.140634734508896, 0.140616022163796, 0.140597338749931, 0.140578684270900, 0.140560058729548, 0.140541462127648, 0.140522894466064, 0.140504355745196, 0.140485845964365, 0.140467365122066, 0.140448913215995, 0.140430490243004, 0.140412096198988, 0.140393731079212, 0.140375394878137, 0.140357087589079, 0.140338809204986, 0.140320559717583, 0.140302339117917, 0.140284147396283, 0.140265984542006, 0.140247850543714, 0.140229745389146, 0.140211669065378, 0.140193621558529, 0.140175602853765, 0.140157612935879, 0.140139651788517, 0.140121719394422, 0.140103815735769, 0.140085940794025, 0.140068094549648, 0.140050276982203, 0.140032488070710, 0.140014727793344, 0.139996996127344, 0.139979293049439, 0.139961618535103, 0.139943972559529, 0.139926355096920, 0.139908766120444, 0.139891205603068, 0.139873673516343, 0.139856169831583, 0.139838694518897, 0.139821247548141, 0.139803828887730, 0.139786438505884, 0.139769076369919, 0.139751742446289, 0.139734436700683, 0.139717159097992, 0.139699909602554, 0.139682688178051, 0.139665494786937, 0.139648329391390, 0.139631191952573, 0.139614082431046, 0.139597000786626, 0.139579946978381, 0.139562920964593, 0.139545922702923, 0.139528952150150, 0.139512009262637, 0.139495093995611, 0.139478206304013, 0.139461346141516, 0.139444513461884, 0.139427708217357, 0.139410930360049, 0.139394179841132, 0.139377456610962, 0.139360760619542, 0.139344091815845, 0.139327450148566, 0.139310835565016, 0.139294248012882, 0.139277687437942, 0.139261153786139, 0.139244647002837, 0.139228167032107, 0.139211713817706, 0.139195287302682, 0.139178887429503, 0.139162514140064, 0.139146167375155, 0.139129847075470, 0.139113553180572, 0.139097285629955, 0.139081044362025, 0.139064829314707, 0.139048640425115, 0.139032477630184, 0.139016340865579, 0.139000230067040, 0.138984145169392, 0.138968086106388, 0.138952052812021, 0.138936045219123, 0.138920063260089, 0.138904106866786, 0.138888175970238, 0.138872270501276, 0.138856390389523, 0.138840535564882, 0.138824705955751, 0.138808901490836, 0.138793122097626, 0.138777367703395, 0.138761638234550, 0.138745933617339, 0.138730253777170, 0.138714598639066, 0.138698968127081, 0.138683362165558, 0.138667780677334, 0.138652223585710, 0.138636690812365, 0.138621182279381, 0.138605697907872, 0.138590237618696, 0.138574801331766, 0.138559388967000, 0.138544000443394, 0.138528635679802, 0.138513294594274, 0.138497977104469, 0.138482683127479, 0.138467412580287, 0.138452165379089, 0.138436941439698, 0.138421740677119, 0.138406563006389, 0.138391408341901, 0.138376276597502, 0.138361167686849, 0.138346081522820, 0.138331018017858, 0.138315977084395, 0.138300958633988, 0.138285962577896, 0.138270988827078, 0.138256037291997, 0.138241107882377, 0.138226200508057, 0.138211315078494, 0.138196451502112, 0.138181609687297, 0.138166789542396, 0.138151990974757, 0.138137213891833, 0.138122458200349, 0.138107723806746, 0.138093010617277, 0.138078318537796, 0.138063647473485, 0.138048997329569, 0.138034368010620, 0.138019759421161, 0.138005171465014, 0.137990604046098, 0.137976057067465, 0.137961530432120, 0.137947024043234, 0.137932537802637, 0.137918071612740, 0.137903625375328, 0.137889198991562, 0.137874792362917, 0.137860405390212, 0.137846037973956, 0.137831690014326, 0.137817361411802, 0.137803052065629, 0.137788761875661, 0.137774490740895, 0.137760238560406, 0.137746005233033, 0.137731790656765, 0.137717594730536, 0.137703417352014, 0.137689258418736, 0.137675117828470, 0.137660995478286, 0.137646891265810, 0.137632805087415, 0.137618736839873, 0.137604686419749, 0.137590653723274, 0.137576638646528, 0.137562641085325, 0.137548660935501, 0.137534698092565, 0.137520752451433, 0.137506823908170, 0.137492912357050, 0.137479017693215, 0.137465139811519, 0.137451278606079, 0.137437433971953, 0.137423605803042, 0.137409793993512, 0.137395998437384, 0.137382219028756, 0.137368455661370, 0.137354708228644, 0.137340976624442, 0.137327260742215, 0.137313560475212, 0.137299875716831, 0.137286206360287, 0.137272552298462, 0.137258913424811, 0.137245289632007, 0.137231680813005, 0.137218086860820, 0.137204507668026, 0.137190943127384, 0.137177393132129, 0.137163857574048, 0.137150336346269, 0.137136829341618, 0.137123336452069, 0.137109857570607, 0.137096392589638, 0.137082941401608, 0.137069503899346, 0.137056079975289, 0.137042669521645, 0.137029272431518, 0.137015888596893, 0.137002517910770, 0.136989160265360, 0.136975815553831, 0.136962483668672, 0.136949164502216, 0.136935857947979, 0.136922563898234, 0.136909282245884, 0.136896012884162, 0.136882755705978, 0.136869510604542, 0.136856277472650, 0.136843056203769, 0.136829846691512, 0.136816648828817, 0.136803462509505, 0.136790287627236, 0.136777124075635, 0.136763971748756, 0.136750830540344, 0.136737700344709, 0.136724581055680, 0.136711472568486, 0.136698374776795, 0.136685287575216, 0.136672210859374, 0.136659144523276, 0.136646088462723, 0.136633042572824, 0.136620006748741, 0.136606980886333, 0.136593964881407, 0.136580958629920, 0.136567962027959, 0.136554974972051, 0.136541997358401, 0.136529029084302, 0.136516070046343, 0.136503120141739, 0.136490179268189, 0.136477247322912, 0.136464324204158, 0.136451409809665, 0.136438504038167, 0.136425606787883, 0.136412717957696, 0.136399837446817, 0.136386965154414, 0.136374100980181, 0.136361244823779, 0.136348396585561, 0.136335556166135, 0.136322723465797, 0.136309898385300, 0.136297080826583, 0.136284270690720, 0.136271467880006, 0.136258672296094, 0.136245883841706, 0.136233102419354, 0.136220327932547, 0.136207560284591, 0.136194799379089, 0.136182045120036, 0.136169297412211, 0.136156556160217, 0.136143821269124, 0.136131092644429, 0.136118370191682, 0.136105653817506, 0.136092943428084, 0.136080238930408, 0.136067540231921, 0.136054847240183, 0.136042159863084, 0.136029478009351, 0.136016801587400, 0.136004130506774, 0.135991464676573, 0.135978804007316, 0.135966148409401, 0.135953497793286, 0.135940852070729, 0.135928211152732, 0.135915574952281, 0.135902943381016, 0.135890316352338, 0.135877693779493, 0.135865075576171, 0.135852461657297, 0.135839851936599, 0.135827246330013, 0.135814644752825, 0.135802047121224, 0.135789453351965, 0.135776863361638, 0.135764277068189, 0.135751694389411, 0.135739115243582, 0.135726539550112, 0.135713967227802, 0.135701398197096, 0.135688832378410, 0.135676269692516, 0.135663710060485, 0.135651153405045, 0.135638599648060, 0.135626048712665, 0.135613500522638, 0.135600955001398, 0.135588412073697, 0.135575871665007, 0.135563333700222, 0.135550798106350, 0.135538264809559, 0.135525733736984, 0.135513204816632, 0.135500677976794, 0.135488153146127, 0.135475630254405, 0.135463109231539, 0.135450590007907, 0.135438072514849, 0.135425556683696, 0.135413042447617, 0.135400529738522, 0.135388018490486, 0.135375508636936, 0.135363000113149, 0.135350492854187, 0.135337986795575, 0.135325481873915, 0.135312978025799, 0.135300475189435, 0.135287973302925, 0.135275472304825, 0.135262972134329, 0.135250472732511, 0.135237974039095, 0.135225475995633, 0.135212978544033, 0.135200481627034, 0.135187985187666, 0.135175489169432, 0.135162993517602, 0.135150498176255, 0.135138003092201, 0.135125508210552, 0.135113013479030, 0.135100518845592, 0.135088024257828, 0.135075529665205, 0.135063035017250, 0.135050540264258, 0.135038045356446, 0.135025550246646, 0.135013054886326, 0.135000559228287, 0.134988063227153, 0.134975566836084, 0.134963070010839, 0.134950572706139, 0.134938074878910, 0.134925576486044, 0.134913077485644, 0.134900577834948, 0.134888077494143, 0.134875576422107, 0.134863074580087, 0.134850571928476, 0.134838068429163, 0.134825564044980, 0.134813058739214, 0.134800552475358, 0.134788045217976, 0.134775536933071, 0.134763027585751, 0.134750517143096, 0.134738005572664, 0.134725492842765, 0.134712978921724, 0.134700463779244, 0.134687947386198, 0.134675429712508, 0.134662910731021, 0.134650390413032, 0.134637868732959, 0.134625345663490, 0.134612821179394, 0.134600295256766, 0.134587767870815, 0.134575238998818, 0.134562708618145, 0.134550176706421, 0.134537643243531, 0.134525108208553, 0.134512571582108, 0.134500033345151, 0.134487493479502, 0.134474951967937, 0.134462408793906, 0.134449863940988, 0.134437317394669, 0.134424769139393, 0.134412219162953, 0.134399667450772, 0.134387113991862, 0.134374558773655, 0.134362001786300, 0.134349443019241, 0.134336882463093, 0.134324320109821, 0.134311755950986, 0.134299189979516, 0.134286622189422, 0.134274052574982, 0.134261481131337, 0.134248907853960, 0.134236332740107, 0.134223755786942, 0.134211176991939, 0.134198596354815, 0.134186013874185, 0.134173429551527, 0.134160843386795, 0.134148255382462, 0.134135665540549, 0.134123073864800, 0.134110480358753, 0.134097885027463, 0.134085287876507, 0.134072688911600, 0.134060088140056, 0.134047485569177, 0.134034881207977, 0.134022275065741, 0.134009667151238, 0.133997057476158, 0.133984446051153, 0.133971832888582, 0.133959218001323, 0.133946601402822, 0.133933983107080, 0.133921363129459, 0.133908741486049, 0.133896118192019, 0.133883493265651, 0.133870866724233, 0.133858238586702, 0.133845608872636, 0.133832977601436, 0.133820344793884, 0.133807710472047, 0.133795074657506, 0.133782437373591, 0.133769798643626, 0.133757158492045, 0.133744516944028, 0.133731874024886, 0.133719229761539, 0.133706584180618, 0.133693937310558, 0.133681289179448, 0.133668639816649, 0.133655989252059, 0.133643337516403, 0.133630684640955, 0.133618030657389, 0.133605375598869, 0.133592719498017, 0.133580062390368, 0.133567404308876, 0.133554745290651, 0.133542085370086, 0.133529424585619, 0.133516762973237, 0.133504100571419, 0.133491437419513, 0.133478773556565, 0.133466109022422, 0.133453443858332, 0.133440778105214, 0.133428111805431, 0.133415445001651, 0.133402777736712, 0.133390110055303, 0.133377442001680, 0.133364773621303, 0.133352104959598, 0.133339436063386, 0.133326766980152, 0.133314097757566, 0.133301428443214, 0.133288759087309, 0.133276089738458, 0.133263420447331, 0.133250751265139, 0.133238082242877, 0.133225413432708, 0.133212744887558, 0.133200076659573, 0.133187408804561, 0.133174741375148, 0.133162074427363, 0.133149408016400, 0.133136742198380, 0.133124077029984, 0.133111412569134, 0.133098748872935, 0.133086086000305, 0.133073424009730, 0.133060762961397, 0.133048102914603, 0.133035443931147, 0.133022786070833, 0.133010129396850, 0.132997473969677, 0.132984819853427, 0.132972167110767, 0.132959515806203, 0.132946866003202, 0.132934217766240, 0.132921571161976, 0.132908926255677, 0.132896283113260, 0.132883641801943, 0.132871002388504, 0.132858364941677, 0.132845729529267, 0.132833096219280, 0.132820465081828, 0.132807836186221, 0.132795209602509, 0.132782585401306, 0.132769963653788, 0.132757344431209, 0.132744727805617, 0.132732113848818, 0.132719502634403, 0.132706894234437, 0.132694288723484, 0.132681686175123, 0.132669086663683, 0.132656490264254, 0.132643897051700, 0.132631307101457, 0.132618720490011, 0.132606137292969, 0.132593557587487, 0.132580981450659, 0.132568408960229, 0.132555840193035, 0.132543275227469, 0.132530714142801, 0.132518157016999, 0.132505603929196, 0.132493054959375, 0.132480510186430, 0.132467969692009, 0.132455433554511, 0.132442901856190, 0.132430374677279, 0.132417852099334, 0.132405334203668, 0.132392821072219, 0.132380312786837, 0.132367809430727, 0.132355311085120, 0.132342817834084, 0.132330329760050, 0.132317846946715, 0.132305369477568, 0.132292897435700, 0.132280430906526, 0.132267969973117, 0.132255514720341, 0.132243065232126, 0.132230621594206, 0.132218183891535, 0.132205752208275, 0.132193326631053, 0.132180907244347, 0.132168494134346, 0.132156087386428, 0.132143687087011, 0.132131293321801, 0.132118906177006, 0.132106525739337, 0.132094152094538, 0.132081785329541, 0.132069425530995, 0.132057072785355, 0.132044727180287, 0.132032388802074, 0.132020057736750, 0.132007734072601, 0.131995417896146, 0.131983109294788, 0.131970808355063, 0.131958515165401, 0.131946229811689, 0.131933952381496, 0.131921682962183, 0.131909421641331, 0.131897168505661, 0.131884923642192, 0.131872687138650, 0.131860459081977, 0.131848239559327, 0.131836028657828, 0.131823826464320, 0.131811633065600, 0.131799448549038, 0.131787273000584, 0.131775106508480, 0.131762949157751, 0.131750801035536, 0.131738662228939, 0.131726532823042, 0.131714412904746, 0.131702302560317, 0.131690201874901, 0.131678110934933, 0.131666029825469, 0.131653958632021, 0.131641897440557, 0.131629846335238, 0.131617805401109, 0.131605774723468, 0.131593754386096, 0.131581744473711, 0.131569745069766, 0.131557756258869, 0.131545778123412, 0.131533810748163, 0.131521854215481, 0.131509908607337, 0.131497974007803, 0.131486050498401, 0.131474138160083, 0.131462237075547, 0.131450347325813, 0.131438468992614, 0.131426602156407, 0.131414746896044, 0.131402903293298, 0.131391071426842, 0.131379251376052, 0.131367443220766, 0.131355647037946, 0.131343862906759, 0.131332090905363, 0.131320331109885, 0.131308583598846, 0.131296848447113, 0.131285125731659, 0.131273415527876, 0.131261717911153, 0.131250032956603, 0.131238360738071, 0.131226701328120, 0.131215054802545, 0.131203421232701, 0.131191800690904, 0.131180193249265, 0.131168598978967, 0.131157017951246, 0.131145450235193, 0.131133895902052, 0.131122355019241, 0.131110827656721, 0.131099313882321, 0.131087813763078, 0.131076327365083, 0.131064854756367, 0.131053396001289, 0.131041951165001, 0.131030520312287, 0.131019103506453, 0.131007700811068, 0.130996312288305, 0.130984937999095, 0.130973578006336, 0.130962232369353, 0.130950901148372, 0.130939584402676, 0.130928282189309, 0.130916994567381, 0.130905721592486, 0.130894463322375, 0.130883219811553, 0.130871991114723, 0.130860777285908, 0.130849578378913, 0.130838394445395, 0.130827225536286, 0.130816071704391, 0.130804932998718, 0.130793809467938, 0.130782701161193, 0.130771608125156, 0.130760530407983, 0.130749468054820, 0.130738421110162, 0.130727389619309, 0.130716373624305, 0.130705373168772, 0.130694388293114, 0.130683419039316, 0.130672465445474, 0.130661527551674, 0.130650605395490, 0.130639699013484, 0.130628808441553, 0.130617933716269, 0.130607074870084, 0.130596231936699, 0.130585404948425, 0.130574593936557, 0.130563798930519, 0.130553019960514, 0.130542257054629, 0.130531510238626, 0.130520779540834, 0.130510064984516, 0.130499366594584, 0.130488684394658, 0.130478018405875, 0.130467368648981, 0.130456735144698, 0.130446117911545, 0.130435516965959, 0.130424932326317, 0.130414364007066, 0.130403812022899, 0.130393276387066, 0.130382757111191, 0.130372254206313, 0.130361767683089, 0.130351297549404, 0.130340843813342, 0.130330406481233, 0.130319985556987, 0.130309581046915, 0.130299192951610, 0.130288821274201, 0.130278466013595, 0.130268127170708, 0.130257804741871, 0.130247498725558, 0.130237209116034, 0.130226935907827, 0.130216679094249, 0.130206438666744, 0.130196214616597, 0.130186006931996, 0.130175815600934, 0.130165640611033, 0.130155481947343, 0.130145339593928, 0.130135213532765, 0.130125103746803, 0.130115010215137, 0.130104932917421, 0.130094871829540, 0.130084826930352, 0.130074798192751, 0.130064785590317, 0.130054789096623, 0.130044808680799, 0.130034844313101, 0.130024895961753, 0.130014963592667, 0.130005047172532, 0.129995146662809, 0.129985262028698, 0.129975393229601, 0.129965540226182, 0.129955702975998, 0.129945881436190, 0.129936075562961, 0.129926285309043, 0.129916510628263, 0.129906751471330, 0.129897007786568, 0.129887279524566, 0.129877566630777, 0.129867869052141, 0.129858186729612, 0.129848519607349, 0.129838867627319, 0.129829230727748, 0.129819608847147, 0.129810001921330, 0.129800409885496, 0.129790832674293, 0.129781270218611, 0.129771722449786, 0.129762189296463, 0.129752670686349, 0.129743166545086, 0.129733676796928, 0.129724201366917, 0.129714740174228, 0.129705293140233, 0.129695860182478, 0.129686441217773, 0.129677036162548, 0.129667644930001, 0.129658267432773, 0.129648903580751, 0.129639553284055, 0.129630216450958, 0.129620892985669, 0.129611582794737, 0.129602285779215, 0.129593001843040, 0.129583730884059, 0.129574472801929, 0.129565227492283, 0.129555994851929, 0.129546774772291, 0.129537567148738, 0.129528371869318, 0.129519188823154, 0.129510017899281, 0.129500858982465, 0.129491711956833, 0.129482576705957, 0.129473453109711, 0.129464341050097, 0.129455240403534, 0.129446151046584, 0.129437072854079, 0.129428005699461, 0.129418949454901, 0.129409903991434, 0.129400869176247, 0.129391844877549, 0.129382830959740, 0.129373827289193, 0.129364833724535, 0.129355850130227, 0.129346876364158, 0.129337912284805, 0.129328957746800, 0.129320012606533, 0.129311076716259, 0.129302149928620, 0.129293232093634, 0.129284323057777, 0.129275422670792, 0.129266530776862, 0.129257647219675, 0.129248771842269, 0.129239904486425, 0.129231044988953, 0.129222193190473, 0.129213348925180, 0.129204512030176, 0.129195682336516, 0.129186859678947, 0.129178043885104, 0.129169234785576, 0.129160432207535, 0.129151635976702, 0.129142845917275, 0.129134061852828, 0.129125283605058, 0.129116510994550, 0.129107743839107, 0.129098981956261, 0.129090225162653, 0.129081473271425, 0.129072726095642, 0.129063983448355, 0.129055245139236, 0.129046510976131, 0.129037780767899, 0.129029054319465, 0.129020331436028, 0.129011611921459, 0.129002895576617, 0.128994182203118, 0.128985471599796, 0.128976763565398, 0.128968057896139, 0.128959354385629, 0.128950652831426, 0.128941953023606, 0.128933254755660, 0.128924557816285, 0.128915861995709, 0.128907167081070, 0.128898472858940, 0.128889779114473, 0.128881085633082, 0.128872392196621, 0.128863698587595, 0.128855004585797, 0.128846309971311, 0.128837614521712, 0.128828918015487, 0.128820220228074, 0.128811520934160, 0.128802819908622, 0.128794116923444, 0.128785411750757, 0.128776704160620, 0.128767993924108, 0.128759280809214, 0.128750564581997, 0.128741845012435, 0.128733121865070, 0.128724394903384, 0.128715663893289, 0.128706928596769, 0.128698188775460, 0.128689444190036, 0.128680694603947, 0.128671939773147, 0.128663179459363, 0.128654413417107, 0.128645641406930, 0.128636863183819, 0.128628078501919, 0.128619287117992, 0.128610488784843, 0.128601683257355, 0.128592870286943, 0.128584049627395, 0.128575221027723, 0.128566384240943, 0.128557539018347, 0.128548685105518, 0.128539822255574, 0.128530950216846, 0.128522068735027, 0.128513177560180, 0.128504276439381, 0.128495365119236, 0.128486443345635, 0.128477510864389, 0.128468567423241, 0.128459612766413, 0.128450646638679, 0.128441668784324, 0.128432678950721, 0.128423676879532, 0.128414662316881, 0.128405635005706, 0.128396594689716, 0.128387541112830, 0.128378474019173, 0.128369393153263, 0.128360298256683, 0.128351189075029, 0.128342065350585, 0.128332926827209, 0.128323773250929, 0.128314604362563, 0.128305419907756, 0.128296219631576, 0.128287003277439, 0.128277770588679, 0.128268521315392, 0.128259255196428, 0.128249971983147, 0.128240671417149, 0.128231353249350, 0.128222017222032, 0.128212663087224, 0.128203290589528, 0.128193899479002, 0.128184489504174, 0.128175060415598, 0.128165611962539, 0.128156143897672, 0.128146655971563, 0.128137147936086, 0.128127619547227, 0.128118070558662, 0.128108500722777, 0.128098909799340, 0.128089297543529, 0.128079663715281, 0.128070008069209, 0.128060330370746, 0.128050630378558, 0.128040907857134, 0.128031162567484, 0.128021394277080, 0.128011602751194, 0.128001787757991, 0.127991949063788, 0.127982086444669, 0.127972199668424, 0.127962288509773, 0.127952352744500, 0.127942392147328, 0.127932406499450, 0.127922395578555, 0.127912359168306, 0.127902297050858, 0.127892209012728, 0.127882094840808, 0.127871954324760, 0.127861787255310, 0.127851593427394, 0.127841372636119, 0.127831124678239, 0.127820849354708, 0.127810546466576, 0.127800215818560, 0.127789857219717, 0.127779470476281, 0.127769055401653, 0.127758611810520, 0.127748139518235, 0.127737638345443, 0.127727108115272, 0.127716548649962, 0.127705959780394, 0.127695341333624, 0.127684693146130, 0.127674015053723, 0.127663306896721, 0.127652568515405, 0.127641799758855, 0.127631000472693, 0.127620170512090, 0.127609309732312, 0.127598417992289, 0.127587495153910, 0.127576541084815, 0.127565555652612, 0.127554538733036, 0.127543490201951, 0.127532409940221, 0.127521297832303, 0.127510153767973, 0.127498977639442, 0.127487769344838, 0.127476528782689, 0.127465255859379, 0.127453950485079, 0.127442612571985, 0.127431242040073, 0.127419838810523, 0.127408402809931, 0.127396933973151, 0.127385432233167, 0.127373897534786, 0.127362329819533, 0.127350729043063, 0.127339095160129, 0.127327428131472, 0.127315727922378, 0.127303994505726, 0.127292227857918, 0.127280427960338, 0.127268594802764, 0.127256728376433, 0.127244828682064, 0.127232895722113, 0.127220929510058, 0.127208930059089, 0.127196897392844, 0.127184831537878, 0.127172732532019, 0.127160600411380, 0.127148435226839, 0.127136237028426, 0.127124005878430, 0.127111741840839, 0.127099444990424, 0.127087115405556, 0.127074753172485, 0.127062358385391, 0.127049931144892, 0.127037471556768, 0.127024979738507, 0.127012455807519, 0.126999899896770, 0.126987312141237, 0.126974692685981, 0.126962041683448, 0.126949359292213, 0.126936645680132, 0.126923901021626, 0.126911125501038, 0.126898319311140, 0.126885482648013, 0.126872615726100, 0.126859718757681, 0.126846791966787, 0.126833835590632, 0.126820849870860, 0.126807835056972, 0.126794791411782, 0.126781719203531, 0.126768618710793, 0.126755490223587, 0.126742334036006, 0.126729150456165, 0.126715939799382, 0.126702702395591, 0.126689438577061, 0.126676148690817, 0.126662833091844, 0.126649492148348, 0.126636126235197, 0.126622735740838, 0.126609321057957, 0.126595882599780, 0.126582420781718, 0.126568936035662, 0.126555428799716, 0.126541899527381, 0.126528348680391, 0.126514776731809, 0.126501184171463, 0.126487571491628, 0.126473939202949, 0.126460287826998, 0.126446617897140, 0.126432929954052, 0.126419224560160, 0.126405502281986, 0.126391763701595, 0.126378009412463, 0.126364240021380, 0.126350456151692, 0.126336658434656, 0.126322847514531, 0.126309024052157, 0.126295188721517, 0.126281342207333, 0.126267485210540, 0.126253618443580, 0.126239742634233, 0.126225858526628, 0.126211966874011, 0.126198068449470, 0.126184164033277, 0.126170254430295, 0.126156340452629, 0.126142422928979, 0.126128502705697, 0.126114580638678, 0.126100657607797, 0.126086734499860, 0.126072812219287, 0.126058891695965, 0.126044973858889, 0.126031059667098, 0.126017150090769, 0.126003246114962, 0.125989348746967, 0.125975458999724, 0.125961577917361, 0.125947706551406, 0.125933845970955, 0.125919997268869, 0.125906161548877, 0.125892339936322, 0.125878533571092, 0.125864743614973, 0.125850971243883, 0.125837217657589, 0.125823484067118, 0.125809771712038, 0.125796081839226, 0.125782415720991, 0.125768774651588, 0.125755159936539, 0.125741572907573, 0.125728014916965, 0.125714487331249, 0.125700991538480, 0.125687528950296, 0.125674100998235, 0.125660709131481, 0.125647354820427, 0.125634039559772, 0.125620764860096, 0.125607532260506, 0.125594343313818, 0.125581199602538, 0.125568102722824, 0.125555054298437, 0.125542055973160, 0.125529109415968, 0.125516216314344, 0.125503378381925, 0.125490597355598, 0.125477874991108, 0.125465213072945, 0.125452613405201, 0.125440077819810, 0.125427608168927, 0.125415206330183, 0.125402874210336, 0.125390613730058, 0.125378426842970, 0.125366315529274, 0.125354281789401, 0.125342327650506, 0.125330455165833, 0.125318666413699, 0.125306963501570, 0.125295348561227, 0.125283823748019, 0.125272391250026, 0.125261053275882, 0.125249812066817, 0.125238669888250, 0.125227629035505, 0.125216691826026, 0.125205860616830, 0.125195137779055, 0.125184525723375, 0.125174026885249, 0.125163643725583, 0.125153378738909, 0.125143234452190, 0.125133213412741, 0.125123318204354, 0.125113551442862, 0.125103915765242, 0.125094413847945, 0.125085048395678, 0.125075822142895, 0.125066737855747, 0.125057798331432, 0.125049006401107, 0.125040364926022, 0.125031876801643, 0.125023544953716, 0.125015372338770, 0.125007361952038, 0.124999516821261, 0.124991840000126, 0.124984334585999, 0.124977003701659, 0.124969850511347, 0.124962878207329, 0.124956090024772, 0.124949489226229, 0.124943079112231, 0.124936863021244, 0.124930844324757, 0.124925026426516, 0.124919412776373, 0.124914006856639, 0.124908812180090, 0.124903832306034, 0.124899070825686, 0.124894531370528, 0.124890217606248, 0.124886133243828, 0.124882282026747, 0.124878667739568, 0.124875294202356, 0.124872165282965, 0.124869284879574, 0.124866656934361, 0.124864285435182, 0.124862174400173, 0.124860327895058, 0.124858750023470, 0.124857444933473, 0.124856416812976, 0.124855669890824, 0.124855208440916, 0.124855036778298, 0.124855159259081, 0.124855580289742, 0.124856304307719, 0.124857335804673, 0.124858679316381, 0.124860339417196, 0.124862320726486, 0.124864627915626, 0.124867265693334, 0.124870238820899, 0.124873552099033, 0.124877210380900, 0.124881218563012, 0.124885581584523, 0.124890304441455, 0.124895392173630, 0.124900849863926, 0.124906682646139, 0.124912895709648, 0.124919494281838, 0.124926483647292, 0.124933869136373, 0.124941656130801, 0.124949850058394, 0.124958456403736, 0.124967480703182, 0.124976928534167]

    freqsPerfEEG = np.array(freqsPerfEEG)
    logPowerPerfEEG = np.array(logPowerPerfEEG)

    return freqsPerfEEG, logPowerPerfEEG

def get_perf_iEEG_chann_avg_regression_vals():

    freqsPerf_iEEG = [1, 1.100, 1.200, 1.300, 1.400, 1.500, 1.600, 1.700, 1.800, 1.900, 2, 2.100, 2.200, 2.300, 2.400, 2.500, 2.600, 2.700, 2.800, 2.900, 3, 3.100, 3.200, 3.300, 3.400, 3.500, 3.600, 3.700, 3.800, 3.900, 4, 4.100, 4.200, 4.300, 4.400, 4.500, 4.600, 4.700, 4.800, 4.900, 5, 5.100, 5.200, 5.300, 5.400, 5.500, 5.600, 5.700, 5.800, 5.900, 6, 6.100, 6.200, 6.300, 6.400, 6.500, 6.600, 6.700, 6.800, 6.900, 7, 7.100, 7.200, 7.300, 7.400, 7.500, 7.600, 7.700, 7.800, 7.900, 8, 8.100, 8.200, 8.300, 8.400, 8.500, 8.600, 8.700, 8.800, 8.900, 9, 9.100, 9.200, 9.300, 9.400, 9.500, 9.600, 9.700, 9.800, 9.900, 10, 10.10, 10.20, 10.30, 10.40, 10.50, 10.60, 10.70, 10.80, 10.90, 11, 11.10, 11.20, 11.30, 11.40, 11.50, 11.60, 11.70, 11.80, 11.90, 12, 12.10, 12.20, 12.30, 12.40, 12.50, 12.60, 12.70, 12.80, 12.90, 13, 13.10, 13.20, 13.30, 13.40, 13.50, 13.60, 13.70, 13.80, 13.90, 14, 14.10, 14.20, 14.30, 14.40, 14.50, 14.60, 14.70, 14.80, 14.90, 15, 15.10, 15.20, 15.30, 15.40, 15.50, 15.60, 15.70, 15.80, 15.90, 16, 16.10, 16.20, 16.30, 16.40, 16.50, 16.60, 16.70, 16.80, 16.90, 17, 17.10, 17.20, 17.30, 17.40, 17.50, 17.60, 17.70, 17.80, 17.90, 18, 18.10, 18.20, 18.30, 18.40, 18.50, 18.60, 18.70, 18.80, 18.90, 19, 19.10, 19.20, 19.30, 19.40, 19.50, 19.60, 19.70, 19.80, 19.90, 20, 20.10, 20.20, 20.30, 20.40, 20.50, 20.60, 20.70, 20.80, 20.90, 21, 21.10, 21.20, 21.30, 21.40, 21.50, 21.60, 21.70, 21.80, 21.90, 22, 22.10, 22.20, 22.30, 22.40, 22.50, 22.60, 22.70, 22.80, 22.90, 23, 23.10, 23.20, 23.30, 23.40, 23.50, 23.60, 23.70, 23.80, 23.90, 24, 24.10, 24.20, 24.30, 24.40, 24.50, 24.60, 24.70, 24.80, 24.90, 25, 25.10, 25.20, 25.30, 25.40, 25.50, 25.60, 25.70, 25.80, 25.90, 26, 26.10, 26.20, 26.30, 26.40, 26.50, 26.60, 26.70, 26.80, 26.90, 27, 27.10, 27.20, 27.30, 27.40, 27.50, 27.60, 27.70, 27.80, 27.90, 28, 28.10, 28.20, 28.30, 28.40, 28.50, 28.60, 28.70, 28.80, 28.90, 29, 29.10, 29.20, 29.30, 29.40, 29.50, 29.60, 29.70, 29.80, 29.90, 30, 30.10, 30.20, 30.30, 30.40, 30.50, 30.60, 30.70, 30.80, 30.90, 31, 31.10, 31.20, 31.30, 31.40, 31.50, 31.60, 31.70, 31.80, 31.90, 32, 32.10, 32.20, 32.30, 32.40, 32.50, 32.60, 32.70, 32.80, 32.90, 33, 33.10, 33.20, 33.30, 33.40, 33.50, 33.60, 33.70, 33.80, 33.90, 34, 34.10, 34.20, 34.30, 34.40, 34.50, 34.60, 34.70, 34.80, 34.90, 35, 35.10, 35.20, 35.30, 35.40, 35.50, 35.60, 35.70, 35.80, 35.90, 36, 36.10, 36.20, 36.30, 36.40, 36.50, 36.60, 36.70, 36.80, 36.90, 37, 37.10, 37.20, 37.30, 37.40, 37.50, 37.60, 37.70, 37.80, 37.90, 38, 38.10, 38.20, 38.30, 38.40, 38.50, 38.60, 38.70, 38.80, 38.90, 39, 39.10, 39.20, 39.30, 39.40, 39.50, 39.60, 39.70, 39.80, 39.90, 40, 40.10, 40.20, 40.30, 40.40, 40.50, 40.60, 40.70, 40.80, 40.90, 41, 41.10, 41.20, 41.30, 41.40, 41.50, 41.60, 41.70, 41.80, 41.90, 42, 42.10, 42.20, 42.30, 42.40, 42.50, 42.60, 42.70, 42.80, 42.90, 43, 43.10, 43.20, 43.30, 43.40, 43.50, 43.60, 43.70, 43.80, 43.90, 44, 44.10, 44.20, 44.30, 44.40, 44.50, 44.60, 44.70, 44.80, 44.90, 45, 45.10, 45.20, 45.30, 45.40, 45.50, 45.60, 45.70, 45.80, 45.90, 46, 46.10, 46.20, 46.30, 46.40, 46.50, 46.60, 46.70, 46.80, 46.90, 47, 47.10, 47.20, 47.30, 47.40, 47.50, 47.60, 47.70, 47.80, 47.90, 48, 48.10, 48.20, 48.30, 48.40, 48.50, 48.60, 48.70, 48.80, 48.90, 49, 49.10, 49.20, 49.30, 49.40, 49.50, 49.60, 49.70, 49.80, 49.90, 50, 50.10, 50.20, 50.30, 50.40, 50.50, 50.60, 50.70, 50.80, 50.90, 51, 51.10, 51.20, 51.30, 51.40, 51.50, 51.60, 51.70, 51.80, 51.90, 52, 52.10, 52.20, 52.30, 52.40, 52.50, 52.60, 52.70, 52.80, 52.90, 53, 53.10, 53.20, 53.30, 53.40, 53.50, 53.60, 53.70, 53.80, 53.90, 54, 54.10, 54.20, 54.30, 54.40, 54.50, 54.60, 54.70, 54.80, 54.90, 55, 55.10, 55.20, 55.30, 55.40, 55.50, 55.60, 55.70, 55.80, 55.90, 56, 56.10, 56.20, 56.30, 56.40, 56.50, 56.60, 56.70, 56.80, 56.90, 57, 57.10, 57.20, 57.30, 57.40, 57.50, 57.60, 57.70, 57.80, 57.90, 58, 58.10, 58.20, 58.30, 58.40, 58.50, 58.60, 58.70, 58.80, 58.90, 59, 59.10, 59.20, 59.30, 59.40, 59.50, 59.60, 59.70, 59.80, 59.90, 60, 60.10, 60.20, 60.30, 60.40, 60.50, 60.60, 60.70, 60.80, 60.90, 61, 61.10, 61.20, 61.30, 61.40, 61.50, 61.60, 61.70, 61.80, 61.90, 62, 62.10, 62.20, 62.30, 62.40, 62.50, 62.60, 62.70, 62.80, 62.90, 63, 63.10, 63.20, 63.30, 63.40, 63.50, 63.60, 63.70, 63.80, 63.90, 64, 64.10, 64.20, 64.30, 64.40, 64.50, 64.60, 64.70, 64.80, 64.90, 65, 65.10, 65.20, 65.30, 65.40, 65.50, 65.60, 65.70, 65.80, 65.90, 66, 66.10, 66.20, 66.30, 66.40, 66.50, 66.60, 66.70, 66.80, 66.90, 67, 67.10, 67.20, 67.30, 67.40, 67.50, 67.60, 67.70, 67.80, 67.90, 68, 68.10, 68.20, 68.30, 68.40, 68.50, 68.60, 68.70, 68.80, 68.90, 69, 69.10, 69.20, 69.30, 69.40, 69.50, 69.60, 69.70, 69.80, 69.90, 70, 70.10, 70.20, 70.30, 70.40, 70.50, 70.60, 70.70, 70.80, 70.90, 71, 71.10, 71.20, 71.30, 71.40, 71.50, 71.60, 71.70, 71.80, 71.90, 72, 72.10, 72.20, 72.30, 72.40, 72.50, 72.60, 72.70, 72.80, 72.90, 73, 73.10, 73.20, 73.30, 73.40, 73.50, 73.60, 73.70, 73.80, 73.90, 74, 74.10, 74.20, 74.30, 74.40, 74.50, 74.60, 74.70, 74.80, 74.90, 75, 75.10, 75.20, 75.30, 75.40, 75.50, 75.60, 75.70, 75.80, 75.90, 76, 76.10, 76.20, 76.30, 76.40, 76.50, 76.60, 76.70, 76.80, 76.90, 77, 77.10, 77.20, 77.30, 77.40, 77.50, 77.60, 77.70, 77.80, 77.90, 78, 78.10, 78.20, 78.30, 78.40, 78.50, 78.60, 78.70, 78.80, 78.90, 79, 79.10, 79.20, 79.30, 79.40, 79.50, 79.60, 79.70, 79.80, 79.90, 80, 80.10, 80.20, 80.30, 80.40, 80.50, 80.60, 80.70, 80.80, 80.90, 81, 81.10, 81.20, 81.30, 81.40, 81.50, 81.60, 81.70, 81.80, 81.90, 82, 82.10, 82.20, 82.30, 82.40, 82.50, 82.60, 82.70, 82.80, 82.90, 83, 83.10, 83.20, 83.30, 83.40, 83.50, 83.60, 83.70, 83.80, 83.90, 84, 84.10, 84.20, 84.30, 84.40, 84.50, 84.60, 84.70, 84.80, 84.90, 85, 85.10, 85.20, 85.30, 85.40, 85.50, 85.60, 85.70, 85.80, 85.90, 86, 86.10, 86.20, 86.30, 86.40, 86.50, 86.60, 86.70, 86.80, 86.90, 87, 87.10, 87.20, 87.30, 87.40, 87.50, 87.60, 87.70, 87.80, 87.90, 88, 88.10, 88.20, 88.30, 88.40, 88.50, 88.60, 88.70, 88.80, 88.90, 89, 89.10, 89.20, 89.30, 89.40, 89.50, 89.60, 89.70, 89.80, 89.90, 90, 90.10, 90.20, 90.30, 90.40, 90.50, 90.60, 90.70, 90.80, 90.90, 91, 91.10, 91.20, 91.30, 91.40, 91.50, 91.60, 91.70, 91.80, 91.90, 92, 92.10, 92.20, 92.30, 92.40, 92.50, 92.60, 92.70, 92.80, 92.90, 93, 93.10, 93.20, 93.30, 93.40, 93.50, 93.60, 93.70, 93.80, 93.90, 94, 94.10, 94.20, 94.30, 94.40, 94.50, 94.60, 94.70, 94.80, 94.90, 95, 95.10, 95.20, 95.30, 95.40, 95.50, 95.60, 95.70, 95.80, 95.90, 96, 96.10, 96.20, 96.30, 96.40, 96.50, 96.60, 96.70, 96.80, 96.90, 97, 97.10, 97.20, 97.30, 97.40, 97.50, 97.60, 97.70, 97.80, 97.90, 98, 98.10, 98.20, 98.30, 98.40, 98.50, 98.60, 98.70, 98.80, 98.90, 99, 99.10, 99.20, 99.30, 99.40, 99.50, 99.60, 99.70, 99.80, 99.90, 100, 100.1, 100.2, 100.3, 100.4, 100.5, 100.6, 100.7, 100.8, 100.9, 101, 101.1, 101.2, 101.3, 101.4, 101.5, 101.6, 101.7, 101.8, 101.9, 102, 102.1, 102.2, 102.3, 102.4, 102.5, 102.6, 102.7, 102.8, 102.9, 103, 103.1, 103.2, 103.3, 103.4, 103.5, 103.6, 103.7, 103.8, 103.9, 104, 104.1, 104.2, 104.3, 104.4, 104.5, 104.6, 104.7, 104.8, 104.9, 105, 105.1, 105.2, 105.3, 105.4, 105.5, 105.6, 105.7, 105.8, 105.9, 106, 106.1, 106.2, 106.3, 106.4, 106.5, 106.6, 106.7, 106.8, 106.9, 107, 107.1, 107.2, 107.3, 107.4, 107.5, 107.6, 107.7, 107.8, 107.9, 108, 108.1, 108.2, 108.3, 108.4, 108.5, 108.6, 108.7, 108.8, 108.9, 109, 109.1, 109.2, 109.3, 109.4, 109.5, 109.6, 109.7, 109.8, 109.9, 110, 110.1, 110.2, 110.3, 110.4, 110.5, 110.6, 110.7, 110.8, 110.9, 111, 111.1, 111.2, 111.3, 111.4, 111.5, 111.6, 111.7, 111.8, 111.9, 112, 112.1, 112.2, 112.3, 112.4, 112.5, 112.6, 112.7, 112.8, 112.9, 113, 113.1, 113.2, 113.3, 113.4, 113.5, 113.6, 113.7, 113.8, 113.9, 114, 114.1, 114.2, 114.3, 114.4, 114.5, 114.6, 114.7, 114.8, 114.9, 115, 115.1, 115.2, 115.3, 115.4, 115.5, 115.6, 115.7, 115.8, 115.9, 116, 116.1, 116.2, 116.3, 116.4, 116.5, 116.6, 116.7, 116.8, 116.9, 117, 117.1, 117.2, 117.3, 117.4, 117.5, 117.6, 117.7, 117.8, 117.9, 118, 118.1, 118.2, 118.3, 118.4, 118.5, 118.6, 118.7, 118.8, 118.9, 119, 119.1, 119.2, 119.3, 119.4, 119.5, 119.6, 119.7, 119.8, 119.9, 120, 120.1, 120.2, 120.3, 120.4, 120.5, 120.6, 120.7, 120.8, 120.9, 121, 121.1, 121.2, 121.3, 121.4, 121.5, 121.6, 121.7, 121.8, 121.9, 122, 122.1, 122.2, 122.3, 122.4, 122.5, 122.6, 122.7, 122.8, 122.9, 123, 123.1, 123.2, 123.3, 123.4, 123.5, 123.6, 123.7, 123.8, 123.9, 124, 124.1, 124.2, 124.3, 124.4, 124.5, 124.6, 124.7, 124.8, 124.9, 125, 125.1, 125.2, 125.3, 125.4, 125.5, 125.6, 125.7, 125.8, 125.9, 126, 126.1, 126.2, 126.3, 126.4, 126.5, 126.6, 126.7, 126.8, 126.9, 127, 127.1, 127.2, 127.3, 127.4, 127.5, 127.6, 127.7, 127.8, 127.9, 128, 128.1, 128.2, 128.3, 128.4, 128.5, 128.6, 128.7, 128.8, 128.9, 129, 129.1, 129.2, 129.3, 129.4, 129.5, 129.6, 129.7, 129.8, 129.9, 130, 130.1, 130.2, 130.3, 130.4, 130.5, 130.6, 130.7, 130.8, 130.9, 131, 131.1, 131.2, 131.3, 131.4, 131.5, 131.6, 131.7, 131.8, 131.9, 132, 132.1, 132.2, 132.3, 132.4, 132.5, 132.6, 132.7, 132.8, 132.9, 133, 133.1, 133.2, 133.3, 133.4, 133.5, 133.6, 133.7, 133.8, 133.9, 134, 134.1, 134.2, 134.3, 134.4, 134.5, 134.6, 134.7, 134.8, 134.9, 135, 135.1, 135.2, 135.3, 135.4, 135.5, 135.6, 135.7, 135.8, 135.9, 136, 136.1, 136.2, 136.3, 136.4, 136.5, 136.6, 136.7, 136.8, 136.9, 137, 137.1, 137.2, 137.3, 137.4, 137.5, 137.6, 137.7, 137.8, 137.9, 138, 138.1, 138.2, 138.3, 138.4, 138.5, 138.6, 138.7, 138.8, 138.9, 139, 139.1, 139.2, 139.3, 139.4, 139.5, 139.6, 139.7, 139.8, 139.9, 140, 140.1, 140.2, 140.3, 140.4, 140.5, 140.6, 140.7, 140.8, 140.9, 141, 141.1, 141.2, 141.3, 141.4, 141.5, 141.6, 141.7, 141.8, 141.9, 142, 142.1, 142.2, 142.3, 142.4, 142.5, 142.6, 142.7, 142.8, 142.9, 143, 143.1, 143.2, 143.3, 143.4, 143.5, 143.6, 143.7, 143.8, 143.9, 144, 144.1, 144.2, 144.3, 144.4, 144.5, 144.6, 144.7, 144.8, 144.9, 145, 145.1, 145.2, 145.3, 145.4, 145.5, 145.6, 145.7, 145.8, 145.9, 146, 146.1, 146.2, 146.3, 146.4, 146.5, 146.6, 146.7, 146.8, 146.9, 147, 147.1, 147.2, 147.3, 147.4, 147.5, 147.6, 147.7, 147.8, 147.9, 148, 148.1, 148.2, 148.3, 148.4, 148.5, 148.6, 148.7, 148.8, 148.9, 149, 149.1, 149.2, 149.3, 149.4, 149.5, 149.6, 149.7, 149.8, 149.9, 150, 150.1, 150.2, 150.3, 150.4, 150.5, 150.6, 150.7, 150.8, 150.9, 151, 151.1, 151.2, 151.3, 151.4, 151.5, 151.6, 151.7, 151.8, 151.9, 152, 152.1, 152.2, 152.3, 152.4, 152.5, 152.6, 152.7, 152.8, 152.9, 153, 153.1, 153.2, 153.3, 153.4, 153.5, 153.6, 153.7, 153.8, 153.9, 154, 154.1, 154.2, 154.3, 154.4, 154.5, 154.6, 154.7, 154.8, 154.9, 155, 155.1, 155.2, 155.3, 155.4, 155.5, 155.6, 155.7, 155.8, 155.9, 156, 156.1, 156.2, 156.3, 156.4, 156.5, 156.6, 156.7, 156.8, 156.9, 157, 157.1, 157.2, 157.3, 157.4, 157.5, 157.6, 157.7, 157.8, 157.9, 158, 158.1, 158.2, 158.3, 158.4, 158.5, 158.6, 158.7, 158.8, 158.9, 159, 159.1, 159.2, 159.3, 159.4, 159.5, 159.6, 159.7, 159.8, 159.9, 160, 160.1, 160.2, 160.3, 160.4, 160.5, 160.6, 160.7, 160.8, 160.9, 161, 161.1, 161.2, 161.3, 161.4, 161.5, 161.6, 161.7, 161.8, 161.9, 162, 162.1, 162.2, 162.3, 162.4, 162.5, 162.6, 162.7, 162.8, 162.9, 163, 163.1, 163.2, 163.3, 163.4, 163.5, 163.6, 163.7, 163.8, 163.9, 164, 164.1, 164.2, 164.3, 164.4, 164.5, 164.6, 164.7, 164.8, 164.9, 165, 165.1, 165.2, 165.3, 165.4, 165.5, 165.6, 165.7, 165.8, 165.9, 166, 166.1, 166.2, 166.3, 166.4, 166.5, 166.6, 166.7, 166.8, 166.9, 167, 167.1, 167.2, 167.3, 167.4, 167.5, 167.6, 167.7, 167.8, 167.9, 168, 168.1, 168.2, 168.3, 168.4, 168.5, 168.6, 168.7, 168.8, 168.9, 169, 169.1, 169.2, 169.3, 169.4, 169.5, 169.6, 169.7, 169.8, 169.9, 170, 170.1, 170.2, 170.3, 170.4, 170.5, 170.6, 170.7, 170.8, 170.9, 171, 171.1, 171.2, 171.3, 171.4, 171.5, 171.6, 171.7, 171.8, 171.9, 172, 172.1, 172.2, 172.3, 172.4, 172.5, 172.6, 172.7, 172.8, 172.9, 173, 173.1, 173.2, 173.3, 173.4, 173.5, 173.6, 173.7, 173.8, 173.9, 174, 174.1, 174.2, 174.3, 174.4, 174.5, 174.6, 174.7, 174.8, 174.9, 175, 175.1, 175.2, 175.3, 175.4, 175.5, 175.6, 175.7, 175.8, 175.9, 176, 176.1, 176.2, 176.3, 176.4, 176.5, 176.6, 176.7, 176.8, 176.9, 177, 177.1, 177.2, 177.3, 177.4, 177.5, 177.6, 177.7, 177.8, 177.9, 178, 178.1, 178.2, 178.3, 178.4, 178.5, 178.6, 178.7, 178.8, 178.9, 179, 179.1, 179.2, 179.3, 179.4, 179.5, 179.6, 179.7, 179.8, 179.9, 180, 180.1, 180.2, 180.3, 180.4, 180.5, 180.6, 180.7, 180.8, 180.9, 181, 181.1, 181.2, 181.3, 181.4, 181.5, 181.6, 181.7, 181.8, 181.9, 182, 182.1, 182.2, 182.3, 182.4, 182.5, 182.6, 182.7, 182.8, 182.9, 183, 183.1, 183.2, 183.3, 183.4, 183.5, 183.6, 183.7, 183.8, 183.9, 184, 184.1, 184.2, 184.3, 184.4, 184.5, 184.6, 184.7, 184.8, 184.9, 185, 185.1, 185.2, 185.3, 185.4, 185.5, 185.6, 185.7, 185.8, 185.9, 186, 186.1, 186.2, 186.3, 186.4, 186.5, 186.6, 186.7, 186.8, 186.9, 187, 187.1, 187.2, 187.3, 187.4, 187.5, 187.6, 187.7, 187.8, 187.9, 188, 188.1, 188.2, 188.3, 188.4, 188.5, 188.6, 188.7, 188.8, 188.9, 189, 189.1, 189.2, 189.3, 189.4, 189.5, 189.6, 189.7, 189.8, 189.9, 190, 190.1, 190.2, 190.3, 190.4, 190.5, 190.6, 190.7, 190.8, 190.9, 191, 191.1, 191.2, 191.3, 191.4, 191.5, 191.6, 191.7, 191.8, 191.9, 192, 192.1, 192.2, 192.3, 192.4, 192.5, 192.6, 192.7, 192.8, 192.9, 193, 193.1, 193.2, 193.3, 193.4, 193.5, 193.6, 193.7, 193.8, 193.9, 194, 194.1, 194.2, 194.3, 194.4, 194.5, 194.6, 194.7, 194.8, 194.9, 195, 195.1, 195.2, 195.3, 195.4, 195.5, 195.6, 195.7, 195.8, 195.9, 196, 196.1, 196.2, 196.3, 196.4, 196.5, 196.6, 196.7, 196.8, 196.9, 197, 197.1, 197.2, 197.3, 197.4, 197.5, 197.6, 197.7, 197.8, 197.9, 198, 198.1, 198.2, 198.3, 198.4, 198.5, 198.6, 198.7, 198.8, 198.9, 199, 199.1, 199.2, 199.3, 199.4, 199.5, 199.6, 199.7, 199.8, 199.9, 200, 200.1, 200.2, 200.3, 200.4, 200.5, 200.6, 200.7, 200.8, 200.9, 201, 201.1, 201.2, 201.3, 201.4, 201.5, 201.6, 201.7, 201.8, 201.9, 202, 202.1, 202.2, 202.3, 202.4, 202.5, 202.6, 202.7, 202.8, 202.9, 203, 203.1, 203.2, 203.3, 203.4, 203.5, 203.6, 203.7, 203.8, 203.9, 204, 204.1, 204.2, 204.3, 204.4, 204.5, 204.6, 204.7, 204.8, 204.9, 205, 205.1, 205.2, 205.3, 205.4, 205.5, 205.6, 205.7, 205.8, 205.9, 206, 206.1, 206.2, 206.3, 206.4, 206.5, 206.6, 206.7, 206.8, 206.9, 207, 207.1, 207.2, 207.3, 207.4, 207.5, 207.6, 207.7, 207.8, 207.9, 208, 208.1, 208.2, 208.3, 208.4, 208.5, 208.6, 208.7, 208.8, 208.9, 209, 209.1, 209.2, 209.3, 209.4, 209.5, 209.6, 209.7, 209.8, 209.9, 210, 210.1, 210.2, 210.3, 210.4, 210.5, 210.6, 210.7, 210.8, 210.9, 211, 211.1, 211.2, 211.3, 211.4, 211.5, 211.6, 211.7, 211.8, 211.9, 212, 212.1, 212.2, 212.3, 212.4, 212.5, 212.6, 212.7, 212.8, 212.9, 213, 213.1, 213.2, 213.3, 213.4, 213.5, 213.6, 213.7, 213.8, 213.9, 214, 214.1, 214.2, 214.3, 214.4, 214.5, 214.6, 214.7, 214.8, 214.9, 215, 215.1, 215.2, 215.3, 215.4, 215.5, 215.6, 215.7, 215.8, 215.9, 216, 216.1, 216.2, 216.3, 216.4, 216.5, 216.6, 216.7, 216.8, 216.9, 217, 217.1, 217.2, 217.3, 217.4, 217.5, 217.6, 217.7, 217.8, 217.9, 218, 218.1, 218.2, 218.3, 218.4, 218.5, 218.6, 218.7, 218.8, 218.9, 219, 219.1, 219.2, 219.3, 219.4, 219.5, 219.6, 219.7, 219.8, 219.9, 220, 220.1, 220.2, 220.3, 220.4, 220.5, 220.6, 220.7, 220.8, 220.9, 221, 221.1, 221.2, 221.3, 221.4, 221.5, 221.6, 221.7, 221.8, 221.9, 222, 222.1, 222.2, 222.3, 222.4, 222.5, 222.6, 222.7, 222.8, 222.9, 223, 223.1, 223.2, 223.3, 223.4, 223.5, 223.6, 223.7, 223.8, 223.9, 224, 224.1, 224.2, 224.3, 224.4, 224.5, 224.6, 224.7, 224.8, 224.9, 225, 225.1, 225.2, 225.3, 225.4, 225.5, 225.6, 225.7, 225.8, 225.9, 226, 226.1, 226.2, 226.3, 226.4, 226.5, 226.6, 226.7, 226.8, 226.9, 227, 227.1, 227.2, 227.3, 227.4, 227.5, 227.6, 227.7, 227.8, 227.9, 228, 228.1, 228.2, 228.3, 228.4, 228.5, 228.6, 228.7, 228.8, 228.9, 229, 229.1, 229.2, 229.3, 229.4, 229.5, 229.6, 229.7, 229.8, 229.9, 230, 230.1, 230.2, 230.3, 230.4, 230.5, 230.6, 230.7, 230.8, 230.9, 231, 231.1, 231.2, 231.3, 231.4, 231.5, 231.6, 231.7, 231.8, 231.9, 232, 232.1, 232.2, 232.3, 232.4, 232.5, 232.6, 232.7, 232.8, 232.9, 233, 233.1, 233.2, 233.3, 233.4, 233.5, 233.6, 233.7, 233.8, 233.9, 234, 234.1, 234.2, 234.3, 234.4, 234.5, 234.6, 234.7, 234.8, 234.9, 235, 235.1, 235.2, 235.3, 235.4, 235.5, 235.6, 235.7, 235.8, 235.9, 236, 236.1, 236.2, 236.3, 236.4, 236.5, 236.6, 236.7, 236.8, 236.9, 237, 237.1, 237.2, 237.3, 237.4, 237.5, 237.6, 237.7, 237.8, 237.9, 238, 238.1, 238.2, 238.3, 238.4, 238.5, 238.6, 238.7, 238.8, 238.9, 239, 239.1, 239.2, 239.3, 239.4, 239.5, 239.6, 239.7, 239.8, 239.9, 240, 240.1, 240.2, 240.3, 240.4, 240.5, 240.6, 240.7, 240.8, 240.9, 241, 241.1, 241.2, 241.3, 241.4, 241.5, 241.6, 241.7, 241.8, 241.9, 242, 242.1, 242.2, 242.3, 242.4, 242.5, 242.6, 242.7, 242.8, 242.9, 243, 243.1, 243.2, 243.3, 243.4, 243.5, 243.6, 243.7, 243.8, 243.9, 244, 244.1, 244.2, 244.3, 244.4, 244.5, 244.6, 244.7, 244.8, 244.9, 245, 245.1, 245.2, 245.3, 245.4, 245.5, 245.6, 245.7, 245.8, 245.9, 246, 246.1, 246.2, 246.3, 246.4, 246.5, 246.6, 246.7, 246.8, 246.9, 247, 247.1, 247.2, 247.3, 247.4, 247.5, 247.6, 247.7, 247.8, 247.9, 248, 248.1, 248.2, 248.3, 248.4, 248.5, 248.6, 248.7, 248.8, 248.9, 249, 249.1, 249.2, 249.3, 249.4, 249.5, 249.6, 249.7, 249.8, 249.9, 250, 250.1, 250.2, 250.3, 250.4, 250.5, 250.6, 250.7, 250.8, 250.9, 251, 251.1, 251.2, 251.3, 251.4, 251.5, 251.6, 251.7, 251.8, 251.9, 252, 252.1, 252.2, 252.3, 252.4, 252.5, 252.6, 252.7, 252.8, 252.9, 253, 253.1, 253.2, 253.3, 253.4, 253.5, 253.6, 253.7, 253.8, 253.9, 254, 254.1, 254.2, 254.3, 254.4, 254.5, 254.6, 254.7, 254.8, 254.9, 255, 255.1, 255.2, 255.3, 255.4, 255.5, 255.6, 255.7, 255.8, 255.9, 256, 256.1, 256.2, 256.3, 256.4, 256.5, 256.6, 256.7, 256.8, 256.9, 257, 257.1, 257.2, 257.3, 257.4, 257.5, 257.6, 257.7, 257.8, 257.9, 258, 258.1, 258.2, 258.3, 258.4, 258.5, 258.6, 258.7, 258.8, 258.9, 259, 259.1, 259.2, 259.3, 259.4, 259.5, 259.6, 259.7, 259.8, 259.9, 260, 260.1, 260.2, 260.3, 260.4, 260.5, 260.6, 260.7, 260.8, 260.9, 261, 261.1, 261.2, 261.3, 261.4, 261.5, 261.6, 261.7, 261.8, 261.9, 262, 262.1, 262.2, 262.3, 262.4, 262.5, 262.6, 262.7, 262.8, 262.9, 263, 263.1, 263.2, 263.3, 263.4, 263.5, 263.6, 263.7, 263.8, 263.9, 264, 264.1, 264.2, 264.3, 264.4, 264.5, 264.6, 264.7, 264.8, 264.9, 265, 265.1, 265.2, 265.3, 265.4, 265.5, 265.6, 265.7, 265.8, 265.9, 266, 266.1, 266.2, 266.3, 266.4, 266.5, 266.6, 266.7, 266.8, 266.9, 267, 267.1, 267.2, 267.3, 267.4, 267.5, 267.6, 267.7, 267.8, 267.9, 268, 268.1, 268.2, 268.3, 268.4, 268.5, 268.6, 268.7, 268.8, 268.9, 269, 269.1, 269.2, 269.3, 269.4, 269.5, 269.6, 269.7, 269.8, 269.9, 270, 270.1, 270.2, 270.3, 270.4, 270.5, 270.6, 270.7, 270.8, 270.9, 271, 271.1, 271.2, 271.3, 271.4, 271.5, 271.6, 271.7, 271.8, 271.9, 272, 272.1, 272.2, 272.3, 272.4, 272.5, 272.6, 272.7, 272.8, 272.9, 273, 273.1, 273.2, 273.3, 273.4, 273.5, 273.6, 273.7, 273.8, 273.9, 274, 274.1, 274.2, 274.3, 274.4, 274.5, 274.6, 274.7, 274.8, 274.9, 275, 275.1, 275.2, 275.3, 275.4, 275.5, 275.6, 275.7, 275.8, 275.9, 276, 276.1, 276.2, 276.3, 276.4, 276.5, 276.6, 276.7, 276.8, 276.9, 277, 277.1, 277.2, 277.3, 277.4, 277.5, 277.6, 277.7, 277.8, 277.9, 278, 278.1, 278.2, 278.3, 278.4, 278.5, 278.6, 278.7, 278.8, 278.9, 279, 279.1, 279.2, 279.3, 279.4, 279.5, 279.6, 279.7, 279.8, 279.9, 280, 280.1, 280.2, 280.3, 280.4, 280.5, 280.6, 280.7, 280.8, 280.9, 281, 281.1, 281.2, 281.3, 281.4, 281.5, 281.6, 281.7, 281.8, 281.9, 282, 282.1, 282.2, 282.3, 282.4, 282.5, 282.6, 282.7, 282.8, 282.9, 283, 283.1, 283.2, 283.3, 283.4, 283.5, 283.6, 283.7, 283.8, 283.9, 284, 284.1, 284.2, 284.3, 284.4, 284.5, 284.6, 284.7, 284.8, 284.9, 285, 285.1, 285.2, 285.3, 285.4, 285.5, 285.6, 285.7, 285.8, 285.9, 286, 286.1, 286.2, 286.3, 286.4, 286.5, 286.6, 286.7, 286.8, 286.9, 287, 287.1, 287.2, 287.3, 287.4, 287.5, 287.6, 287.7, 287.8, 287.9, 288, 288.1, 288.2, 288.3, 288.4, 288.5, 288.6, 288.7, 288.8, 288.9, 289, 289.1, 289.2, 289.3, 289.4, 289.5, 289.6, 289.7, 289.8, 289.9, 290, 290.1, 290.2, 290.3, 290.4, 290.5, 290.6, 290.7, 290.8, 290.9, 291, 291.1, 291.2, 291.3, 291.4, 291.5, 291.6, 291.7, 291.8, 291.9, 292, 292.1, 292.2, 292.3, 292.4, 292.5, 292.6, 292.7, 292.8, 292.9, 293, 293.1, 293.2, 293.3, 293.4, 293.5, 293.6, 293.7, 293.8, 293.9, 294, 294.1, 294.2, 294.3, 294.4, 294.5, 294.6, 294.7, 294.8, 294.9, 295, 295.1, 295.2, 295.3, 295.4, 295.5, 295.6, 295.7, 295.8, 295.9, 296, 296.1, 296.2, 296.3, 296.4, 296.5, 296.6, 296.7, 296.8, 296.9, 297, 297.1, 297.2, 297.3, 297.4, 297.5, 297.6, 297.7, 297.8, 297.9, 298, 298.1, 298.2, 298.3, 298.4, 298.5, 298.6, 298.7, 298.8, 298.9, 299, 299.1, 299.2, 299.3, 299.4, 299.5, 299.6, 299.7, 299.8, 299.9, 300, 300.1, 300.2, 300.3, 300.4, 300.5, 300.6, 300.7, 300.8, 300.9, 301, 301.1, 301.2, 301.3, 301.4, 301.5, 301.6, 301.7, 301.8, 301.9, 302, 302.1, 302.2, 302.3, 302.4, 302.5, 302.6, 302.7, 302.8, 302.9, 303, 303.1, 303.2, 303.3, 303.4, 303.5, 303.6, 303.7, 303.8, 303.9, 304, 304.1, 304.2, 304.3, 304.4, 304.5, 304.6, 304.7, 304.8, 304.9, 305, 305.1, 305.2, 305.3, 305.4, 305.5, 305.6, 305.7, 305.8, 305.9, 306, 306.1, 306.2, 306.3, 306.4, 306.5, 306.6, 306.7, 306.8, 306.9, 307, 307.1, 307.2, 307.3, 307.4, 307.5, 307.6, 307.7, 307.8, 307.9, 308, 308.1, 308.2, 308.3, 308.4, 308.5, 308.6, 308.7, 308.8, 308.9, 309, 309.1, 309.2, 309.3, 309.4, 309.5, 309.6, 309.7, 309.8, 309.9, 310, 310.1, 310.2, 310.3, 310.4, 310.5, 310.6, 310.7, 310.8, 310.9, 311, 311.1, 311.2, 311.3, 311.4, 311.5, 311.6, 311.7, 311.8, 311.9, 312, 312.1, 312.2, 312.3, 312.4, 312.5, 312.6, 312.7, 312.8, 312.9, 313, 313.1, 313.2, 313.3, 313.4, 313.5, 313.6, 313.7, 313.8, 313.9, 314, 314.1, 314.2, 314.3, 314.4, 314.5, 314.6, 314.7, 314.8, 314.9, 315, 315.1, 315.2, 315.3, 315.4, 315.5, 315.6, 315.7, 315.8, 315.9, 316, 316.1, 316.2, 316.3, 316.4, 316.5, 316.6, 316.7, 316.8, 316.9, 317, 317.1, 317.2, 317.3, 317.4, 317.5, 317.6, 317.7, 317.8, 317.9, 318, 318.1, 318.2, 318.3, 318.4, 318.5, 318.6, 318.7, 318.8, 318.9, 319, 319.1, 319.2, 319.3, 319.4, 319.5, 319.6, 319.7, 319.8, 319.9, 320, 320.1, 320.2, 320.3, 320.4, 320.5, 320.6, 320.7, 320.8, 320.9, 321, 321.1, 321.2, 321.3, 321.4, 321.5, 321.6, 321.7, 321.8, 321.9, 322, 322.1, 322.2, 322.3, 322.4, 322.5, 322.6, 322.7, 322.8, 322.9, 323, 323.1, 323.2, 323.3, 323.4, 323.5, 323.6, 323.7, 323.8, 323.9, 324, 324.1, 324.2, 324.3, 324.4, 324.5, 324.6, 324.7, 324.8, 324.9, 325, 325.1, 325.2, 325.3, 325.4, 325.5, 325.6, 325.7, 325.8, 325.9, 326, 326.1, 326.2, 326.3, 326.4, 326.5, 326.6, 326.7, 326.8, 326.9, 327, 327.1, 327.2, 327.3, 327.4, 327.5, 327.6, 327.7, 327.8, 327.9, 328, 328.1, 328.2, 328.3, 328.4, 328.5, 328.6, 328.7, 328.8, 328.9, 329, 329.1, 329.2, 329.3, 329.4, 329.5, 329.6, 329.7, 329.8, 329.9, 330, 330.1, 330.2, 330.3, 330.4, 330.5, 330.6, 330.7, 330.8, 330.9, 331, 331.1, 331.2, 331.3, 331.4, 331.5, 331.6, 331.7, 331.8, 331.9, 332, 332.1, 332.2, 332.3, 332.4, 332.5, 332.6, 332.7, 332.8, 332.9, 333, 333.1, 333.2, 333.3, 333.4, 333.5, 333.6, 333.7, 333.8, 333.9, 334, 334.1, 334.2, 334.3, 334.4, 334.5, 334.6, 334.7, 334.8, 334.9, 335, 335.1, 335.2, 335.3, 335.4, 335.5, 335.6, 335.7, 335.8, 335.9, 336, 336.1, 336.2, 336.3, 336.4, 336.5, 336.6, 336.7, 336.8, 336.9, 337, 337.1, 337.2, 337.3, 337.4, 337.5, 337.6, 337.7, 337.8, 337.9, 338, 338.1, 338.2, 338.3, 338.4, 338.5, 338.6, 338.7, 338.8, 338.9, 339, 339.1, 339.2, 339.3, 339.4, 339.5, 339.6, 339.7, 339.8, 339.9, 340, 340.1, 340.2, 340.3, 340.4, 340.5, 340.6, 340.7, 340.8, 340.9, 341, 341.1, 341.2, 341.3];
    
    
    logPowerPerf_iEEG = [1.000000,	0.996306,	0.992635,	0.988988,	0.985363,	0.981763,	0.978186,	0.974631,	0.971100,	0.967590,	0.964104,	0.960640,	0.957198,	0.953779,	0.950381,	0.947006,	0.943652,	0.940320,	0.937009,	0.933720,	0.930452,	0.927205,	0.923979,	0.920774,	0.917589,	0.914425,	0.911282,	0.908159,	0.905056,	0.901973,	0.898911,	0.895867,	0.892844,	0.889841,	0.886857,	0.883892,	0.880946,	0.878020,	0.875112,	0.872224,	0.869354,	0.866502,	0.863670,	0.860856,	0.858061,	0.855282,	0.852523,	0.849781,	0.847057,	0.844351,	0.841663,	0.838991,	0.836338,	0.833702,	0.831083,	0.828480,	0.825895,	0.823326,	0.820775,	0.818241,	0.815722,	0.813219,	0.810734,	0.808265,	0.805811,	0.803374,	0.800953,	0.798547,	0.796157,	0.793783,	0.791423,	0.789080,	0.786753,	0.784440,	0.782141,	0.779858,	0.777591,	0.775337,	0.773099,	0.770875,	0.768666,	0.766471,	0.764290,	0.762124,	0.759972,	0.757834,	0.755710,	0.753599,	0.751503,	0.749420,	0.747351,	0.745294,	0.743252,	0.741224,	0.739207,	0.737205,	0.735215,	0.733239,	0.731275,	0.729323,	0.727386,	0.725460,	0.723547,	0.721646,	0.719758,	0.717882,	0.716018,	0.714166,	0.712327,	0.710500,	0.708684,	0.706880,	0.705087,	0.703307,	0.701537,	0.699781,	0.698034,	0.696300,	0.694575,	0.692863,	0.691162,	0.689471,	0.687792,	0.686124,	0.684466,	0.682819,	0.681182,	0.679557,	0.677942,	0.676337,	0.674742,	0.673157,	0.671584,	0.670020,	0.668466,	0.666922,	0.665388,	0.663864,	0.662350,	0.660845,	0.659351,	0.657864,	0.656389,	0.654922,	0.653465,	0.652018,	0.650579,	0.649150,	0.647729,	0.646319,	0.644917,	0.643523,	0.642138,	0.640763,	0.639396,	0.638037,	0.636688,	0.635347,	0.634014,	0.632691,	0.631375,	0.630067,	0.628768,	0.627477,	0.626193,	0.624918,	0.623652,	0.622392,	0.621141,	0.619898,	0.618662,	0.617434,	0.616214,	0.615002,	0.613796,	0.612599,	0.611409,	0.610226,	0.609051,	0.607883,	0.606721,	0.605568,	0.604421,	0.603281,	0.602148,	0.601023,	0.599903,	0.598792,	0.597686,	0.596588,	0.595496,	0.594411,	0.593332,	0.592260,	0.591194,	0.590136,	0.589083,	0.588036,	0.586996,	0.585961,	0.584934,	0.583913,	0.582896,	0.581888,	0.580884,	0.579886,	0.578895,	0.577909,	0.576929,	0.575955,	0.574987,	0.574023,	0.573066,	0.572116,	0.571170,	0.570229,	0.569294,	0.568364,	0.567441,	0.566521,	0.565607,	0.564700,	0.563797,	0.562899,	0.562006,	0.561119,	0.560236,	0.559358,	0.558486,	0.557617,	0.556755,	0.555897,	0.555044,	0.554195,	0.553351,	0.552512,	0.551678,	0.550848,	0.550023,	0.549202,	0.548386,	0.547575,	0.546768,	0.545965,	0.545166,	0.544372,	0.543582,	0.542796,	0.542015,	0.541238,	0.540465,	0.539696,	0.538930,	0.538169,	0.537413,	0.536660,	0.535912,	0.535166,	0.534425,	0.533687,	0.532954,	0.532225,	0.531498,	0.530776,	0.530057,	0.529343,	0.528632,	0.527923,	0.527220,	0.526519,	0.525822,	0.525128,	0.524438,	0.523752,	0.523067,	0.522387,	0.521710,	0.521038,	0.520368,	0.519700,	0.519037,	0.518376,	0.517718,	0.517064,	0.516413,	0.515765,	0.515119,	0.514478,	0.513839,	0.513202,	0.512569,	0.511939,	0.511311,	0.510686,	0.510064,	0.509446,	0.508829,	0.508216,	0.507605,	0.506997,	0.506393,	0.505790,	0.505190,	0.504592,	0.503998,	0.503405,	0.502816,	0.502228,	0.501643,	0.501061,	0.500481,	0.499904,	0.499329,	0.498757,	0.498186,	0.497618,	0.497053,	0.496489,	0.495928,	0.495370,	0.494812,	0.494258,	0.493707,	0.493156,	0.492608,	0.492063,	0.491519,	0.490978,	0.490438,	0.489900,	0.489364,	0.488831,	0.488300,	0.487770,	0.487242,	0.486717,	0.486194,	0.485672,	0.485152,	0.484633,	0.484117,	0.483603,	0.483090,	0.482579,	0.482070,	0.481563,	0.481057,	0.480553,	0.480051,	0.479551,	0.479052,	0.478555,	0.478059,	0.477564,	0.477073,	0.476581,	0.476092,	0.475605,	0.475119,	0.474634,	0.474151,	0.473669,	0.473189,	0.472710,	0.472233,	0.471756,	0.471282,	0.470809,	0.470337,	0.469867,	0.469398,	0.468931,	0.468464,	0.467999,	0.467536,	0.467074,	0.466612,	0.466152,	0.465693,	0.465236,	0.464780,	0.464325,	0.463871,	0.463418,	0.462968,	0.462517,	0.462068,	0.461619,	0.461173,	0.460727,	0.460282,	0.459839,	0.459396,	0.458954,	0.458514,	0.458075,	0.457636,	0.457198,	0.456763,	0.456328,	0.455893,	0.455459,	0.455027,	0.454596,	0.454165,	0.453735,	0.453307,	0.452879,	0.452451,	0.452025,	0.451601,	0.451176,	0.450753,	0.450330,	0.449908,	0.449487,	0.449067,	0.448648,	0.448230,	0.447811,	0.447394,	0.446978,	0.446563,	0.446148,	0.445735,	0.445321,	0.444909,	0.444497,	0.444086,	0.443675,	0.443266,	0.442857,	0.442449,	0.442041,	0.441634,	0.441228,	0.440822,	0.440417,	0.440013,	0.439609,	0.439206,	0.438804,	0.438402,	0.438001,	0.437600,	0.437199,	0.436800,	0.436400,	0.436003,	0.435605,	0.435207,	0.434811,	0.434414,	0.434019,	0.433623,	0.433229,	0.432834,	0.432441,	0.432047,	0.431655,	0.431263,	0.430871,	0.430479,	0.430088,	0.429698,	0.429308,	0.428919,	0.428530,	0.428142,	0.427753,	0.427366,	0.426978,	0.426591,	0.426205,	0.425819,	0.425433,	0.425048,	0.424663,	0.424278,	0.423894,	0.423511,	0.423127,	0.422744,	0.422362,	0.421979,	0.421598,	0.421216,	0.420835,	0.420454,	0.420073,	0.419693,	0.419313,	0.418934,	0.418554,	0.418175,	0.417797,	0.417418,	0.417040,	0.416663,	0.416286,	0.415908,	0.415531,	0.415155,	0.414779,	0.414403,	0.414027,	0.413652,	0.413277,	0.412902,	0.412526,	0.412152,	0.411779,	0.411405,	0.411031,	0.410657,	0.410284,	0.409911,	0.409538,	0.409166,	0.408793,	0.408421,	0.408050,	0.407678,	0.407307,	0.406936,	0.406564,	0.406194,	0.405824,	0.405453,	0.405084,	0.404714,	0.404344,	0.403975,	0.403605,	0.403236,	0.402867,	0.402498,	0.402130,	0.401761,	0.401394,	0.401025,	0.400657,	0.400290,	0.399922,	0.399555,	0.399188,	0.398821,	0.398455,	0.398088,	0.397722,	0.397355,	0.396989,	0.396622,	0.396257,	0.395892,	0.395526,	0.395161,	0.394796,	0.394430,	0.394065,	0.393701,	0.393336,	0.392971,	0.392607,	0.392242,	0.391879,	0.391515,	0.391151,	0.390788,	0.390424,	0.390060,	0.389697,	0.389334,	0.388972,	0.388608,	0.388245,	0.387883,	0.387521,	0.387158,	0.386796,	0.386434,	0.386071,	0.385710,	0.385348,	0.384986,	0.384625,	0.384264,	0.383902,	0.383541,	0.383180,	0.382819,	0.382459,	0.382098,	0.381737,	0.381377,	0.381017,	0.380656,	0.380296,	0.379936,	0.379576,	0.379216,	0.378857,	0.378497,	0.378137,	0.377778,	0.377419,	0.377059,	0.376701,	0.376342,	0.375983,	0.375625,	0.375266,	0.374907,	0.374549,	0.374190,	0.373833,	0.373475,	0.373117,	0.372759,	0.372401,	0.372044,	0.371686,	0.371328,	0.370971,	0.370613,	0.370257,	0.369900,	0.369543,	0.369187,	0.368830,	0.368473,	0.368116,	0.367761,	0.367404,	0.367048,	0.366692,	0.366336,	0.365980,	0.365625,	0.365269,	0.364913,	0.364558,	0.364203,	0.363848,	0.363493,	0.363137,	0.362783,	0.362428,	0.362074,	0.361719,	0.361364,	0.361010,	0.360656,	0.360302,	0.359948,	0.359595,	0.359240,	0.358887,	0.358533,	0.358181,	0.357827,	0.357474,	0.357121,	0.356768,	0.356415,	0.356063,	0.355710,	0.355357,	0.355006,	0.354653,	0.354301,	0.353949,	0.353597,	0.353246,	0.352893,	0.352542,	0.352191,	0.351839,	0.351489,	0.351138,	0.350788,	0.350436,	0.350086,	0.349736,	0.349385,	0.349035,	0.348685,	0.348335,	0.347984,	0.347635,	0.347286,	0.346936,	0.346587,	0.346237,	0.345889,	0.345540,	0.345191,	0.344843,	0.344494,	0.344145,	0.343797,	0.343449,	0.343101,	0.342754,	0.342406,	0.342059,	0.341711,	0.341364,	0.341017,	0.340670,	0.340323,	0.339977,	0.339630,	0.339284,	0.338938,	0.338592,	0.338246,	0.337900,	0.337554,	0.337209,	0.336863,	0.336518,	0.336174,	0.335829,	0.335484,	0.335140,	0.334795,	0.334451,	0.334108,	0.333763,	0.333420,	0.333076,	0.332732,	0.332390,	0.332046,	0.331703,	0.331360,	0.331018,	0.330676,	0.330333,	0.329992,	0.329649,	0.329307,	0.328966,	0.328625,	0.328283,	0.327943,	0.327601,	0.327261,	0.326920,	0.326581,	0.326240,	0.325900,	0.325560,	0.325221,	0.324881,	0.324542,	0.324203,	0.323864,	0.323525,	0.323187,	0.322848,	0.322510,	0.322172,	0.321834,	0.321497,	0.321160,	0.320822,	0.320485,	0.320149,	0.319811,	0.319475,	0.319139,	0.318803,	0.318467,	0.318131,	0.317796,	0.317461,	0.317126,	0.316791,	0.316455,	0.316121,	0.315787,	0.315453,	0.315119,	0.314786,	0.314452,	0.314119,	0.313786,	0.313453,	0.313120,	0.312788,	0.312456,	0.312124,	0.311792,	0.311460,	0.311128,	0.310797,	0.310466,	0.310136,	0.309805,	0.309475,	0.309145,	0.308815,	0.308486,	0.308156,	0.307827,	0.307497,	0.307169,	0.306840,	0.306512,	0.306183,	0.305856,	0.305528,	0.305201,	0.304873,	0.304546,	0.304220,	0.303893,	0.303567,	0.303240,	0.302915,	0.302589,	0.302264,	0.301938,	0.301613,	0.301288,	0.300964,	0.300640,	0.300316,	0.299992,	0.299669,	0.299345,	0.299022,	0.298700,	0.298376,	0.298054,	0.297732,	0.297411,	0.297089,	0.296767,	0.296446,	0.296125,	0.295804,	0.295484,	0.295164,	0.294844,	0.294524,	0.294204,	0.293885,	0.293566,	0.293247,	0.292930,	0.292611,	0.292293,	0.291975,	0.291658,	0.291340,	0.291023,	0.290706,	0.290390,	0.290074,	0.289758,	0.289442,	0.289126,	0.288812,	0.288497,	0.288182,	0.287867,	0.287553,	0.287239,	0.286926,	0.286612,	0.286299,	0.285986,	0.285673,	0.285361,	0.285049,	0.284737,	0.284425,	0.284114,	0.283803,	0.283492,	0.283181,	0.282871,	0.282561,	0.282251,	0.281942,	0.281633,	0.281323,	0.281015,	0.280706,	0.280398,	0.280091,	0.279782,	0.279475,	0.279168,	0.278861,	0.278555,	0.278249,	0.277943,	0.277637,	0.277330,	0.277025,	0.276720,	0.276415,	0.276112,	0.275808,	0.275504,	0.275200,	0.274897,	0.274593,	0.274290,	0.273988,	0.273686,	0.273384,	0.273082,	0.272780,	0.272480,	0.272179,	0.271878,	0.271578,	0.271278,	0.270978,	0.270679,	0.270379,	0.270081,	0.269781,	0.269483,	0.269185,	0.268888,	0.268589,	0.268293,	0.267995,	0.267699,	0.267402,	0.267106,	0.266810,	0.266514,	0.266219,	0.265924,	0.265630,	0.265334,	0.265040,	0.264746,	0.264453,	0.264160,	0.263866,	0.263573,	0.263281,	0.262988,	0.262696,	0.262405,	0.262113,	0.261822,	0.261531,	0.261241,	0.260950,	0.260660,	0.260370,	0.260081,	0.259793,	0.259504,	0.259215,	0.258926,	0.258638,	0.258351,	0.258063,	0.257776,	0.257488,	0.257203,	0.256916,	0.256629,	0.256344,	0.256058,	0.255774,	0.255488,	0.255204,	0.254919,	0.254636,	0.254351,	0.254068,	0.253785,	0.253502,	0.253219,	0.252937,	0.252655,	0.252372,	0.252091,	0.251810,	0.251530,	0.251248,	0.250968,	0.250688,	0.250408,	0.250129,	0.249849,	0.249570,	0.249291,	0.249013,	0.248735,	0.248457,	0.248180,	0.247902,	0.247625,	0.247349,	0.247072,	0.246796,	0.246520,	0.246244,	0.245970,	0.245695,	0.245420,	0.245145,	0.244871,	0.244598,	0.244324,	0.244050,	0.243777,	0.243505,	0.243232,	0.242959,	0.242688,	0.242416,	0.242145,	0.241874,	0.241603,	0.241332,	0.241063,	0.240792,	0.240523,	0.240253,	0.239984,	0.239715,	0.239447,	0.239179,	0.238910,	0.238643,	0.238376,	0.238109,	0.237841,	0.237575,	0.237309,	0.237043,	0.236777,	0.236511,	0.236246,	0.235981,	0.235716,	0.235452,	0.235188,	0.234924,	0.234661,	0.234398,	0.234135,	0.233872,	0.233609,	0.233347,	0.233085,	0.232823,	0.232563,	0.232302,	0.232041,	0.231780,	0.231520,	0.231260,	0.231000,	0.230741,	0.230481,	0.230223,	0.229964,	0.229707,	0.229448,	0.229190,	0.228933,	0.228675,	0.228419,	0.228162,	0.227906,	0.227649,	0.227394,	0.227138,	0.226883,	0.226628,	0.226373,	0.226119,	0.225864,	0.225610,	0.225357,	0.225104,	0.224850,	0.224597,	0.224345,	0.224093,	0.223840,	0.223588,	0.223337,	0.223085,	0.222834,	0.222584,	0.222333,	0.222083,	0.221833,	0.221584,	0.221334,	0.221085,	0.220836,	0.220587,	0.220339,	0.220091,	0.219843,	0.219595,	0.219349,	0.219101,	0.218854,	0.218608,	0.218361,	0.218116,	0.217870,	0.217624,	0.217378,	0.217134,	0.216889,	0.216644,	0.216401,	0.216156,	0.215913,	0.215669,	0.215427,	0.215183,	0.214941,	0.214698,	0.214457,	0.214214,	0.213973,	0.213731,	0.213490,	0.213250,	0.213009,	0.212769,	0.212529,	0.212288,	0.212049,	0.211810,	0.211571,	0.211331,	0.211093,	0.210855,	0.210617,	0.210379,	0.210141,	0.209904,	0.209667,	0.209430,	0.209193,	0.208957,	0.208721,	0.208485,	0.208249,	0.208014,	0.207779,	0.207544,	0.207309,	0.207075,	0.206841,	0.206608,	0.206374,	0.206140,	0.205907,	0.205674,	0.205441,	0.205209,	0.204977,	0.204745,	0.204514,	0.204282,	0.204050,	0.203819,	0.203589,	0.203358,	0.203129,	0.202898,	0.202668,	0.202439,	0.202209,	0.201981,	0.201751,	0.201523,	0.201294,	0.201067,	0.200838,	0.200611,	0.200384,	0.200156,	0.199930,	0.199704,	0.199476,	0.199251,	0.199025,	0.198798,	0.198573,	0.198348,	0.198123,	0.197898,	0.197673,	0.197449,	0.197225,	0.197001,	0.196778,	0.196553,	0.196330,	0.196107,	0.195885,	0.195662,	0.195440,	0.195218,	0.194996,	0.194774,	0.194553,	0.194332,	0.194111,	0.193890,	0.193669,	0.193450,	0.193230,	0.193010,	0.192790,	0.192571,	0.192351,	0.192133,	0.191914,	0.191695,	0.191477,	0.191260,	0.191041,	0.190824,	0.190606,	0.190389,	0.190172,	0.189956,	0.189739,	0.189522,	0.189307,	0.189091,	0.188876,	0.188660,	0.188445,	0.188230,	0.188015,	0.187800,	0.187586,	0.187372,	0.187158,	0.186944,	0.186731,	0.186518,	0.186304,	0.186092,	0.185878,	0.185666,	0.185454,	0.185243,	0.185030,	0.184819,	0.184608,	0.184397,	0.184185,	0.183975,	0.183764,	0.183554,	0.183344,	0.183134,	0.182924,	0.182715,	0.182505,	0.182296,	0.182088,	0.181879,	0.181671,	0.181462,	0.181254,	0.181046,	0.180839,	0.180631,	0.180424,	0.180217,	0.180010,	0.179803,	0.179597,	0.179392,	0.179185,	0.178979,	0.178774,	0.178568,	0.178363,	0.178159,	0.177954,	0.177749,	0.177544,	0.177341,	0.177136,	0.176932,	0.176729,	0.176526,	0.176322,	0.176120,	0.175917,	0.175714,	0.175512,	0.175309,	0.175108,	0.174905,	0.174703,	0.174502,	0.174301,	0.174100,	0.173900,	0.173699,	0.173499,	0.173298,	0.173098,	0.172898,	0.172699,	0.172498,	0.172300,	0.172101,	0.171901,	0.171703,	0.171505,	0.171306,	0.171108,	0.170910,	0.170713,	0.170514,	0.170317,	0.170120,	0.169923,	0.169727,	0.169529,	0.169333,	0.169137,	0.168941,	0.168745,	0.168550,	0.168353,	0.168158,	0.167963,	0.167768,	0.167573,	0.167379,	0.167185,	0.166990,	0.166796,	0.166603,	0.166409,	0.166215,	0.166022,	0.165829,	0.165637,	0.165444,	0.165251,	0.165059,	0.164867,	0.164675,	0.164483,	0.164292,	0.164100,	0.163909,	0.163717,	0.163526,	0.163336,	0.163146,	0.162955,	0.162765,	0.162575,	0.162385,	0.162195,	0.162006,	0.161817,	0.161627,	0.161439,	0.161250,	0.161062,	0.160873,	0.160685,	0.160497,	0.160309,	0.160122,	0.159934,	0.159747,	0.159559,	0.159373,	0.159186,	0.158999,	0.158812,	0.158627,	0.158441,	0.158255,	0.158069,	0.157883,	0.157698,	0.157513,	0.157328,	0.157143,	0.156959,	0.156774,	0.156589,	0.156406,	0.156221,	0.156037,	0.155854,	0.155671,	0.155488,	0.155304,	0.155121,	0.154939,	0.154756,	0.154574,	0.154391,	0.154209,	0.154027,	0.153845,	0.153664,	0.153483,	0.153302,	0.153121,	0.152940,	0.152759,	0.152579,	0.152398,	0.152218,	0.152038,	0.151858,	0.151678,	0.151499,	0.151320,	0.151140,	0.150961,	0.150782,	0.150604,	0.150425,	0.150247,	0.150069,	0.149891,	0.149713,	0.149535,	0.149358,	0.149181,	0.149004,	0.148826,	0.148649,	0.148474,	0.148297,	0.148120,	0.147944,	0.147768,	0.147592,	0.147416,	0.147242,	0.147066,	0.146890,	0.146716,	0.146541,	0.146367,	0.146192,	0.146018,	0.145844,	0.145669,	0.145496,	0.145322,	0.145149,	0.144975,	0.144803,	0.144629,	0.144457,	0.144285,	0.144111,	0.143940,	0.143767,	0.143595,	0.143423,	0.143252,	0.143080,	0.142909,	0.142738,	0.142567,	0.142396,	0.142225,	0.142055,	0.141884,	0.141714,	0.141544,	0.141375,	0.141204,	0.141035,	0.140866,	0.140697,	0.140528,	0.140358,	0.140190,	0.140021,	0.139853,	0.139685,	0.139517,	0.139349,	0.139181,	0.139013,	0.138846,	0.138679,	0.138512,	0.138345,	0.138178,	0.138012,	0.137845,	0.137679,	0.137513,	0.137347,	0.137181,	0.137015,	0.136850,	0.136684,	0.136519,	0.136355,	0.136190,	0.136025,	0.135861,	0.135696,	0.135532,	0.135368,	0.135205,	0.135041,	0.134877,	0.134714,	0.134550,	0.134388,	0.134225,	0.134062,	0.133900,	0.133737,	0.133575,	0.133412,	0.133251,	0.133089,	0.132927,	0.132766,	0.132605,	0.132443,	0.132283,	0.132122,	0.131962,	0.131801,	0.131641,	0.131481,	0.131320,	0.131161,	0.131001,	0.130842,	0.130682,	0.130523,	0.130365,	0.130205,	0.130047,	0.129888,	0.129730,	0.129571,	0.129414,	0.129256,	0.129098,	0.128940,	0.128782,	0.128625,	0.128469,	0.128311,	0.128155,	0.127998,	0.127842,	0.127685,	0.127529,	0.127374,	0.127218,	0.127062,	0.126906,	0.126751,	0.126596,	0.126441,	0.126286,	0.126131,	0.125977,	0.125822,	0.125668,	0.125514,	0.125361,	0.125207,	0.125052,	0.124899,	0.124746,	0.124593,	0.124440,	0.124287,	0.124135,	0.123982,	0.123830,	0.123678,	0.123526,	0.123374,	0.123222,	0.123070,	0.122920,	0.122769,	0.122618,	0.122467,	0.122316,	0.122165,	0.122014,	0.121864,	0.121715,	0.121565,	0.121415,	0.121265,	0.121115,	0.120967,	0.120817,	0.120668,	0.120519,	0.120371,	0.120222,	0.120074,	0.119926,	0.119778,	0.119630,	0.119483,	0.119335,	0.119187,	0.119041,	0.118893,	0.118747,	0.118600,	0.118453,	0.118307,	0.118160,	0.118015,	0.117868,	0.117723,	0.117577,	0.117432,	0.117286,	0.117141,	0.116996,	0.116851,	0.116706,	0.116562,	0.116418,	0.116273,	0.116130,	0.115985,	0.115842,	0.115699,	0.115555,	0.115412,	0.115269,	0.115125,	0.114983,	0.114841,	0.114698,	0.114555,	0.114413,	0.114272,	0.114130,	0.113988,	0.113846,	0.113705,	0.113564,	0.113423,	0.113283,	0.113141,	0.113001,	0.112861,	0.112720,	0.112580,	0.112441,	0.112301,	0.112161,	0.112022,	0.111883,	0.111743,	0.111604,	0.111465,	0.111326,	0.111188,	0.111051,	0.110911,	0.110774,	0.110636,	0.110499,	0.110361,	0.110223,	0.110086,	0.109949,	0.109812,	0.109675,	0.109538,	0.109403,	0.109266,	0.109130,	0.108994,	0.108858,	0.108723,	0.108587,	0.108452,	0.108316,	0.108182,	0.108047,	0.107912,	0.107777,	0.107643,	0.107509,	0.107374,	0.107241,	0.107107,	0.106973,	0.106840,	0.106707,	0.106574,	0.106441,	0.106307,	0.106175,	0.106042,	0.105911,	0.105778,	0.105646,	0.105514,	0.105383,	0.105251,	0.105120,	0.104988,	0.104857,	0.104726,	0.104596,	0.104465,	0.104334,	0.104204,	0.104074,	0.103944,	0.103815,	0.103684,	0.103555,	0.103426,	0.103296,	0.103167,	0.103039,	0.102909,	0.102781,	0.102653,	0.102525,	0.102396,	0.102268,	0.102141,	0.102013,	0.101885,	0.101758,	0.101631,	0.101504,	0.101377,	0.101250,	0.101124,	0.100997,	0.100871,	0.100745,	0.100619,	0.100493,	0.100367,	0.100242,	0.100117,	0.099991,	0.099866,	0.099742,	0.099617,	0.099492,	0.099368,	0.099244,	0.099119,	0.098995,	0.098871,	0.098748,	0.098624,	0.098501,	0.098377,	0.098255,	0.098132,	0.098009,	0.097887,	0.097764,	0.097642,	0.097520,	0.097398,	0.097276,	0.097154,	0.097032,	0.096912,	0.096790,	0.096669,	0.096549,	0.096428,	0.096307,	0.096186,	0.096066,	0.095946,	0.095826,	0.095706,	0.095586,	0.095467,	0.095348,	0.095228,	0.095110,	0.094990,	0.094872,	0.094753,	0.094635,	0.094516,	0.094397,	0.094280,	0.094161,	0.094044,	0.093926,	0.093809,	0.093691,	0.093575,	0.093458,	0.093341,	0.093224,	0.093107,	0.092991,	0.092875,	0.092759,	0.092643,	0.092526,	0.092411,	0.092296,	0.092181,	0.092065,	0.091950,	0.091836,	0.091720,	0.091606,	0.091491,	0.091377,	0.091262,	0.091148,	0.091035,	0.090921,	0.090807,	0.090693,	0.090580,	0.090467,	0.090354,	0.090242,	0.090129,	0.090016,	0.089903,	0.089791,	0.089679,	0.089567,	0.089455,	0.089343,	0.089232,	0.089120,	0.089009,	0.088898,	0.088787,	0.088676,	0.088565,	0.088455,	0.088344,	0.088234,	0.088124,	0.088013,	0.087903,	0.087794,	0.087684,	0.087574,	0.087465,	0.087357,	0.087247,	0.087138,	0.087029,	0.086921,	0.086812,	0.086703,	0.086596,	0.086488,	0.086379,	0.086271,	0.086164,	0.086057,	0.085949,	0.085842,	0.085735,	0.085628,	0.085521,	0.085415,	0.085308,	0.085201,	0.085094,	0.084989,	0.084883,	0.084777,	0.084671,	0.084565,	0.084460,	0.084354,	0.084249,	0.084144,	0.084039,	0.083935,	0.083830,	0.083726,	0.083621,	0.083517,	0.083412,	0.083309,	0.083204,	0.083101,	0.082997,	0.082894,	0.082790,	0.082687,	0.082583,	0.082481,	0.082377,	0.082275,	0.082173,	0.082070,	0.081968,	0.081866,	0.081763,	0.081662,	0.081559,	0.081458,	0.081357,	0.081256,	0.081154,	0.081053,	0.080952,	0.080850,	0.080750,	0.080649,	0.080549,	0.080448,	0.080348,	0.080248,	0.080148,	0.080047,	0.079948,	0.079848,	0.079749,	0.079650,	0.079551,	0.079451,	0.079352,	0.079253,	0.079154,	0.079056,	0.078957,	0.078859,	0.078761,	0.078663,	0.078564,	0.078466,	0.078368,	0.078270,	0.078173,	0.078075,	0.077978,	0.077881,	0.077784,	0.077687,	0.077590,	0.077493,	0.077396,	0.077300,	0.077203,	0.077107,	0.077011,	0.076914,	0.076819,	0.076723,	0.076628,	0.076532,	0.076436,	0.076341,	0.076245,	0.076150,	0.076056,	0.075961,	0.075866,	0.075771,	0.075676,	0.075582,	0.075488,	0.075394,	0.075299,	0.075205,	0.075111,	0.075018,	0.074924,	0.074830,	0.074736,	0.074644,	0.074550,	0.074457,	0.074363,	0.074271,	0.074178,	0.074085,	0.073993,	0.073900,	0.073807,	0.073716,	0.073623,	0.073531,	0.073440,	0.073347,	0.073256,	0.073164,	0.073072,	0.072981,	0.072890,	0.072799,	0.072707,	0.072616,	0.072525,	0.072434,	0.072344,	0.072253,	0.072163,	0.072072,	0.071982,	0.071891,	0.071802,	0.071711,	0.071622,	0.071531,	0.071442,	0.071352,	0.071263,	0.071173,	0.071084,	0.070994,	0.070905,	0.070817,	0.070727,	0.070639,	0.070549,	0.070461,	0.070373,	0.070284,	0.070196,	0.070108,	0.070019,	0.069931,	0.069842,	0.069755,	0.069667,	0.069579,	0.069492,	0.069404,	0.069316,	0.069229,	0.069142,	0.069055,	0.068967,	0.068881,	0.068794,	0.068706,	0.068620,	0.068533,	0.068447,	0.068361,	0.068274,	0.068187,	0.068101,	0.068016,	0.067929,	0.067843,	0.067757,	0.067672,	0.067586,	0.067499,	0.067414,	0.067329,	0.067244,	0.067158,	0.067073,	0.066987,	0.066902,	0.066817,	0.066733,	0.066648,	0.066563,	0.066479,	0.066394,	0.066310,	0.066224,	0.066140,	0.066056,	0.065972,	0.065888,	0.065804,	0.065720,	0.065636,	0.065552,	0.065469,	0.065385,	0.065301,	0.065218,	0.065134,	0.065051,	0.064968,	0.064885,	0.064801,	0.064718,	0.064635,	0.064552,	0.064470,	0.064387,	0.064304,	0.064221,	0.064139,	0.064056,	0.063974,	0.063891,	0.063809,	0.063727,	0.063644,	0.063562,	0.063481,	0.063399,	0.063317,	0.063235,	0.063153,	0.063071,	0.062990,	0.062908,	0.062826,	0.062746,	0.062664,	0.062583,	0.062502,	0.062420,	0.062339,	0.062258,	0.062178,	0.062097,	0.062016,	0.061935,	0.061854,	0.061773,	0.061693,	0.061612,	0.061532,	0.061451,	0.061370,	0.061291,	0.061210,	0.061130,	0.061050,	0.060969,	0.060890,	0.060810,	0.060730,	0.060650,	0.060571,	0.060491,	0.060411,	0.060331,	0.060252,	0.060172,	0.060092,	0.060013,	0.059934,	0.059855,	0.059775,	0.059697,	0.059617,	0.059538,	0.059458,	0.059380,	0.059301,	0.059222,	0.059144,	0.059065,	0.058985,	0.058907,	0.058829,	0.058750,	0.058672,	0.058593,	0.058514,	0.058436,	0.058358,	0.058279,	0.058202,	0.058123,	0.058046,	0.057967,	0.057889,	0.057811,	0.057733,	0.057655,	0.057577,	0.057499,	0.057422,	0.057344,	0.057266,	0.057189,	0.057111,	0.057034,	0.056956,	0.056878,	0.056801,	0.056723,	0.056647,	0.056569,	0.056491,	0.056415,	0.056337,	0.056260,	0.056183,	0.056106,	0.056029,	0.055953,	0.055875,	0.055798,	0.055721,	0.055644,	0.055568,	0.055491,	0.055415,	0.055337,	0.055261,	0.055184,	0.055108,	0.055031,	0.054955,	0.054878,	0.054802,	0.054726,	0.054650,	0.054573,	0.054497,	0.054420,	0.054345,	0.054268,	0.054192,	0.054116,	0.054040,	0.053964,	0.053888,	0.053812,	0.053736,	0.053660,	0.053585,	0.053508,	0.053433,	0.053357,	0.053281,	0.053205,	0.053130,	0.053055,	0.052978,	0.052903,	0.052827,	0.052752,	0.052676,	0.052601,	0.052525,	0.052450,	0.052375,	0.052299,	0.052224,	0.052149,	0.052074,	0.051998,	0.051923,	0.051848,	0.051773,	0.051698,	0.051622,	0.051548,	0.051472,	0.051397,	0.051323,	0.051247,	0.051173,	0.051097,	0.051023,	0.050948,	0.050873,	0.050799,	0.050723,	0.050649,	0.050574,	0.050499,	0.050425,	0.050351,	0.050275,	0.050201,	0.050126,	0.050052,	0.049978,	0.049902,	0.049828,	0.049753,	0.049679,	0.049605,	0.049530,	0.049456,	0.049382,	0.049307,	0.049233,	0.049159,	0.049084,	0.049010,	0.048937,	0.048862,	0.048788,	0.048713,	0.048639,	0.048566,	0.048491,	0.048417,	0.048343,	0.048269,	0.048195,	0.048121,	0.048047,	0.047973,	0.047899,	0.047825,	0.047751,	0.047678,	0.047603,	0.047530,	0.047456,	0.047382,	0.047308,	0.047235,	0.047162,	0.047087,	0.047014,	0.046940,	0.046866,	0.046793,	0.046719,	0.046645,	0.046572,	0.046499,	0.046424,	0.046351,	0.046278,	0.046204,	0.046131,	0.046057,	0.045984,	0.045910,	0.045837,	0.045764,	0.045690,	0.045617,	0.045544,	0.045471,	0.045397,	0.045324,	0.045251,	0.045177,	0.045104,	0.045031,	0.044958,	0.044884,	0.044811,	0.044738,	0.044665,	0.044592,	0.044519,	0.044446,	0.044372,	0.044300,	0.044227,	0.044154,	0.044080,	0.044008,	0.043935,	0.043862,	0.043789,	0.043716,	0.043644,	0.043571,	0.043497,	0.043425,	0.043352,	0.043280,	0.043206,	0.043134,	0.043061,	0.042989,	0.042915,	0.042843,	0.042770,	0.042698,	0.042625,	0.042552,	0.042480,	0.042408,	0.042334,	0.042262,	0.042190,	0.042117,	0.042044,	0.041972,	0.041900,	0.041827,	0.041755,	0.041682,	0.041610,	0.041538,	0.041466,	0.041394,	0.041321,	0.041248,	0.041176,	0.041104,	0.041031,	0.040959,	0.040887,	0.040815,	0.040744,	0.040671,	0.040599,	0.040527,	0.040455,	0.040383,	0.040310,	0.040238,	0.040167,	0.040095,	0.040023,	0.039950,	0.039879,	0.039807,	0.039735,	0.039664,	0.039592,	0.039519,	0.039448,	0.039376,	0.039305,	0.039233,	0.039162,	0.039089,	0.039018,	0.038946,	0.038875,	0.038803,	0.038732,	0.038659,	0.038588,	0.038517,	0.038445,	0.038374,	0.038303,	0.038231,	0.038159,	0.038088,	0.038017,	0.037946,	0.037875,	0.037804,	0.037731,	0.037660,	0.037589,	0.037518,	0.037447,	0.037376,	0.037305,	0.037233,	0.037162,	0.037092,	0.037021,	0.036950,	0.036879,	0.036808,	0.036738,	0.036667,	0.036595,	0.036524,	0.036454,	0.036383,	0.036312,	0.036242,	0.036171,	0.036101,	0.036030,	0.035960,	0.035888,	0.035818,	0.035748,	0.035677,	0.035607,	0.035537,	0.035466,	0.035396,	0.035326,	0.035256,	0.035185,	0.035115,	0.035045,	0.034975,	0.034904,	0.034834,	0.034764,	0.034694,	0.034624,	0.034554,	0.034484,	0.034414,	0.034345,	0.034275,	0.034205,	0.034135,	0.034066,	0.033996,	0.033927,	0.033857,	0.033787,	0.033718,	0.033648,	0.033579,	0.033510,	0.033440,	0.033371,	0.033302,	0.033232,	0.033163,	0.033094,	0.033025,	0.032955,	0.032886,	0.032817,	0.032748,	0.032679,	0.032610,	0.032541,	0.032473,	0.032404,	0.032335,	0.032266,	0.032197,	0.032129,	0.032060,	0.031992,	0.031923,	0.031855,	0.031787,	0.031719,	0.031650,	0.031582,	0.031513,	0.031445,	0.031377,	0.031309,	0.031241,	0.031173,	0.031104,	0.031037,	0.030969,	0.030902,	0.030834,	0.030766,	0.030698,	0.030630,	0.030562,	0.030495,	0.030428,	0.030361,	0.030293,	0.030226,	0.030158,	0.030091,	0.030023,	0.029957,	0.029890,	0.029823,	0.029756,	0.029688,	0.029621,	0.029555,	0.029488,	0.029422,	0.029355,	0.029288,	0.029222,	0.029155,	0.029089,	0.029022,	0.028956,	0.028890,	0.028824,	0.028757,	0.028691,	0.028626,	0.028559,	0.028493,	0.028427,	0.028362,	0.028296,	0.028230,	0.028164,	0.028099,	0.028033,	0.027968,	0.027902,	0.027837,	0.027772,	0.027706,	0.027642,	0.027576,	0.027511,	0.027447,	0.027381,	0.027316,	0.027252,	0.027187,	0.027122,	0.027058,	0.026993,	0.026929,	0.026865,	0.026800,	0.026735,	0.026672,	0.026607,	0.026543,	0.026480,	0.026415,	0.026352,	0.026288,	0.026224,	0.026161,	0.026097,	0.026034,	0.025970,	0.025907,	0.025843,	0.025780,	0.025717,	0.025653,	0.025591,	0.025527,	0.025465,	0.025402,	0.025340,	0.025276,	0.025214,	0.025151,	0.025089,	0.025027,	0.024965,	0.024902,	0.024840,	0.024778,	0.024716,	0.024654,	0.024592,	0.024530,	0.024469,	0.024407,	0.024346,	0.024285,	0.024223,	0.024162,	0.024100,	0.024039,	0.023978,	0.023917,	0.023856,	0.023795,	0.023735,	0.023673,	0.023613,	0.023553,	0.023492,	0.023432,	0.023372,	0.023311,	0.023251,	0.023192,	0.023131,	0.023072,	0.023012,	0.022952,	0.022893,	0.022833,	0.022774,	0.022714,	0.022655,	0.022596,	0.022536,	0.022478,	0.022419,	0.022360,	0.022301,	0.022242,	0.022184,	0.022126,	0.022068,	0.022009,	0.021951,	0.021893,	0.021835,	0.021777,	0.021718,	0.021661,	0.021603,	0.021546,	0.021488,	0.021431,	0.021374,	0.021316,	0.021259,	0.021202,	0.021145,	0.021088,	0.021032,	0.020975,	0.020919,	0.020862,	0.020806,	0.020750,	0.020692,	0.020636,	0.020580,	0.020524,	0.020469,	0.020413,	0.020357,	0.020302,	0.020246,	0.020191,	0.020136,	0.020080,	0.020025,	0.019970,	0.019915,	0.019860,	0.019807,	0.019752,	0.019697,	0.019643,	0.019589,	0.019534,	0.019480,	0.019426,	0.019372,	0.019318,	0.019264,	0.019211,	0.019158,	0.019104,	0.019051,	0.018997,	0.018944,	0.018892,	0.018839,	0.018785,	0.018733,	0.018680,	0.018628,	0.018575,	0.018523,	0.018470,	0.018418,	0.018367,	0.018314,	0.018262,	0.018210,	0.018159,	0.018107,	0.018056,	0.018005,	0.017953,	0.017902,	0.017852,	0.017800,	0.017749,	0.017699,	0.017648,	0.017597,	0.017547,	0.017497,	0.017446,	0.017397,	0.017346,	0.017297,	0.017247,	0.017196,	0.017147,	0.017097,	0.017049,	0.016999,	0.016949,	0.016901,	0.016851,	0.016803,	0.016753,	0.016705,	0.016656,	0.016608,	0.016559,	0.016511,	0.016463,	0.016415,	0.016366,	0.016319,	0.016272,	0.016223,	0.016176,	0.016128,	0.016081,	0.016033,	0.015986,	0.015940,	0.015892,	0.015846,	0.015798,	0.015752,	0.015706,	0.015658,	0.015612,	0.015566,	0.015520,	0.015474,	0.015428,	0.015382,	0.015336,	0.015291,	0.015245,	0.015199,	0.015154,	0.015109,	0.015063,	0.015018,	0.014974,	0.014929,	0.014883,	0.014839,	0.014795,	0.014750,	0.014706,	0.014661,	0.014617,	0.014573,	0.014529,	0.014486,	0.014442,	0.014397,	0.014354,	0.014311,	0.014267,	0.014224,	0.014181,	0.014138,	0.014095,	0.014051,	0.014009,	0.013966,	0.013923,	0.013881,	0.013839,	0.013796,	0.013754,	0.013712,	0.013670,	0.013628,	0.013586,	0.013544,	0.013503,	0.013461,	0.013420,	0.013378,	0.013337,	0.013296,	0.013255,	0.013213,	0.013172,	0.013132,	0.013091,	0.013050,	0.013009,	0.012969,	0.012928,	0.012889,	0.012849,	0.012808,	0.012768,	0.012728,	0.012688,	0.012648,	0.012608,	0.012570,	0.012530,	0.012491,	0.012451,	0.012412,	0.012372,	0.012334,	0.012295,	0.012256,	0.012217,	0.012178,	0.012140,	0.012101,	0.012062,	0.012024,	0.011985,	0.011948,	0.011909,	0.011871,	0.011833,	0.011796,	0.011757,	0.011719,	0.011682,	0.011645,	0.011607,	0.011569,	0.011532,	0.011495,	0.011457,	0.011421,	0.011383,	0.011346,	0.011310,	0.011273,	0.011236,	0.011200,	0.011163,	0.011126,	0.011090,	0.011053,	0.011017,	0.010981,	0.010944,	0.010909,	0.010872,	0.010837,	0.010801,	0.010764,	0.010729,	0.010693,	0.010658,	0.010622,	0.010587,	0.010551,	0.010515,	0.010481,	0.010445,	0.010410,	0.010375,	0.010340,	0.010305,	0.010270,	0.010235,	0.010201,	0.010166,	0.010131,	0.010096,	0.010062,	0.010027,	0.009993,	0.009958,	0.009925,	0.009890,	0.009856,	0.009821,	0.009788,	0.009753,	0.009720,	0.009686,	0.009652,	0.009618,	0.009584,	0.009551,	0.009516,	0.009483,	0.009450,	0.009416,	0.009383,	0.009349,	0.009316,	0.009283,	0.009249,	0.009216,	0.009182,	0.009150,	0.009117,	0.009083,	0.009051,	0.009017,	0.008984,	0.008952,	0.008918,	0.008886,	0.008854,	0.008820,	0.008788,	0.008754,	0.008722,	0.008690,	0.008657,	0.008625,	0.008592,	0.008559,	0.008527,	0.008495,	0.008462,	0.008430,	0.008398,	0.008365,	0.008333,	0.008302,	0.008269,	0.008237,	0.008205,	0.008172,	0.008141,	0.008109,	0.008077,	0.008045,	0.008013,	0.007981,	0.007949,	0.007917,	0.007886,	0.007853,	0.007822,	0.007790,	0.007759,	0.007726,	0.007695,	0.007663,	0.007631,	0.007600,	0.007568,	0.007537,	0.007504,	0.007473,	0.007442,	0.007410,	0.007378,	0.007347,	0.007316,	0.007284,	0.007252,	0.007221,	0.007190,	0.007158,	0.007127,	0.007095,	0.007063,	0.007032,	0.007001,	0.006970,	0.006938,	0.006907,	0.006875,	0.006844,	0.006812,	0.006781,	0.006750,	0.006719,	0.006687,	0.006656,	0.006625,	0.006594,	0.006562,	0.006531,	0.006499,	0.006467,	0.006436,	0.006405,	0.006374,	0.006342,	0.006311,	0.006280,	0.006249,	0.006217,	0.006186,	0.006155,	0.006124,	0.006092,	0.006061,	0.006030,	0.005998,	0.005966,	0.005935,	0.005904,	0.005873,	0.005841,	0.005810,	0.005779,	0.005748,	0.005716,	0.005685,	0.005653,	0.005621,	0.005590,	0.005559,	0.005528,	0.005496,	0.005465,	0.005433,	0.005402,	0.005370,	0.005339,	0.005308,	0.005277,	0.005244,	0.005213,	0.005182,	0.005150,	0.005119,	0.005087,	0.005056,	0.005024,	0.004993,	0.004961,	0.004930,	0.004898,	0.004866,	0.004835,	0.004803,	0.004772,	0.004740,	0.004709,	0.004677,	0.004645,	0.004614,	0.004583,	0.004550,	0.004519,	0.004488,	0.004455,	0.004424,	0.004393,	0.004361,	0.004329,	0.004298,	0.004266,	0.004234,	0.004203,	0.004171,	0.004140,	0.004107,	0.004076,	0.004045,	0.004012,	0.003981,	0.003950,	0.003918,	0.003886,	0.003855,	0.003823,	0.003791,	0.003760,	0.003728,	0.003697,	0.003665,	0.003633,	0.003602,	0.003571,	0.003538,	0.003507,	0.003476,	0.003444,	0.003412,	0.003381,	0.003350,	0.003318,	0.003287,	0.003255,	0.003223,	0.003192,	0.003161,	0.003130,	0.003098,	0.003067,	0.003036,	0.003005,	0.002973,	0.002942,	0.002911,	0.002880,	0.002848,	0.002818,	0.002787,	0.002756,	0.002725,	0.002694,	0.002663,	0.002632,	0.002602,	0.002570,	0.002540,	0.002510,	0.002479,	0.002449,	0.002418,	0.002387,	0.002357,	0.002327,	0.002297,	0.002266,	0.002236,	0.002206,	0.002177,	0.002147,	0.002117,	0.002088,	0.002057,	0.002028,	0.001998,	0.001969,	0.001940,	0.001911,	0.001882,	0.001853,	0.001824,	0.001794,	0.001765,	0.001737,	0.001708,	0.001680,	0.001652,	0.001623,	0.001595,	0.001567,	0.001539,	0.001512,	0.001484,	0.001457,	0.001429,	0.001402,	0.001375,	0.001348,	0.001321,	0.001294,	0.001268,	0.001242,	0.001216,	0.001190,	0.001164,	0.001138,	0.001112,	0.001087,	0.001062,	0.001037,	0.001012,	0.000987,	0.000963,	0.000939,	0.000915,	0.000891,	0.000868,	0.000844,	0.000820,	0.000798,	0.000775,	0.000752,	0.000730,	0.000708,	0.000687,	0.000665,	0.000643,	0.000622,	0.000601,	0.000581,	0.000560,	0.000540,	0.000521,	0.000501,	0.000482,	0.000462,	0.000444,	0.000426,	0.000407,	0.000390,	0.000373,	0.000356,	0.000338,	0.000322,	0.000306,	0.000291,	0.000275,	0.000260,	0.000246,	0.000231,	0.000217,	0.000204,	0.000191,	0.000178,	0.000165,	0.000153,	0.000141,	0.000130,	0.000120,	0.000109,	0.000099,	0.000089,	0.000080,	0.000071,	0.000064,	0.000056,	0.000048,	0.000042,	0.000036,	0.000029,	0.000024,	0.000019,	0.000015,	0.000011,	0.000008,	0.000005,	0.000003,	0.000001,	0.000000,	0.000000,	0.000000,	0.000000,	0.000001,	0.000003,	0.000005,	0.000009,	0.000012,	0.000016,	0.000022,	0.000027,	0.000033,	0.000040,	0.000047,	0.000056,	0.000065,	0.000074,	0.000084,	0.000095,	0.000107,	0.000120,	0.000133,	0.000147,	0.000162,	0.000177,	0.000193,	0.000210,	0.000227,	0.000247,	0.000266,	0.000287,	0.000307,	0.000330,	0.000352,	0.000376,	0.000401,	0.000426,	0.000453,	0.000480,	0.000508,	0.000537,	0.000567,	0.000598,	0.000630,	0.000664,	0.000697,	0.000733,	0.000768,	0.000805,	0.000844,	0.000883,	0.000924,	0.000965,	0.001007,	0.001051,	0.001095,	0.001141,	0.001188,	0.001236,	0.001285,	0.001335,	0.001387,	0.001440,	0.001494,	0.001550,	0.001606,	0.001664,	0.001722,	0.001783];

    freqsPerfiEEG = np.array(freqsPerf_iEEG)
    logPowerPerfiEEG = np.array(logPowerPerf_iEEG)

    return freqsPerfiEEG, logPowerPerfiEEG

if __name__ == '__main__':

    fs = 1024
    mtg_labels = ['ch1','ch2','ch3','ch4','ch5']
    data = np.random.rand(len(mtg_labels), fs*3600)
    mtg_eeg_data = {'fs':1024, 'mtg_labels':mtg_labels, 'data':data}
    time_vec, chspec_ni_vec, chavg_ni_vec = get_noise_index_vec(fs=fs, mtg_labels=mtg_labels,
                        mtg_signals=mtg_eeg_data['data'], notched=False, notch_freqs=None, is_ieeg=False)

    pass
